/*
								+----------------------------------+
								|                                  |
								|  *** Atmosphere phys model  ***  |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|          Atmosphere.cpp          |
								|                                  |
								+----------------------------------+
*/

#include <assert.h>
#define _ASSERTE(x) assert(x)
#include <string.h>
#include <math.h>
#include "MiniVector.h"
#include "Atmosphere.h"

extern const float f_pi;

/*
 *								=== CAtmosphere::COpticalDepthFunction ===
 */

class CAtmosphere::COpticalDepthFunction {
protected:
	Vector3f m_v_org, m_v_step;
	float m_f_min_height, m_f_inv_height_scale;

public:
	inline COpticalDepthFunction(Vector3f v_org, Vector3f v_dir,
		float f_planet_radius, float f_inv_height_scale)
		:m_v_org(v_org), m_v_step(v_dir), m_f_min_height(f_planet_radius),
		m_f_inv_height_scale(f_inv_height_scale)
	{}

	inline float operator ()(float f_t) const
	{
		Vector3f v_pos = m_v_org + m_v_step * f_t;
		return (float)exp(-(v_pos.f_Length() - m_f_min_height) * m_f_inv_height_scale);
	}
};

/*
 *								=== ~CAtmosphere::COpticalDepthFunction ===
 */

/*
 *								=== CAtmosphere ===
 */

/*
 *	CAtmosphere::CAtmosphere(float f_planet_radius, float f_atmosphere_radius,
 *		Vector3f v_sun_dir, float f_u = .7f, float f_k_mie = .25e-9f,
 *		float f_k_raileigh = .15e-31f)
 *		- default constructor
 *		- initializes atmosphere with given values
 */
CAtmosphere::CAtmosphere(float f_planet_radius, float f_atmosphere_radius,
	Vector3f v_sun_dir, float f_u, float f_k_mie, float f_k_raileigh)
	:m_f_planet_radius(f_planet_radius), m_f_atmosphere_radius(f_atmosphere_radius),
	m_f_inv_radius_scale(1.0f / (f_atmosphere_radius - f_planet_radius)),
	m_v_sun_dir(v_sun_dir), m_f_k_raileigh(f_k_raileigh), m_f_k_mie(f_k_mie),
	m_f_4pi_k_raileigh(4 * f_pi * f_k_raileigh), m_f_4pi_k_mie(4 * f_pi * f_k_mie)
{
	float f_u2 = f_u * f_u;
	float f_x = (.55556f + .171468f * f_u2) * f_u +
		(float)sqrt(2.37037f + (-1.337449f + .571559f * f_u2) * f_u2);
	float f_x_one_third = (float)pow(f_x, .33333f);
	m_f_g = .55556f * f_u - (1.33333f - .308642f * f_u2) / f_x_one_third + f_x_one_third;
	m_f_g2 = m_f_g * m_f_g;
	// temps for phase equations
}

/*
 *	float CAtmosphere::f_Intensity(Vector3f v_camera, Vector3f v_target, float f_lambda) const
 *		- returns relative intensity for light of wavelength f_lambda meters
 *		  observed at v_camera and coming from v_target
 */
float CAtmosphere::f_Intensity(Vector3f v_camera, Vector3f v_target, float f_lambda) const
{
	Vector3f v_ray = v_target - v_camera;
	float f_far = v_ray.f_Length();
	v_ray /= f_far;
	// far distance is distance to the point

	float f_near;
	if(!IntersectSphere(v_camera, v_ray, m_f_atmosphere_radius, f_near))
		return 0; // looking to space, no light there
	// near distance is distance to atmosphere

	bool b_in_atmosphere;
	if(f_near < 0) {
		b_in_atmosphere = true;
		f_near = 0;
		// inside atmosphere
	} else {
		b_in_atmosphere = false;
		v_camera += v_ray * f_near;
		f_far -= f_near;
		f_near = 0;
		// outside atmosphere, move it just to touch it
	}
	// determine wheter camera is inside atmosphere

	double f_integral_raileigh = 0;
	double f_integral_mie = 0;
	// integral values

	double f_lambda4 = f_lambda * f_lambda;
	f_lambda4 *= f_lambda4;
	// precalc lambda^4

	double f_rl_depth_dcale = m_f_4pi_k_raileigh / f_lambda4;
	// precalc

	const int n_sample_num = 64;
	float f_sample_step = f_far / n_sample_num; // f_near is zero
	Vector3f v_pos = v_camera + v_ray * (.5f * f_sample_step);
	for(int i = 0; i < n_sample_num; ++ i, v_pos += v_ray * f_sample_step) {
		if(b_Shadow(v_pos))
			continue;
		// is this point in shadow?

		float f_sun_point_depth;
		float f_t;
		if(IntersectSphere(v_pos, -m_v_sun_dir, m_f_atmosphere_radius, f_t)) {
			Vector3f v_sun_pos = v_pos + m_v_sun_dir * -f_t;
			f_sun_point_depth = f_Depth(v_sun_pos, v_pos);
		} else
			f_sun_point_depth = 0;
		float f_camera_point_depth = f_Depth(v_camera, v_pos);
		// calculate atmosphere densities from sun and from camera

		double f_density_ratio = exp(-(v_pos.f_Length() - m_f_atmosphere_radius) *
			m_f_inv_radius_scale);
		// density ratio (ro at the start of the equation)

		double f_rl_depth = (f_camera_point_depth + f_sun_point_depth) * f_rl_depth_dcale;
		double f_mie_depth = f_sun_point_depth * m_f_4pi_k_mie;
		// calculate optical densities for mie and railieigh

		f_integral_raileigh += f_density_ratio * exp(-f_rl_depth);
		f_integral_mie += f_density_ratio * exp(-f_mie_depth);
		// sum it up
	}
	f_integral_raileigh *= f_sample_step;
	f_integral_mie *= f_sample_step;
	// integrate

	float f_cos = -v_ray.f_Dot(m_v_sun_dir);
	float f_cos2 = f_cos * f_cos;
	// calculate cosine of the angle

	float f_phase_raileigh = .75f * (1 + f_cos2);
	float f_phase_mie = 1.5f * ((1 - m_f_g2) * (1 + f_cos2)) /
		((2 + m_f_g2) * (float)pow(1 + m_f_g2 - 2 * m_f_g * f_cos, 1.5f));
	// calculate raileigh and mie phase functions

	return (float)(f_integral_raileigh * f_phase_raileigh * m_f_k_raileigh / f_lambda4 +
		f_integral_mie * f_phase_mie * m_f_k_mie);
	// return relative intensity
}

/*
 *	CIntensity CAtmosphere::Intensity(Vector3f v_camera, Vector3f v_target) const
 *		- returns relative intensity function for light observed at v_camera
 *		  and coming from v_target
 */
CAtmosphere::CIntensity CAtmosphere::Intensity(Vector3f v_camera, Vector3f v_target) const
{
	Vector3f v_ray = v_target - v_camera;
	float f_far = v_ray.f_Length();
	v_ray /= f_far;
	// far distance is distance to the point

	float f_near;
	if(!IntersectSphere(v_camera, v_ray, m_f_atmosphere_radius, f_near))
		return CIntensity(0, 0, 0, 0, 0); // looking to space, no light there
	// near distance is distance to atmosphere

	bool b_in_atmosphere;
	if(f_near < 0) {
		b_in_atmosphere = true;
		f_near = 0;
		// inside atmosphere
	} else {
		b_in_atmosphere = false;
		v_camera += v_ray * f_near;
		f_far -= f_near;
		f_near = 0;
		// outside atmosphere, move it just to touch it
	}
	// determine wheter camera is inside atmosphere

	double f_integral_mie = 0;
	// integral values

	const int n_sample_num = 64;

	double *p_raileigh_depth_density_ratio;
	if(!(p_raileigh_depth_density_ratio = new double[2 * n_sample_num]))
		return CIntensity(0, 0, 0, 0, 0);
	// alloc interleaved data buffer

	float f_sample_step = f_far / n_sample_num; // f_near is zero
	Vector3f v_pos = v_camera + v_ray * (.5f * f_sample_step);
	for(int i = 0; i < n_sample_num; ++ i, v_pos += v_ray * f_sample_step) {
		if(b_Shadow(v_pos)) {
			p_raileigh_depth_density_ratio[2 * i] = 0; // exp result is 1
			p_raileigh_depth_density_ratio[2 * i + 1] = 0; // 1 times 0 is 0 which is our purpose
			continue;
		}
		// is this point in shadow?

		float f_sun_point_depth;
		float f_t;
		if(IntersectSphere(v_pos, -m_v_sun_dir, m_f_atmosphere_radius, f_t)) {
			Vector3f v_sun_pos = v_pos + m_v_sun_dir * -f_t;
			f_sun_point_depth = f_Depth(v_sun_pos, v_pos);
		} else
			f_sun_point_depth = 0;
		float f_camera_point_depth = f_Depth(v_camera, v_pos);
		// calculate atmosphere densities from sun and from camera

		double f_density_ratio = exp(-(v_pos.f_Length() - m_f_atmosphere_radius) *
			m_f_inv_radius_scale);
		// density ratio (ro at the start of the equation)

		double f_mie_depth = f_sun_point_depth * m_f_4pi_k_mie;
		// calculate optical density for mie

		p_raileigh_depth_density_ratio[2 * i] = f_density_ratio;
		p_raileigh_depth_density_ratio[2 * i + 1] = f_camera_point_depth + f_sun_point_depth;
		// store precalculated part of integral

		f_integral_mie += f_density_ratio * exp(-f_mie_depth);
		// sum it up
	}
	f_integral_mie *= f_sample_step;
	// integrate

	float f_cos = -v_ray.f_Dot(m_v_sun_dir);
	float f_cos2 = f_cos * f_cos;
	// calculate cosine of the angle

	float f_phase_raileigh = .75f * (1 + f_cos2);
	float f_phase_mie = 1.5f * ((1 - m_f_g2) * (1 + f_cos2)) /
		((2 + m_f_g2) * (float)pow(1 + m_f_g2 - 2 * m_f_g * f_cos, 1.5f));
	// calculate raileigh and mie phase functions

	double f_raileigh_integral_multiplier = f_phase_raileigh * m_f_k_raileigh * f_sample_step;
	// raileigh integral is supposed to be multiplied by this and then divided by f_lambda4

	return CIntensity(n_sample_num, p_raileigh_depth_density_ratio, m_f_4pi_k_raileigh,
		f_raileigh_integral_multiplier, f_integral_mie * f_phase_mie * m_f_k_mie);
	// return the intensity function
}

/*
 *	float CAtmosphere::f_Depth(Vector3f v_a, Vector3f v_b) const
 *		- returns optical depth for atmosphere between points v_a and v_b,
 *		  given they both lie inside atmosphere
 */
float CAtmosphere::f_Depth(Vector3f v_a, Vector3f v_b) const
{
	/*do {
		float f_input_h = v_a.f_Length();
		if(fabs(f_input_h - m_f_atmosphere_radius) > 100) {
			float f_h2 = v_b.f_Length();
			if(f_h2 > f_input_h) {
				Vector3f v_tmp(v_b);
				v_b = v_a;
				v_a = v_tmp;
			}
			// make sure v_a is above v_b

			Vector3f v_ray(v_b - v_a);
			// ray goes -down-

			float f_t;
			IntersectSphere(v_b, v_ray, m_f_atmosphere_radius, f_t);
			Vector3f v_top = v_b + v_ray * f_t;
			// calculate point at the top of the atmosphere

			return f_Depth(v_top, v_b) - f_Depth(v_top, v_a);
			// subtract two depths
		}
	} while(0);*/
	// should be posible to subdivide to two intervals so one point is always on
	// top of the atmosphere, then it breaks to 2D function of second point height and
	// direction steepness which could be easily precalculated to 2D image

	v_a -= v_b;
	float f_length = v_a.f_Length();
	v_a /= f_length;
	return f_Integrate_s(COpticalDepthFunction(v_b, v_a,
		m_f_planet_radius, m_f_inv_radius_scale), 0, f_length, 1e-2f);
}

/*
 *	bool CAtmosphere::b_Shadow(Vector3f v_pos) const
 *		- returns true in case v_pos is in shadow of lower sphere, otherwise false
 */
bool CAtmosphere::b_Shadow(Vector3f v_pos) const
{
	float f_side = m_v_sun_dir.f_Dot(v_pos);
	if(f_side > 0)
		return false; // position is on the same side as sun

	Vector3f v_project = m_v_sun_dir * f_side;
	// project v_pos to m_v_sun_dir axis

	return (v_project - v_pos).f_Length2() <= m_f_planet_radius * m_f_planet_radius;
	// it's shadowed in case it lies inside cylinder
	// with axis m_v_sun_dir and radius m_f_planet_radius (widen the cylinder a little
	// bit so the crappy mesh is not shadowed so irregularly)
}

/*
 *	static bool CAtmosphere::IntersectSphere(const Vector3f &r_v_pos,
 *		const Vector3f &r_v_dir, float f_radius, float &r_f_intersection)
 *		- calculates intersection of ray starting at r_v_pos with r_v_dir direction
 *		  with sphere having center at O and radius f_radius
 *		- returns true in case intersection occurs, otherwise false
 *		- fills r_f_intersection with distance to the first intersection
 *		  (even if it's behind ray origin)
 */
bool CAtmosphere::IntersectSphere(const Vector3f &r_v_pos,
	const Vector3f &r_v_dir, float f_radius, float &r_f_intersection)
{
	float f_a = r_v_dir.f_Length2();
	float f_b = r_v_dir.f_Dot(r_v_pos);
	float f_c = r_v_pos.f_Length2() - f_radius * f_radius;
	float f_d = f_b * f_b - f_a * f_c;
	if(f_d < 0)
		return false;
	// quadratic equation of ray-sphere intersection

	r_f_intersection = (-f_b - (float)sqrt(f_d)) / f_a;
	// find closest root

	return true;
}

/*
 *								=== ~CAtmosphere ===
 */
