/*
								+----------------------------------+
								|                                  |
								|  *** Atmosphere phys model  ***  |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|           Atmosphere.h           |
								|                                  |
								+----------------------------------+
*/

/*
 *	this file was written based on Nishita's Siggraph '93 paper
 */

#ifndef __ATMOSPHERE_PHYSICS_INCLUDED
#define __ATMOSPHERE_PHYSICS_INCLUDED

/*
 *	template <class CFunctor, class CScalarType>
 *	class CIntegrator
 *		- simple one-dimensional integrator with max error parameter
 */
template <class CFunctor, class CScalarType>
class CIntegrator {
protected:
	CFunctor m_function;
	CScalarType m_f_min, m_f_scale;
	CScalarType m_f_max_error;

public:
	/*
	 *	inline CIntegrator(CFunctor function, CScalarType f_min,
	 *		CScalarType f_max, CScalarType f_max_error)
	 *		- default constructor; function is function being integrated over range f_min to f_max
	 *		  with maximal relative error f_max_error
	 */
	inline CIntegrator(CFunctor function, CScalarType f_min,
		CScalarType f_max, CScalarType f_max_error)
		:m_function(function), m_f_min(f_min), m_f_scale(f_max - f_min), m_f_max_error(f_max_error)
	{}

	inline operator CScalarType () const
	{
		CScalarType f_scale = m_f_scale;
		CScalarType f_first_integral((m_function(0) + m_function(1)) * f_scale * CScalarType(.5));
		// first integral approximation (trapezoid rule)

		for(int n_pass = 1; n_pass < 8; ++ n_pass) {
			CScalarType f_integral(f_first_integral * CScalarType(.5));
			// prepare integration variable

			f_scale *= CScalarType(.5);
			// scale is going down with every refinement

			CScalarType f_refinement(0);
			for(int i = 0, n_mask = (1 << n_pass) - 1, n = (1 << n_pass) + 1; i < n; ++ i) {
				if(i & n_mask)
					f_refinement += m_function(i * f_scale + m_f_min);
			}
			f_refinement *= f_scale;
			f_integral += f_refinement;
			// refine integral sollution

			if((f_first_integral >= 0 && fabs(f_first_integral - f_integral) <=
			   m_f_max_error * f_first_integral) || (f_first_integral < 0 &&
			   fabs(f_first_integral - f_integral) >= m_f_max_error * f_first_integral))
				return f_integral;
			// is error in required range now?

			f_first_integral = f_integral;
		}
		// adaptive subdivision to fit error

		return f_first_integral;
	}
};

/*
 *	template <class CFunctor>
 *	inline double f_Integrate(CFunctor function, double f_min, double f_max, double f_max_error)
 *		- integrates one-dimensional function over range f_min to f_max (inclusive)
 *		  while maintaining maximal relative error f_max_error
 *		- returns the value of the integral
 */
template <class CFunctor>
inline double f_Integrate(CFunctor function, double f_min, double f_max, double f_max_error)
{
	return CIntegrator<CFunctor, double>(function, f_min, f_max, f_max_error);
}

/*
 *	template <class CFunctor>
 *	inline float f_Integrate_s(CFunctor function, float f_min, float f_max, float f_max_error)
 *		- integrates one-dimensional function over range f_min to f_max (inclusive)
 *		  while maintaining maximal relative error f_max_error
 *		- returns the value of the integral
 */
template <class CFunctor>
inline float f_Integrate_s(CFunctor function, float f_min, float f_max, float f_max_error)
{
	return CIntegrator<CFunctor, float>(function, f_min, f_max, f_max_error);
}

/*
 *	class CAtmosphere
 *		- can calculate intensity of certain wavelength passing from target point to the camera
 */
class CAtmosphere {
protected:
	float m_f_planet_radius;
	float m_f_atmosphere_radius;
	float m_f_inv_radius_scale;
	Vector3f m_v_sun_dir;
	double m_f_k_raileigh;
	double m_f_k_mie;
	double m_f_4pi_k_raileigh;
	double m_f_4pi_k_mie;
	float m_f_g, m_f_g2;

	class COpticalDepthFunction;

public:
	/*
	 *	CAtmosphere::CAtmosphere(float f_planet_radius, float f_atmosphere_radius,
	 *		Vector3f v_sun_dir, float f_u = .7f, float f_k_mie = .25e-9f,
	 *		float f_k_raileigh = .15e-31f)
	 *		- default constructor
	 *		- initializes atmosphere with given values
	 */
	CAtmosphere(float f_planet_radius, float f_atmosphere_radius, Vector3f v_sun_dir,
		float f_u = .7f, float f_k_mie = .25e-9f, float f_k_raileigh = .15e-31f);

	/*
	 *	float CAtmosphere::f_Intensity(Vector3f v_camera, Vector3f v_target, float f_lambda) const
	 *		- returns relative intensity for light of wavelength f_lambda meters
	 *		  observed at v_camera and coming from v_target
	 */
	float f_Intensity(Vector3f v_camera, Vector3f v_target, float f_lambda) const;

	/*
	 *	class CAtmosphere::CIntensity
	 *		- intensity as a function of wavelenght
	 */
	class CIntensity {
	protected:
		double m_f_4pi_k_raileigh;
		double m_f_raileigh_integral_multiplier;
		double m_f_mie_integral;
		double *m_p_raileigh_depth_density_ratio;
		int m_n_point_num;

	public:
		/*
		 *	CIntensity(int n_point_num, double *p_raileigh_depth_density_ratio,
		 *		double f_4pi_k_raileigh, double f_raileigh_integral_multiplier,
		 *		float f_mie_integral)
		 *		- default constructor, initializes intensity function with given values
		 *		  (names should be self-explanatory)
		 *		- note p_raileigh_depth_density_ratio array is not copied, but
		 *		  directly referenced instead and will be deleted by destructor
		 */
		CIntensity(int n_point_num, double *p_raileigh_depth_density_ratio,
			double f_4pi_k_raileigh, double f_raileigh_integral_multiplier, double f_mie_integral)
			:m_f_4pi_k_raileigh(f_4pi_k_raileigh),
			m_f_raileigh_integral_multiplier(f_raileigh_integral_multiplier),
			m_f_mie_integral(f_mie_integral),
			m_p_raileigh_depth_density_ratio(p_raileigh_depth_density_ratio),
			m_n_point_num(n_point_num)
		{}

		/*
		 *	CIntensity(const CIntensity &r_intensity)
		 *		- copy-constructor
		 *		- call b_Status() to see wheter operation succeeded
		 */
		CIntensity(const CIntensity &r_intensity)
			:m_f_4pi_k_raileigh(r_intensity.m_f_4pi_k_raileigh),
			m_f_raileigh_integral_multiplier(r_intensity.m_f_raileigh_integral_multiplier),
			m_f_mie_integral(r_intensity.m_f_mie_integral),
			m_p_raileigh_depth_density_ratio(0), m_n_point_num(r_intensity.m_n_point_num)
		{
			if((m_p_raileigh_depth_density_ratio = new double[2 * m_n_point_num])) {
				memcpy(m_p_raileigh_depth_density_ratio,
					r_intensity.m_p_raileigh_depth_density_ratio,
					2 * m_n_point_num * sizeof(double));
			}
			// copy array
		}

		/*
		 *	~CIntensity()
		 *		- destructor
		 */
		~CIntensity()
		{
			if(m_p_raileigh_depth_density_ratio)
				delete[] m_p_raileigh_depth_density_ratio;
		}

		/*
		 *	CIntensity &operator =(const CIntensity &r_intensity)
		 *		- copy operator
		 *		- call b_Status() to see wheter operation succeeded
		 */
		CIntensity &operator =(const CIntensity &r_intensity)
		{
			if(m_n_point_num != r_intensity.m_n_point_num && m_p_raileigh_depth_density_ratio) {
				delete[] m_p_raileigh_depth_density_ratio;
				m_p_raileigh_depth_density_ratio = 0;
			}
			if(!m_p_raileigh_depth_density_ratio)
				m_p_raileigh_depth_density_ratio = new double[2 * r_intensity.m_n_point_num];
			// realloc density ratio if neccesary

			m_f_4pi_k_raileigh = r_intensity.m_f_4pi_k_raileigh;
			m_f_raileigh_integral_multiplier = r_intensity.m_f_raileigh_integral_multiplier;
			m_f_mie_integral = r_intensity.m_f_mie_integral;
			m_n_point_num = r_intensity.m_n_point_num;
			memcpy(m_p_raileigh_depth_density_ratio, r_intensity.m_p_raileigh_depth_density_ratio,
				2 * m_n_point_num * sizeof(double));
			// copy data

			return *this;
		}

		/*
		 *	bool b_Status() const
		 *		- returns true if constructor succeeded and the structure is ready to be used,
		 *		  otherwise false
		 *		- calling operator () while this returns false causes access violation
		 */
		bool b_Status() const
		{
			return m_p_raileigh_depth_density_ratio != 0;
		}

		/*
		 *	float CIntensity::operator ()(float f_lambda) const
		 *		- returns relative intensity for light of wavelength f_lambda meters
		 */
		float operator ()(float f_lambda) const
		{
			double f_lambda4 = f_lambda * f_lambda;
			f_lambda4 *= f_lambda4;
			// lambda^4

			double f_rl_depth_dcale = m_f_4pi_k_raileigh / f_lambda4;
			// precalc

			double f_integral_raileigh = 0;
			const double *p_raileigh_depth_density_ratio = m_p_raileigh_depth_density_ratio;
			for(int i = 0, n = m_n_point_num; i < n; ++ i) {
				double f_density_ratio = *p_raileigh_depth_density_ratio;
				++ p_raileigh_depth_density_ratio;
				// density ratio (ro at the start of the equation)

				double f_rl_depth = *p_raileigh_depth_density_ratio * f_rl_depth_dcale;
				++ p_raileigh_depth_density_ratio;
				// calculate optical density for railieigh

				f_integral_raileigh += f_density_ratio * exp(-f_rl_depth);
				// sum it up
			}
			// integral loop

			return (float)(f_integral_raileigh * m_f_raileigh_integral_multiplier /
				f_lambda4 + m_f_mie_integral);
			// return intensity
		}
	};

	/*
	 *	CIntensity CAtmosphere::Intensity(Vector3f v_camera, Vector3f v_target) const
	 *		- returns relative intensity function for light observed at v_camera
	 *		  and coming from v_target
	 */
	CIntensity Intensity(Vector3f v_camera, Vector3f v_target) const;

	/*
	 *	float CAtmosphere::f_Depth(Vector3f v_a, Vector3f v_b) const
	 *		- returns optical depth for atmosphere between points v_a and v_b,
	 *		  given they both lie inside atmosphere
	 */
	float f_Depth(Vector3f v_a, Vector3f v_b) const;

	/*
	 *	bool CAtmosphere::b_Shadow(Vector3f v_pos) const
	 *		- returns true in case v_pos is in shadow of lower sphere, otherwise false
	 */
	bool b_Shadow(Vector3f v_pos) const;

	/*
	 *	static bool CAtmosphere::IntersectSphere(const Vector3f &r_v_pos,
	 *		const Vector3f &r_v_dir, float f_radius, float &r_f_intersection)
	 *		- calculates intersection of ray starting at r_v_pos with r_v_dir direction
	 *		  with sphere having center at O and radius f_radius
	 *		- returns true in case intersection occurs, otherwise false
	 *		- fills r_f_intersection with distance to the first intersection
	 *		  (even if it's behind ray origin)
	 */
	static bool IntersectSphere(const Vector3f &r_v_pos,
		const Vector3f &r_v_dir, float f_radius, float &r_f_intersection);
};

#endif //__ATMOSPHERE_PHYSICS_INCLUDED
