/*
								+----------------------------------+
								|                                  |
								| *** Color system conversions *** |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|            Color.cpp             |
								|                                  |
								+----------------------------------+
*/

#include <assert.h>
#define _ASSERTE(x) assert(x)
#include <math.h>
#include "Color.h"

/*
 *								=== CColorSystem ===
 */

const float CColorSystem::f_gamma_REC709 = 0;

/*
 *	CColorSystem::CColorSystem(Vector2f v_red, Vector2f v_green,
 *		Vector2f v_blue, Vector2f v_white, float f_gamma = 1)
 *		- default constructor
 *		- constructs a new color system, given by CIE gamut coordinates
 *		  of red, green, blue and whitepoint
 *		- f_gamma is gamma correction
 */
CColorSystem::CColorSystem(Vector2f v_red, Vector2f v_green,
	Vector2f v_blue, Vector2f v_white, float f_gamma)
	:m_v_red(v_red), m_v_green(v_green), m_v_blue(v_blue),
	m_v_white(v_white), m_f_gamma(f_gamma)
{}

/*
 *	CColorSystem::CColorSystem(const CColorSystem &r_color_sys)
 *		- copy-constructor
 */
CColorSystem::CColorSystem(const CColorSystem &r_color_sys)
	:m_v_red(r_color_sys.m_v_red), m_v_green(r_color_sys.m_v_green),
	m_v_blue(r_color_sys.m_v_blue), m_v_white(r_color_sys.m_v_white),
	m_f_gamma(r_color_sys.m_f_gamma)
{}

/*
 *	CColorSystem &CColorSystem::operator =(const CColorSystem &r_color_sys)
 *		- copy-operator, returns reference to this
 */
CColorSystem &CColorSystem::operator =(const CColorSystem &r_color_sys)
{
	m_v_red = r_color_sys.m_v_red;
	m_v_green = r_color_sys.m_v_green;
	m_v_blue = r_color_sys.m_v_blue;
	m_v_white = r_color_sys.m_v_white;
	m_f_gamma = r_color_sys.m_f_gamma;
	return *this;
}

/*
 *	static CColorSystem::CColorSystem Make_System(int n_type)
 *		- creates a new color system of type n_type, which is one of
 *		  system_NTSC, system_EBU, system_SMPTE, system_HDTV,
 *		  system_CIE or system_Rec709
 */
CColorSystem CColorSystem::Make_System(int n_type)
{
	const struct {
		float f_xr, f_yr, f_xg, f_yg, f_xb, f_yb, f_xw, f_yw, f_gamma;
	} p_color_sys[] = {
		{0.67f, 0.33f, 0.21f, 0.71f, 0.14f, 0.08f, 0.3101f, 0.3162f, f_gamma_REC709},
		{0.64f, 0.33f, 0.29f, 0.6f, 0.15f, 0.06f, 0.3127f, 0.3291f, f_gamma_REC709},
		{0.63f, 0.34f, 0.31f, 0.595f, 0.155f, 0.07f, 0.3127f, 0.3291f, f_gamma_REC709},
		{0.67f, 0.33f, 0.21f, 0.71f, 0.15f, 0.06f, 0.3127f, 0.3291f, f_gamma_REC709},
		{0.7355f, 0.2645f, 0.2658f, 0.7243f, 0.1669f, 0.0085f, 0.33333333f, 0.33333333f, f_gamma_REC709},
		{0.64f, 0.33f, 0.3f, 0.6f, 0.15f, 0.06f, 0.3127f, 0.3291f, f_gamma_REC709}
	}, *p_sys = p_color_sys + n_type;
	// preset table

	_ASSERTE(n_type >= 0 && (unsigned)n_type < sizeof(p_color_sys) / sizeof(p_color_sys[0]));

	return CColorSystem(Vector2f(p_sys->f_xr, p_sys->f_yr),
		Vector2f(p_sys->f_xg, p_sys->f_yg), Vector2f(p_sys->f_xb, p_sys->f_yb),
		Vector2f(p_sys->f_xw, p_sys->f_yw), p_sys->f_gamma);
}

/*
 *	Vector3f CColorSystem::v_MapColor(Vector3f v_xyz_color) const
 *		- maps v_xyz_color into color system gamut and returns corresponding RGB color
 *		- note in case color is undisplayable, some of it's coordinates can be negative
 */
Vector3f CColorSystem::v_MapColor(Vector3f v_xyz_color) const
{
	Vector3f v_r(m_v_red.f_X(), m_v_red.f_Y(), 1 - m_v_red.f_X() - m_v_red.f_Y());
	Vector3f v_g(m_v_green.f_X(), m_v_green.f_Y(), 1 - m_v_green.f_X() - m_v_green.f_Y());
	Vector3f v_b(m_v_blue.f_X(), m_v_blue.f_Y(), 1 - m_v_blue.f_X() - m_v_blue.f_Y());
	Vector3f v_w(m_v_white.f_X(), m_v_white.f_Y(), 1 - m_v_white.f_X() - m_v_white.f_Y());
	// expand coordinates to 3-space (maybe do that in constructor?)

	Vector3f v_mat0(v_g.v_Cross(v_b));
	Vector3f v_mat1(v_b.v_Cross(v_r));
	Vector3f v_mat2(v_r.v_Cross(v_g));
	// three-column conversion matrix

	v_mat0 *= v_w.f_Y() / v_mat0.f_Dot(v_w);
	v_mat1 *= v_w.f_Y() / v_mat1.f_Dot(v_w);
	v_mat2 *= v_w.f_Y() / v_mat2.f_Dot(v_w);
	// rescale matrix to correctly display white

	return Vector3f(v_xyz_color.f_Dot(v_mat0),
		v_xyz_color.f_Dot(v_mat1), v_xyz_color.f_Dot(v_mat2));
	// transform color by the matrix
}

/*
 *	bool CColorSystem::b_InsideGamut_XYZ(Vector3f v_xyz_color) const
 *		- returns true in case v_xyz_color CIE color is inside
 *		  displayable gamut, otherwise false
 *		- note this function performs color conversion to RGB,
 *		  if RGB equivalent of v_xyz_color is present, it is
 *		  more suitable to use b_InsideGamut_RGB()
 */
bool CColorSystem::b_InsideGamut_XYZ(Vector3f v_xyz_color) const
{
	Vector3f v_rgb = v_MapColor(v_xyz_color);
	return v_rgb.f_X() >= 0 && v_rgb.f_Y() >= 0 && v_rgb.f_Z() >= 0;
}

/*
 *	bool CColorSystem::b_InsideGamut_RGB(Vector3f v_rgb_color) const
 *		- returns true in case v_rgb_color RGB color is inside
 *		  displayable gamut, otherwise false
 */
bool CColorSystem::b_InsideGamut_RGB(Vector3f v_rgb_color) const
{
	return v_rgb_color.f_X() >= 0 && v_rgb_color.f_Y() >= 0 && v_rgb_color.f_Z() >= 0;
}

/*
 *	Vector3f CColorSystem::v_Constrain(Vector3f v_rgb_color) const
 *		- constrains RGB color so it's inside gamut
 *		- note this doesn't affect colors inside gamut
 */
Vector3f CColorSystem::v_Constrain(Vector3f v_rgb_color) const
{
	float w = 0;
	for(int n = 0; n < 3; ++ n)
		w = (w > v_rgb_color[n])? v_rgb_color[n] : w;
	// find minimum

	if(w < 0)
		return v_rgb_color - Vector3f(w, w, w);
	else
		return v_rgb_color;
	// desaturate to fit inside gamut
}

/*
 *	float CColorSystem::f_GammaCorrect(float f_intensity) const
 *		- returns gamma-corrected intensity f_intensity
 */
float CColorSystem::f_GammaCorrect(float f_intensity) const
{
	if(m_f_gamma == f_gamma_REC709) {
		const float f_cc = .018f;
		if(f_intensity < f_cc)
			return f_intensity * ((1.099f * (float)pow(f_cc, .45f)) - .099f) / f_cc;
		else
			return (1.099f * (float)pow(f_intensity, .45f)) - .099f;
		// rec. 709 gamma correction
	} else if(m_f_gamma != 1.0f)
		return (float)pow(f_intensity, 1.0f / m_f_gamma); // std gamma correction
	else
		return f_intensity; // no gamma correction
}

/*
 *	Vector3f CColorSystem::v_GammaCorrect(Vector3f v_rgb_color) const
 *		- returns gamma-corrected color v_rgb_color
 */
Vector3f CColorSystem::v_GammaCorrect(Vector3f v_rgb_color) const
{
	return Vector3f(f_GammaCorrect(v_rgb_color.f_X()),
		f_GammaCorrect(v_rgb_color.f_Y()), f_GammaCorrect(v_rgb_color.f_Z()));
}

/*
 *								=== ~CColorSystem ===
 */

/*
 *								=== CSpectrumToCIE ===
 */

const float CSpectrumToCIE::m_p_table[3 * 81] = {
    /*{*/0.0014f, 0.0000f, 0.0065f/*}*/, /*{*/0.0022f, 0.0001f, 0.0105f/*}*/,
	/*{*/0.0042f, 0.0001f, 0.0201f/*}*/, /*{*/0.0076f, 0.0002f, 0.0362f/*}*/,
	/*{*/0.0143f, 0.0004f, 0.0679f/*}*/, /*{*/0.0232f, 0.0006f, 0.1102f/*}*/, 
    /*{*/0.0435f, 0.0012f, 0.2074f/*}*/, /*{*/0.0776f, 0.0022f, 0.3713f/*}*/,
	/*{*/0.1344f, 0.0040f, 0.6456f/*}*/, /*{*/0.2148f, 0.0073f, 1.0391f/*}*/,
	/*{*/0.2839f, 0.0116f, 1.3856f/*}*/, /*{*/0.3285f, 0.0168f, 1.6230f/*}*/, 
    /*{*/0.3483f, 0.0230f, 1.7471f/*}*/, /*{*/0.3481f, 0.0298f, 1.7826f/*}*/,
	/*{*/0.3362f, 0.0380f, 1.7721f/*}*/, /*{*/0.3187f, 0.0480f, 1.7441f/*}*/,
	/*{*/0.2908f, 0.0600f, 1.6692f/*}*/, /*{*/0.2511f, 0.0739f, 1.5281f/*}*/, 
    /*{*/0.1954f, 0.0910f, 1.2876f/*}*/, /*{*/0.1421f, 0.1126f, 1.0419f/*}*/,
	/*{*/0.0956f, 0.1390f, 0.8130f/*}*/, /*{*/0.0580f, 0.1693f, 0.6162f/*}*/,
	/*{*/0.0320f, 0.2080f, 0.4652f/*}*/, /*{*/0.0147f, 0.2586f, 0.3533f/*}*/, 
    /*{*/0.0049f, 0.3230f, 0.2720f/*}*/, /*{*/0.0024f, 0.4073f, 0.2123f/*}*/,
	/*{*/0.0093f, 0.5030f, 0.1582f/*}*/, /*{*/0.0291f, 0.6082f, 0.1117f/*}*/,
	/*{*/0.0633f, 0.7100f, 0.0782f/*}*/, /*{*/0.1096f, 0.7932f, 0.0573f/*}*/, 
    /*{*/0.1655f, 0.8620f, 0.0422f/*}*/, /*{*/0.2257f, 0.9149f, 0.0298f/*}*/,
	/*{*/0.2904f, 0.9540f, 0.0203f/*}*/, /*{*/0.3597f, 0.9803f, 0.0134f/*}*/,
	/*{*/0.4334f, 0.9950f, 0.0087f/*}*/, /*{*/0.5121f, 1.0000f, 0.0057f/*}*/, 
    /*{*/0.5945f, 0.9950f, 0.0039f/*}*/, /*{*/0.6784f, 0.9786f, 0.0027f/*}*/,
	/*{*/0.7621f, 0.9520f, 0.0021f/*}*/, /*{*/0.8425f, 0.9154f, 0.0018f/*}*/,
	/*{*/0.9163f, 0.8700f, 0.0017f/*}*/, /*{*/0.9786f, 0.8163f, 0.0014f/*}*/, 
    /*{*/1.0263f, 0.7570f, 0.0011f/*}*/, /*{*/1.0567f, 0.6949f, 0.0010f/*}*/,
	/*{*/1.0622f, 0.6310f, 0.0008f/*}*/, /*{*/1.0456f, 0.5668f, 0.0006f/*}*/,
	/*{*/1.0026f, 0.5030f, 0.0003f/*}*/, /*{*/0.9384f, 0.4412f, 0.0002f/*}*/, 
    /*{*/0.8544f, 0.3810f, 0.0002f/*}*/, /*{*/0.7514f, 0.3210f, 0.0001f/*}*/,
	/*{*/0.6424f, 0.2650f, 0.0000f/*}*/, /*{*/0.5419f, 0.2170f, 0.0000f/*}*/,
	/*{*/0.4479f, 0.1750f, 0.0000f/*}*/, /*{*/0.3608f, 0.1382f, 0.0000f/*}*/, 
    /*{*/0.2835f, 0.1070f, 0.0000f/*}*/, /*{*/0.2187f, 0.0816f, 0.0000f/*}*/,
	/*{*/0.1649f, 0.0610f, 0.0000f/*}*/, /*{*/0.1212f, 0.0446f, 0.0000f/*}*/,
	/*{*/0.0874f, 0.0320f, 0.0000f/*}*/, /*{*/0.0636f, 0.0232f, 0.0000f/*}*/, 
    /*{*/0.0468f, 0.0170f, 0.0000f/*}*/, /*{*/0.0329f, 0.0119f, 0.0000f/*}*/,
	/*{*/0.0227f, 0.0082f, 0.0000f/*}*/, /*{*/0.0158f, 0.0057f, 0.0000f/*}*/,
	/*{*/0.0114f, 0.0041f, 0.0000f/*}*/, /*{*/0.0081f, 0.0029f, 0.0000f/*}*/, 
    /*{*/0.0058f, 0.0021f, 0.0000f/*}*/, /*{*/0.0041f, 0.0015f, 0.0000f/*}*/,
	/*{*/0.0029f, 0.0010f, 0.0000f/*}*/, /*{*/0.0020f, 0.0007f, 0.0000f/*}*/,
	/*{*/0.0014f, 0.0005f, 0.0000f/*}*/, /*{*/0.0010f, 0.0004f, 0.0000f/*}*/, 
    /*{*/0.0007f, 0.0002f, 0.0000f/*}*/, /*{*/0.0005f, 0.0002f, 0.0000f/*}*/,
	/*{*/0.0003f, 0.0001f, 0.0000f/*}*/, /*{*/0.0002f, 0.0001f, 0.0000f/*}*/,
	/*{*/0.0002f, 0.0001f, 0.0000f/*}*/, /*{*/0.0001f, 0.0000f, 0.0000f/*}*/, 
    /*{*/0.0001f, 0.0000f, 0.0000f/*}*/, /*{*/0.0001f, 0.0000f, 0.0000f/*}*/,
	/*{*/0.0000f, 0.0000f, 0.0000f/*}*/
};

const float CSpectrumToCIE::m_f_min_wavelength = 380e-9f;
const float CSpectrumToCIE::m_f_wavelength_step = 5e-9f;

/*
 *	static Vector3f CSpectrumToCIE::v_CIE(float f_wavelength)
 *		- returns CIE xyz color for a single given wavelength f_wavelength (in meters)
 *		- returns zero vectors for out-of-range wavelengths
 */
Vector3f CSpectrumToCIE::v_CIE(float f_wavelength)
{
	const int n_table_size = 81;

	f_wavelength -= m_f_min_wavelength;
	f_wavelength /= m_f_wavelength_step;
	int n_index = (int)f_wavelength;
	float f_lerp = f_wavelength - n_index;
	// determine piecewise curve segment

	Vector3f v_cie(0);
	n_index *= 3;
	if(n_index >= 0 && n_index < n_table_size * 3) {
		v_cie = Vector3f(m_p_table[n_index],
			m_p_table[n_index + 1], m_p_table[n_index + 2]) * (1 - f_lerp);
	}
	n_index += 3;
	if(n_index >= 0 && n_index < n_table_size * 3) {
		v_cie += Vector3f(m_p_table[n_index],
			m_p_table[n_index + 1], m_p_table[n_index + 2]) * f_lerp;
	}
	// piecewise linear interpolation

	float f_size = v_cie.f_Sum();
	if(f_size != 0) // negative spectrum values?
		v_cie /= f_size;
	// normalize if nonzero

	return v_cie;
}

/*
 *								=== ~CSpectrumToCIE ===
 */

/*
 *								=== CSunSpectrum ===
 */

const float CSunSpectrum::m_p_sun_radiation[401] = {
	1.1520E+00f, 1.2490E+00f, 9.5426E-01f, 7.3603E-01f, 8.2100E-01f, 1.0802E+00f, 9.9113E-01f,
	1.0354E+00f, 1.0067E+00f, 1.0810E+00f, 1.2519E+00f, 1.3322E+00f, 1.2400E+00f, 7.4500E-01f,
	7.6675E-01f, 1.2450E+00f, 1.1626E+00f, 6.5200E-01f, 1.3010E+00f, 1.6245E+00f, 1.6885E+00f,
	1.7520E+00f, 1.8140E+00f, 1.7400E+00f, 1.7630E+00f, 1.7150E+00f, 1.6660E+00f, 1.6300E+00f,
	1.6990E+00f, 1.8090E+00f, 1.5370E+00f, 1.7150E+00f, 1.8160E+00f, 1.7392E+00f, 1.7144E+00f,
	1.7688E+00f, 1.8150E+00f, 1.7660E+00f, 1.6850E+00f, 1.7490E+00f, 1.5990E+00f, 1.8110E+00f,
	1.7820E+00f, 1.7210E+00f, 1.7080E+00f, 1.7550E+00f, 1.6990E+00f, 1.6380E+00f, 1.6510E+00f,
	1.5230E+00f, 1.2120E+00f, 1.0990E+00f, 1.8220E+00f, 1.6913E+00f, 1.5600E+00f, 1.7090E+00f,
	1.8680E+00f, 1.9000E+00f, 1.6630E+00f, 1.6010E+00f, 1.8300E+00f, 1.7990E+00f, 1.9220E+00f,
	1.9490E+00f, 1.8941E+00f, 1.9650E+00f, 1.7557E+00f, 1.9900E+00f, 2.0140E+00f, 2.0010E+00f,
	2.0690E+00f, 2.1420E+00f, 2.0470E+00f, 1.8864E+00f, 2.0180E+00f, 2.0010E+00f, 2.0630E+00f,
	2.0770E+00f, 2.0320E+00f, 2.0120E+00f, 1.9973E+00f, 2.0639E+00f, 2.0780E+00f, 2.0840E+00f,
	2.0150E+00f, 1.9840E+00f, 2.0210E+00f, 1.9310E+00f, 2.0120E+00f, 2.0180E+00f, 1.9390E+00f,
	1.9690E+00f, 2.0700E+00f, 1.9882E+00f, 2.0120E+00f, 2.0800E+00f, 2.0120E+00f, 2.0250E+00f,
	2.0860E+00f, 2.0400E+00f, 2.0680E+00f, 2.0610E+00f, 2.0623E+00f, 2.0310E+00f, 1.9890E+00f,
	1.9790E+00f, 1.6010E+00f, 1.7890E+00f, 1.9350E+00f, 1.8224E+00f, 2.0320E+00f, 1.9490E+00f,
	1.8560E+00f, 1.9830E+00f, 1.9339E+00f, 2.0510E+00f, 1.9490E+00f, 1.9800E+00f, 1.9240E+00f,
	1.9230E+00f, 1.9160E+00f, 1.8580E+00f, 1.8600E+00f, 1.9490E+00f, 1.8330E+00f, 1.9472E+00f,
	2.0250E+00f, 1.9354E+00f, 1.8800E+00f, 1.9650E+00f, 1.9100E+00f, 1.9410E+00f, 1.9890E+00f,
	1.8660E+00f, 1.8240E+00f, 1.8750E+00f, 1.8910E+00f, 1.5390E+00f, 1.7590E+00f, 1.7040E+00f,
	1.8600E+00f, 1.8730E+00f, 1.9150E+00f, 1.8040E+00f, 1.9410E+00f, 1.9280E+00f, 1.8740E+00f,
	1.6410E+00f, 1.8800E+00f, 1.9690E+00f, 1.8920E+00f, 1.9950E+00f, 1.9580E+00f, 1.7470E+00f,
	1.8690E+00f, 1.8950E+00f, 1.9740E+00f, 1.8240E+00f, 1.9130E+00f, 1.8640E+00f, 1.8000E+00f,
	1.7340E+00f, 1.8880E+00f, 1.8510E+00f, 1.9190E+00f, 1.8740E+00f, 1.8609E+00f, 1.8820E+00f,
	1.8260E+00f, 1.8800E+00f, 1.8630E+00f, 1.8590E+00f, 1.8960E+00f, 1.8420E+00f, 1.8780E+00f,
	1.8890E+00f, 1.8570E+00f, 1.8120E+00f, 1.8530E+00f, 1.7550E+00f, 1.7860E+00f, 1.8900E+00f,
	1.8010E+00f, 1.8710E+00f, 1.8360E+00f, 1.8490E+00f, 1.7500E+00f, 1.8680E+00f, 1.8590E+00f,
	1.8310E+00f, 1.8280E+00f, 1.7620E+00f, 1.8720E+00f, 1.8810E+00f, 1.8730E+00f, 1.8340E+00f,
	1.8180E+00f, 1.8620E+00f, 1.7990E+00f, 1.8160E+00f, 1.8340E+00f, 1.8330E+00f, 1.8520E+00f,
	1.8630E+00f, 1.8540E+00f, 1.8360E+00f, 1.7920E+00f, 1.8380E+00f, 1.8210E+00f, 1.6240E+00f,
	1.7218E+00f, 1.8090E+00f, 1.7880E+00f, 1.7920E+00f, 1.7890E+00f, 1.7780E+00f, 1.7960E+00f,
	1.8060E+00f, 1.7720E+00f, 1.7640E+00f, 1.7700E+00f, 1.7420E+00f, 1.7150E+00f, 1.7490E+00f,
	1.7790E+00f, 1.7730E+00f, 1.7580E+00f, 1.7620E+00f, 1.7510E+00f, 1.7340E+00f, 1.7240E+00f,
	1.7120E+00f, 1.7360E+00f, 1.7100E+00f, 1.6550E+00f, 1.7120E+00f, 1.6640E+00f, 1.6410E+00f,
	1.7020E+00f, 1.7090E+00f, 1.7110E+00f, 1.7240E+00f, 1.6784E+00f, 1.6820E+00f, 1.6670E+00f,
	1.6440E+00f, 1.6400E+00f, 1.6930E+00f, 1.6930E+00f, 1.6870E+00f, 1.6650E+00f, 1.6590E+00f,
	1.5901E+00f, 1.6740E+00f, 1.6370E+00f, 1.6520E+00f, 1.6093E+00f, 1.6610E+00f, 1.6650E+00f,
	1.6530E+00f, 1.6130E+00f, 1.6100E+00f, 1.6090E+00f, 1.6250E+00f, 1.6140E+00f, 1.6270E+00f,
	1.5910E+00f, 1.6060E+00f, 1.6020E+00f, 1.5510E+00f, 1.5260E+00f, 1.6130E+00f, 1.5910E+00f,
	1.5980E+00f, 1.5750E+00f, 1.5230E+00f, 1.3233E+00f, 1.3840E+00f, 1.5390E+00f, 1.5420E+00f,
	1.5580E+00f, 1.5660E+00f, 1.5710E+00f, 1.5630E+00f, 1.5540E+00f, 1.5670E+00f, 1.5550E+00f,
	1.5354E+00f, 1.5348E+00f, 1.5580E+00f, 1.5340E+00f, 1.5290E+00f, 1.5060E+00f, 1.5170E+00f,
	1.5130E+00f, 1.4990E+00f, 1.5150E+00f, 1.5000E+00f, 1.5070E+00f, 1.4930E+00f, 1.4940E+00f,
	1.4870E+00f, 1.4930E+00f, 1.4760E+00f, 1.4660E+00f, 1.4650E+00f, 1.4330E+00f, 1.4720E+00f,
	1.4760E+00f, 1.4780E+00f, 1.4790E+00f, 1.4680E+00f, 1.4540E+00f, 1.4580E+00f, 1.4570E+00f,
	1.4350E+00f, 1.4420E+00f, 1.4380E+00f, 1.4170E+00f, 1.4340E+00f, 1.4220E+00f, 1.4131E+00f,
	1.3987E+00f, 1.4095E+00f, 1.4187E+00f, 1.4330E+00f, 1.4138E+00f, 1.4040E+00f, 1.3990E+00f,
	1.3900E+00f, 1.4040E+00f, 1.3970E+00f, 1.3818E+00f, 1.3702E+00f, 1.3819E+00f, 1.3502E+00f,
	1.3694E+00f, 1.3650E+00f, 1.3570E+00f, 1.3010E+00f, 1.3487E+00f, 1.3480E+00f, 1.3600E+00f,
	1.3510E+00f, 1.3607E+00f, 1.3465E+00f, 1.3429E+00f, 1.3444E+00f, 1.3370E+00f, 1.2796E+00f,
	1.3357E+00f, 1.3104E+00f, 1.3333E+00f, 1.3108E+00f, 1.3390E+00f, 1.3271E+00f, 1.3100E+00f,
	1.3120E+00f, 1.3000E+00f, 1.2646E+00f, 1.2830E+00f, 1.2707E+00f, 1.2649E+00f, 1.2892E+00f,
	1.2955E+00f, 1.2920E+00f, 1.2892E+00f, 1.2890E+00f, 1.2808E+00f, 1.2760E+00f, 1.2740E+00f,
	1.2680E+00f, 1.2720E+00f, 1.2650E+00f, 1.2809E+00f, 1.2771E+00f, 1.2610E+00f, 1.2598E+00f,
	1.2680E+00f, 1.2500E+00f, 1.2590E+00f, 1.2487E+00f, 1.2460E+00f, 1.2543E+00f, 1.2566E+00f,
	1.2452E+00f, 1.2000E+00f, 1.2120E+00f, 1.2310E+00f, 1.2142E+00f, 1.2146E+00f, 1.2073E+00f,
	1.2120E+00f, 1.2100E+00f, 1.2090E+00f, 1.2080E+00f, 1.2102E+00f, 1.2028E+00f, 1.2014E+00f,
	1.2067E+00f, 1.1930E+00f
};

/*
 *								=== ~CSunSpectrum ===
 */
