/*
								+----------------------------------+
								|                                  |
								| *** Color system conversions *** |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|             Color.h              |
								|                                  |
								+----------------------------------+
*/

#ifndef __COLOR_SYSTEMS_INCLUDED
#define __COLOR_SYSTEMS_INCLUDED

#include <math.h>
#include "MiniVector.h"

/*
 *	class CColorSystem
 *		- describes color system CIE gamut
 */
class CColorSystem {
protected:
	Vector2f m_v_red, m_v_green, m_v_blue, m_v_white;
	float m_f_gamma;

public:
	/*
	 *	static const float f_gamma_REC709
	 *		- name for rec. 709 gamma correction, pass to constructor as f_gamma
	 */
	static const float f_gamma_REC709;

	/*
	 *	CColorSystem::CColorSystem(Vector2f v_red, Vector2f v_green,
	 *		Vector2f v_blue, Vector2f v_white, float f_gamma = 1)
	 *		- default constructor
	 *		- constructs a new color system, given by CIE gamut coordinates
	 *		  of red, green, blue and whitepoint
	 *		- f_gamma is gamma correction
	 */
	CColorSystem(Vector2f v_red, Vector2f v_green,
		Vector2f v_blue, Vector2f v_white, float f_gamma = 1);

	/*
	 *	CColorSystem::CColorSystem(const CColorSystem &r_color_sys)
	 *		- copy-constructor
	 */
	CColorSystem(const CColorSystem &r_color_sys);

	/*
	 *	CColorSystem &CColorSystem::operator =(const CColorSystem &r_color_sys)
	 *		- copy-operator, returns reference to this
	 */
	CColorSystem &operator =(const CColorSystem &r_color_sys);

	/*
	 *	inline Vector2f CColorSystem::v_Red() const
	 *		- returns position of red color in CIE gamut
	 */
	inline Vector2f v_Red() const
	{
		return m_v_red;
	}

	/*
	 *	inline Vector2f CColorSystem::v_Green() const
	 *		- returns position of green color in CIE gamut
	 */
	inline Vector2f v_Green() const
	{
		return m_v_green;
	}

	/*
	 *	inline Vector2f CColorSystem::v_Blue() const
	 *		- returns position of blue color in CIE gamut
	 */
	inline Vector2f v_Blue() const
	{
		return m_v_blue;
	}

	/*
	 *	inline Vector2f CColorSystem::v_White() const
	 *		- returns position of white-point in CIE gamut
	 */
	inline Vector2f v_White() const
	{
		return m_v_white;
	}

	/*
	 *	inline float CColorSystem::f_Gamma() const
	 *		- returns color system gamma
	 */
	inline float f_Gamma() const
	{
		return m_f_gamma;
	}

	/*
	 *	CColorSystem::<unnamed>
	 *		- names of some standard color systems
	 */
	enum {
		system_NTSC = 0,
		system_EBU, // EBU (PAL/SECAM)
		system_SMPTE,
		system_HDTV,
		system_CIE,
		system_Rec709 // CIE REC 709
	};

	/*
	 *	static CColorSystem::CColorSystem Make_System(int n_type)
	 *		- creates a new color system of type n_type, which is one of
	 *		  system_NTSC, system_EBU, system_SMPTE, system_HDTV,
	 *		  system_CIE or system_Rec709
	 */
	static CColorSystem Make_System(int n_type);

	/*
	 *	Vector3f CColorSystem::v_MapColor(Vector3f v_xyz_color) const
	 *		- maps v_xyz_color into color system gamut and returns corresponding RGB color
	 *		- note in case color is undisplayable, some of it's coordinates can be negative
	 */
	Vector3f v_MapColor(Vector3f v_xyz_color) const;

	/*
	 *	bool CColorSystem::b_InsideGamut_XYZ(Vector3f v_xyz_color) const
	 *		- returns true in case v_xyz_color CIE color is inside
	 *		  displayable gamut, otherwise false
	 *		- note this function performs color conversion to RGB,
	 *		  if RGB equivalent of v_xyz_color is present, it is
	 *		  more suitable to use b_InsideGamut_RGB()
	 */
	bool b_InsideGamut_XYZ(Vector3f v_xyz_color) const;

	/*
	 *	bool CColorSystem::b_InsideGamut_RGB(Vector3f v_rgb_color) const
	 *		- returns true in case v_rgb_color RGB color is inside
	 *		  displayable gamut, otherwise false
	 */
	bool b_InsideGamut_RGB(Vector3f v_rgb_color) const;

	/*
	 *	Vector3f CColorSystem::v_Constrain(Vector3f v_rgb_color) const
	 *		- constrains RGB color so it's inside gamut
	 *		- note this doesn't affect colors inside gamut
	 */
	Vector3f v_Constrain(Vector3f v_rgb_color) const;

	/*
	 *	float CColorSystem::f_GammaCorrect(float f_intensity) const
	 *		- returns gamma-corrected intensity f_intensity
	 */
	float f_GammaCorrect(float f_intensity) const;

	/*
	 *	Vector3f CColorSystem::v_GammaCorrect(Vector3f v_rgb_color) const
	 *		- returns gamma-corrected color v_rgb_color
	 */
	Vector3f v_GammaCorrect(Vector3f v_rgb_color) const;
};

/*
 *	class CBlackBodySpectrum
 *		- simple spectrum class for black body
 */
class CBlackBodySpectrum {
protected:
	float m_f_temp;

public:
	/*
	 *	CBlackBodySpectrum(float f_temperature)
	 *		- default constructor; f_temperature is temperature in kelvins
	 */
	inline CBlackBodySpectrum(float f_temperature)
		:m_f_temp(f_temperature)
	{}

	/*
	 *	float operator ()(float f_wavelength) const
	 *		- returns light intensity at wavelength f_wavelength meters (not nanometers!)
	 *		- intensity units may be only relative for conversion to xyz color because
	 *		  it will be normalized in the end
	 */
	inline float operator ()(float f_wavelength) const
	{
		return (float)((3.74183e-16 * pow(f_wavelength, -5.0)) /
			(exp(1.4388e-2 / (f_wavelength * m_f_temp)) - 1.0)); // Planck's radiation law
	}
};

/*
 *	class CSunSpectrum
 *		- extraterrestial sun spectrum (the visible part 380nm to 780nm with 1nm stepping)
 *		- note the returned values are in W.m^-2.nm^-1,
 *		  obtained from the 'ASTM G173-03 Reference Spectra'
 */
class CSunSpectrum {
protected:
	static const float m_p_sun_radiation[401];

public:
	/*
	 *	float operator ()(float f_wavelength) const
	 *		- returns light intensity at wavelength f_wavelength meters (not nanometers!)
	 *		- intensity units are W.m^-2.nm^-1
	 */
	inline float operator ()(float f_wavelength) const
	{
		f_wavelength *= 1e9f;
		f_wavelength -= 380;
		// to table units

		int n_segment = (int)f_wavelength;
		f_wavelength -= n_segment;
		// calc table index and lerp weight

		if(n_segment >= 0 && n_segment < 400) {
			return m_p_sun_radiation[n_segment] * (1 - f_wavelength) +
				m_p_sun_radiation[n_segment + 1] * f_wavelength;
		} else if(n_segment == 400)
			return m_p_sun_radiation[n_segment];
		else
			return 0;
		// check table bounds, interpolate
	}
};

/*
 *	class CSpectrumToCIE
 *		- class with two static functions  for converting light spectra to CIE color
 */
class CSpectrumToCIE {
protected:
	static const float m_p_table[3 * 81];
	static const float m_f_min_wavelength /*= 380e-9f*/;
	static const float m_f_wavelength_step /*= 5e-9f*/;

public:
	/*
	 *	template <class CSpectrum>
	 *	static Vector3f CSpectrumToCIE::v_CIE(CSpectrum spectrum)
	 *		- converts light given by spectrum to CIE xyz color
	 *		- spectrum is function object that is supposed to have public member function
	 *		  operator () which inputs float wavelength in meters and outputs light intensity
	 *		  on that wavelength
	 */
	template <class CSpectrum>
	static Vector3f v_CIE(CSpectrum spectrum)
	{
		const int n_table_size = 81; // (780 - 380) / 5 + 1;

		Vector3<double> v_cie(0);
		// use double for summing up lots of small numbers, float not precise enough

		float f_wavelength = m_f_min_wavelength;
		const float f_step = m_f_wavelength_step; // fetch
		for(const float *p_color = m_p_table, *p_end = m_p_table + 3 * n_table_size;
		   p_color != p_end; p_color += 3, f_wavelength += f_step) {
			v_cie += Vector3<double>(*p_color, *(p_color + 1),
				*(p_color + 2)) * spectrum(f_wavelength);
		}
		// sum up CIE equivs for different wavelengths

		double f_size = v_cie.f_Sum();
		if(f_size != 0) // negative spectrum values?
			v_cie /= f_size;
		// normalize if nonzero

		return Vector3f(v_cie.f_X(), v_cie.f_Y(), v_cie.f_Z());
	}
	/*
	 *	template <class CSpectrum>
	 *	static Vector3f CSpectrumToCIE::v_CIE(CSpectrum spectrum, double &r_power_integral)
	 *		- converts light given by spectrum to CIE xyz color
	 *		- spectrum is function object that is supposed to have public member function
	 *		  operator () which inputs float wavelength in meters and outputs light intensity
	 *		  on that wavelength
	 */
	template <class CSpectrum>
	static Vector3f v_CIE(CSpectrum spectrum, double &r_power_integral)
	{
		const int n_table_size = 81; // (780 - 380) / 5 + 1;

		Vector3<double> v_cie(0);
		// use double for summing up lots of small numbers, float not precise enough

		float f_wavelength = m_f_min_wavelength;
		const float f_step = m_f_wavelength_step; // fetch
		double f_power_integral = 0;
		for(const float *p_color = m_p_table, *p_end = m_p_table + 3 * n_table_size;
		   p_color != p_end; p_color += 3, f_wavelength += f_step) {
			double f_power = spectrum(f_wavelength);
			f_power_integral += f_power;
			v_cie += Vector3<double>(*p_color, *(p_color + 1),
				*(p_color + 2)) * f_power;
		}
		r_power_integral = f_power_integral * f_step; // finish integration
		// sum up CIE equivs for different wavelengths

		double f_size = v_cie.f_Sum();
		if(f_size != 0) // negative spectrum values?
			v_cie /= f_size;
		// normalize if nonzero

		return Vector3f(v_cie.f_X(), v_cie.f_Y(), v_cie.f_Z());
	}

	/*
	 *	static Vector3f CSpectrumToCIE::v_CIE(float f_wavelength)
	 *		- returns CIE xyz color for a single given wavelength f_wavelength (in meters)
	 *		- returns zero vectors for out-of-range wavelengths
	 */
	static Vector3f v_CIE(float f_wavelength);
};

#endif //__COLOR_SYSTEMS_INCLUDED
