/*
								+---------------------------------+
								|                                 |
								|   ***  2D image template  ***   |
								|                                 |
								|   Written by Lukas Polok 2007   |
								|                                 |
								|             Image.h             |
								|                                 |
								+---------------------------------+
*/

#ifndef __IMAGE_2D_INCLUDED
#define __IMAGE_2D_INCLUDED

/*
 *	template <class CPixelType>
 *	class CImage2D
 *		- template for two-dimensional scalar/vector field
 */
template <class CPixelType>
class CImage2D {
protected:
	int m_n_width, m_n_height;
	int m_n_width_mask, m_n_height_mask, m_n_height_shift;
	CPixelType *m_p_data;

public:
	/*
	 *	CImage2D(int n_width, int n_height)
	 *		- default constructor, allocates field n_width per n_height elements
	 *		- note dimensions have to be power of two
	 *		- call b_Status() to see wheter allocation succeeded
	 */
	CImage2D(int n_width, int n_height)
		:m_n_width(n_width), m_n_height(n_height),
		m_p_data(new CPixelType[n_width * n_height])
	{
		_ASSERTE(n_width == (1 << n_Bit_Num(n_width)));
		_ASSERTE(n_height == (1 << n_Bit_Num(n_height)));
		// want powers of two only

		m_n_width_mask = n_width - 1;
		m_n_height_mask = n_height - 1;
		m_n_height_shift = n_Bit_Num(n_width);
		m_n_height_mask <<= m_n_height_shift;
		// calculate masks and shifts to accelerate sampling
	}

	/*
	 *	~CImage2D()
	 *		- default destructor
	 */
	~CImage2D()
	{
		if(m_p_data)
			delete[] m_p_data;
	}

	/*
	 *	bool b_Status() const
	 *		- if field was successfuly allocated and is ready
	 *		  to use, returns true, otherwise returns false
	 */
	bool b_Status() const
	{
		return m_p_data != 0;
	}

	/*
	 *	inline int n_Width() const
	 *		- returns field width
	 */
	inline int n_Width() const
	{
		return m_n_width;
	}

	/*
	 *	inline int n_Height() const
	 *		- returns field height
	 */
	inline int n_Height() const
	{
		return m_n_height;
	}

	/*
	 *	inline CPixelType &f_At(int x, int y)
	 *		- returns reference to pixel at location x, y
	 *		- note x and y are clamped so they always point inside the field (in repeat fashion)
	 */
	inline CPixelType &f_At(int x, int y)
	{
		return m_p_data[(x & m_n_width_mask) | ((y << m_n_height_shift) & m_n_height_mask)];
	}

	/*
	 *	inline CPixelType f_At(int x, int y) const
	 *		- returns value of pixel at location x, y
	 *		- note x and y are clamped so they always point inside the field (in repeat fashion)
	 */
	inline CPixelType f_At(int x, int y) const
	{
		return m_p_data[(x & m_n_width_mask) | ((y << m_n_height_shift) & m_n_height_mask)];
	}

	/*
	 *	inline CPixelType f_Get(int x, int y) const
	 *		- returns value of pixel at location x, y
	 *		- note x and y are not clamped (optimization)
	 */
	inline CPixelType f_Get(int x, int y) const
	{
		_ASSERTE(x >= 0 && x < m_n_width && y >= 0 && y < m_n_height);
		return m_p_data[x | (y << m_n_height_shift)];
	}

	/*
	 *	inline void Set(int x, int y, CPixelType f_value) const
	 *		- sets value of pixel at location x, y to f_value
	 *		- note x and y are not clamped (optimization)
	 */
	inline void Set(int x, int y, CPixelType f_value) const
	{
		_ASSERTE(x >= 0 && x < m_n_width && y >= 0 && y < m_n_height);
		m_p_data[x | (y << m_n_height_shift)] = f_value;
	}

	/*
	 *	CPixelType f_Sample(float f_x, float f_y) const
	 *		- returns bilinear sample value at coordinates f_x and f_y (in pixels)
	 *		- note f_x and f_y are clamped so they always
	 *		  point inside the field (in repeat fashion)
	 */
	CPixelType f_Sample(float f_x, float f_y) const
	{
		//return f_At(f_x, f_y); // nearest neighbor :(

		int n_x = (int)f_x;
		int n_y = (int)f_y;
		float f_frac_x = f_x - n_x;
		float f_frac_y = f_y - n_y;

		int n_w_m = m_n_width_mask;
		int n_h_m = m_n_height_mask; // antialiass

		n_x &= n_w_m;
		n_y <<= m_n_height_shift;
		n_y &= n_h_m;

		int n_x1 = (n_x + 1) & n_w_m;
		int n_y1 = (n_y + m_n_width) & n_h_m;

		return (m_p_data[n_x | n_y] * (1 - f_frac_x) +
			   m_p_data[n_x1 | n_y] * f_frac_x) * (1 - f_frac_y) +
			   (m_p_data[n_x | n_y1] * (1 - f_frac_x) +
			   m_p_data[n_x1 | n_y1] * f_frac_x) * f_frac_y;
	}

	/*
	 *	static inline int n_Bit_Num(int n_x)
	 *		- returns number of bits taken by n_x
	 */
	static inline int n_Bit_Num(int n_x)
	{
		int n_result = -1;
		for(; n_x; ++ n_result, n_x >>= 1)
			;
		return n_result;
	}
};

#endif //__IMAGE_2D_INCLUDED
