/*
								+---------------------------------+
								|                                 |
								|    ***  CubeMap sampler   ***   |
								|                                 |
								|   Written by Lukas Polok 2007   |
								|                                 |
								|           ImageCube.h           |
								|                                 |
								+---------------------------------+
*/

#ifndef __IMAGE_CUBE_MAP_SAMPLER_INCLUDED
#define __IMAGE_CUBE_MAP_SAMPLER_INCLUDED

#include "Image.h"

/*
 *	template <class CPixelType>
 *	class CCubeSampler
 *		- simple class, implementing OpenGL-style cubemap sampling
 */
template <class CPixelType>
class CCubeSampler {
protected:
	CImage2D<CPixelType> *m_p_image_list[6];
	float m_f_texcoord_max;
	int m_n_size;

public:
	/*
	 *	CCubeSampler::<unnamed>
	 *		- cube side names
	 */
	enum {
		side_Neg_X = 0,
		side_Pos_X,
		side_Neg_Y,
		side_Pos_Y,
		side_Neg_Z,
		side_Pos_Z
	};

	/*
	 *	CCubeSampler::CCubeSampler(int n_size)
	 *		- constructor
	 *		- creates a new cube-map with six new images with size
	 *		  n_size per n_size pixels and undefined contents
	 *		- call b_Status() to see if there was enough memory for images
	 */
	CCubeSampler(int n_size)
		:m_n_size(n_size)
	{
		for(int i = 0; i < 6; ++ i)
			m_p_image_list[i] = new CImage2D<CPixelType>(n_size, n_size);
		// init image list

		m_f_texcoord_max = 1 - 1.0f / n_size;
		// calculate texcoord clamp limits
	}

	/*
	 *	CCubeSampler::CCubeSampler(CImage2D<CPixelType> *p_neg_x, CImage2D<CPixelType> *p_pos_x,
	 *		CImage2D<CPixelType> *p_neg_y, CImage2D<CPixelType> *p_pos_y,
	 *		CImage2D<CPixelType> *p_neg_z, CImage2D<CPixelType> *p_pos_z)
	 *		- constructor
	 *		- creates a new cubemap referencing given images
	 *		- note all images should be square, same size and must not be 0
	 */
	CCubeSampler(CImage2D<CPixelType> *p_neg_x, CImage2D<CPixelType> *p_pos_x,
		CImage2D<CPixelType> *p_neg_y, CImage2D<CPixelType> *p_pos_y,
		CImage2D<CPixelType> *p_neg_z, CImage2D<CPixelType> *p_pos_z)
	{
		m_p_image_list[0] = p_neg_x;
		m_p_image_list[1] = p_pos_x;
		m_p_image_list[2] = p_neg_y;
		m_p_image_list[3] = p_pos_y;
		m_p_image_list[4] = p_neg_z;
		m_p_image_list[5] = p_pos_z;
		// init image list

		m_n_size = p_neg_x->n_Width();
		m_f_texcoord_max = 1 - 1.0f / m_n_size;
		// calculate texcoord clamp limits
	}

	/*
	 *	CCubeSampler::~CCubeSampler()
	 *		- destructor
	 */
	~CCubeSampler()
	{
		for(int i = 0; i < 6; ++ i) {
			if(m_p_image_list[i])
				delete m_p_image_list[i];
		}
	}

	/*
	 *	int CCubeSampler::n_Size() const
	 *		- returns cube sampler resolution
	 */
	int n_Size() const
	{
		return m_n_size;
	}

	/*
	 *	bool CCubeSampler::b_Status() const
	 *		- returns true in case all images are nonzero
	 *		  and ready to use, otherwise returns false
	 *		- note in case this returns false, calling f_Sample()
	 *		  will most likely cause memory access violation
	 */
	bool b_Status() const
	{
		for(int i = 0; i < 6; ++ i) {
			if(!m_p_image_list[i] || !m_p_image_list[i]->b_Status())
				return false;
		}
		return true;
	}

	/*
	 *	inline CCubeSampler::CImage2D<CPixelType> &r_Image(int n_side)
	 *		- returns reference to 2D image on n_side side of cube
	 *		- n_side is one of side_Neg_X, side_Pos_X, side_Neg_Y,
	 *		  side_Pos_Y, side_Neg_Z or side_Pos_Z
	 */
	inline CImage2D<CPixelType> &r_Image(int n_side)
	{
		return *m_p_image_list[n_side];
	}

	/*
	 *	inline const CCubeSampler::CImage2D<CPixelType> &r_Image(int n_side) const
	 *		- returns constant reference to 2D image on n_side side of cube
	 *		- n_side is one of side_Neg_X, side_Pos_X, side_Neg_Y,
	 *		  side_Pos_Y, side_Neg_Z or side_Pos_Z
	 */
	inline const CImage2D<CPixelType> &r_Image(int n_side) const
	{
		return *m_p_image_list[n_side];
	}

	/*
	 *	CPixelType CCubeSampler::f_Sample(float f_s, float f_t, float f_r) const
	 *		- returns bilinear-filtered sample at coordinates (f_s, f_t, f_r)
	 *		- note texture wrap mode is "clamp to edge", coordinates
	 *		  don't have to be normalized and textures aren't filtered
	 *		  accross the edges (mimicking OpenGL style)
	 */
	CPixelType f_Sample(float f_s, float f_t, float f_r) const
	{
		float f_inv_len = 1.0f / (float)sqrt(f_s * f_s + f_t * f_t + f_r * f_r);
		f_s *= f_inv_len;
		f_t *= f_inv_len;
		f_r *= f_inv_len;
		// normalize texcoords

		float f_abs_s = (float)fabs(f_s);
		float f_abs_t = (float)fabs(f_t);
		float f_abs_r = (float)fabs(f_r);
		// calculate abs values

		bool b_s_greater_t = f_abs_s > f_abs_t;
		if(b_s_greater_t && f_abs_s > f_abs_r) {
			if(f_s > 0) {
				return m_p_image_list[1]->f_Sample(
					f_Clamp((1 - f_r / f_abs_s) * .5f), f_Clamp((1 - f_t / f_abs_s) * .5f));
			} else {
				return m_p_image_list[0]->f_Sample(
					f_Clamp((f_r / f_abs_s + 1) * .5f), f_Clamp((1 - f_t / f_abs_s) * .5f));
			}
		} else if(!b_s_greater_t && f_abs_t > f_abs_r) {
			if(f_t > 0) {
				return m_p_image_list[3]->f_Sample(
					f_Clamp((f_s / f_abs_t + 1) * .5f), f_Clamp((f_r / f_abs_t + 1) * .5f));
			} else {
				return m_p_image_list[2]->f_Sample(
					f_Clamp((f_s / f_abs_t + 1) * .5f), f_Clamp((1 - f_r / f_abs_t) * .5f));
			}
		} else {
			if(f_r > 0) {
				return m_p_image_list[5]->f_Sample(
					f_Clamp((f_s / f_abs_r + 1) * .5f), f_Clamp((1 - f_t / f_abs_r) * .5f));
			} else {
				return m_p_image_list[4]->f_Sample(
					f_Clamp((1 - f_s / f_abs_r) * .5f), f_Clamp((1 - f_t / f_abs_r) * .5f));
			}
		}
		// decide greatest component and calculate correct 2D sample coords
	}

protected:
	inline float f_Clamp(float f_coord) const
	{
		f_coord -= .5f / m_n_size; // how 'bout that?
		return ((f_coord >= 0)? ((f_coord <= m_f_texcoord_max)?
			f_coord : m_f_texcoord_max) : 0) * m_n_size;
	}
};

#endif //__IMAGE_CUBE_MAP_SAMPLER_INCLUDED
