/*
								+----------------------------------+
								|                                  |
								|    ***  SkyBox generator  ***    |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|             Main.cpp             |
								|                                  |
								+----------------------------------+
*/

#if defined(_MSC_VER) && !defined(__MWERKS__)
#pragma warning(disable:4786)
#endif
// disable warning C4786 "identifier was truncated to '255' characters in the debug information"

#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#include <crtdbg.h>
#define R_OK 04
#include <io.h> // _access()
#else
#include <assert.h>
#define _ASSERTE(x) assert(x)
#include <unistd.h> // access()
#define _access(a,b) access((a),(b))
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <GL/glext.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <vector>
#include <string>
#include <algorithm>
#include <functional>
#include "Raytracer.h"
#include "Dither.h"
#include "TgaOutput.h"
#include "BmpOutput.h"

static float f_mouse_x, f_mouse_y;
// mouse cursor in OpenGL coordinates

static float f_scene_z = 0;
static float f_angle_y = 1.8f, f_angle_x = 0;
// scene rotation

static int n_width = 640, n_height = 480;
// window size

// here we usualy draw glViewport and all commands
// to be called just once before drawing many frames
void onResize(int n_new_width, int n_new_height)
{
	glViewport(0, 0, n_width = n_new_width, n_height = n_new_height);
	// set new viewport

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(90, (float)n_width / n_height, 10, 10000);
	// set perspective projection

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// set camera

	glColor3f(1, 1, 1);
	// white

	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
}

// here we render frame
void onDisplay()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// reset matrix

	glTranslatef(0, 0, f_scene_z * 1000);
	glRotatef(f_angle_x * 90, 1, 0, 0);
	glRotatef(f_angle_y * 90, 0, 1, 0);
	// rotate/translate scene mouse-wisely

	const float p_cube_verts[8][3] = {
		{-1,  1, -1}, // l t f
		{-1, -1, -1}, // l b f
		{ 1, -1, -1}, // r b f
		{ 1,  1, -1}, // r t f

		{-1,  1,  1}, // l t b
		{-1, -1,  1}, // l b b
		{ 1, -1,  1}, // r b b
		{ 1,  1,  1}, // r t b
	};
	const int p_cube_faces[6][4] = {
		{3, 2, 6, 7}, // right
		{4, 5, 1, 0}, // left
		{4, 0, 3, 7}, // top
		{1, 5, 6, 2}, // bottom
		{7, 6, 5, 4}, // back
		{0, 1, 2, 3}, // front
	};

	glColor3f(1, 1, 1);
	glEnable(GL_TEXTURE_CUBE_MAP_ARB);
	glBegin(GL_QUADS);
	for(int i = 0; i < 6; ++ i) {
		for(int j = 0; j < 4; ++ j) {
			int n_vertex = p_cube_faces[i][j];
			glTexCoord3f(p_cube_verts[n_vertex][0], p_cube_verts[n_vertex][1],
				p_cube_verts[n_vertex][2]);
			glVertex3f(p_cube_verts[n_vertex][0] * 100, p_cube_verts[n_vertex][1] * 100,
				p_cube_verts[n_vertex][2] * 100);
		}
	}
	glEnd();
	// draw box with cubemap texcoords

	glutSwapBuffers();
	// display results
}

static bool b_left_button = false;

// here we receive notifications on mouse clicks
// x and y are cursor coordinates in pixels
void onMouseClick(int button, int state, int x, int y)
{
	f_mouse_x = (2.0f * x) / n_width - 1;
	f_mouse_y = 1 - (2.0f * y) / n_height;
	// get current mouse position

	state = 1; // shut up compiler warning about unused params

	switch(button) {
    case GLUT_LEFT_BUTTON:
		b_left_button = true;
		break;
    case GLUT_MIDDLE_BUTTON:
    case GLUT_RIGHT_BUTTON:
		b_left_button = false;
		break;
	};
	// in case button is going to be important for dragging, it's necessary to remember it
}

// here we receive notifications on mouse motion when dragging
// something, x and y are cursor coordinates in pixels
void onMouseMotion(int x, int y)
{
	float f_new_mouse_x = (2.0f * x) / n_width - 1;
	float f_new_mouse_y = 1 - (2.0f * y) / n_height;
	// get current mouse position

	if(b_left_button) {
		f_angle_y += f_new_mouse_x - f_mouse_x;
		f_angle_x += f_new_mouse_y - f_mouse_y;
	} else
		f_scene_z += f_new_mouse_x - f_mouse_x;
	// rotate scene

	f_mouse_x = f_new_mouse_x;
	f_mouse_y = f_new_mouse_y;
	// update mouse coords

	glutPostRedisplay();
	// asume we changed something and now we want to redraw ...
}

// here we receive notifications on mouse motion when just moving
// the mouse, x and y are cursor coordinates in pixels
void onPassiveMouseMotion(int x, int y)
{
	f_mouse_x = (2.0f * x) / n_width - 1;
	f_mouse_y = 1 - (2.0f * y) / n_height;
	// get current mouse position
}

bool b_HaveExtension(const char *p_s_ext_name)
{
	const char *p_s_exts = (const char*)glGetString(GL_EXTENSIONS);
	const char *p_s_ptr;
	if(p_s_exts && (p_s_ptr = strstr(p_s_exts, p_s_ext_name))) {
		unsigned int n_length = strlen(p_s_ext_name);
		if((p_s_ptr == p_s_exts || isspace(p_s_ptr[-1])) &&
		   (p_s_ptr[n_length] == 0 || isspace(p_s_ptr[n_length])))
			return true;
		// in case before is whitespace or beginning and after is whitespace or end, we have it
	}
	// determine wheter extension is supported

	return false;
}

void Dither_Image(unsigned int *p_raster, CImage2D<Vector3f> &r_cube_face)
{
	_ASSERTE(r_cube_face.n_Width() == r_cube_face.n_Height());

	for(int y = 0, n_resolution = r_cube_face.n_Width();
	   y < n_resolution; ++ y, p_raster += n_resolution) {
		int n_x_start = (y & 1)? n_resolution - 1 : 0;
		int n_x_stop = (y & 1)? -1 : n_resolution;
		int n_x_step = (y & 1)? -1 : 1;
		// have to alternate line processing direction because of dithering

		for(int x = n_x_start; x != n_x_stop; x += n_x_step) {
			Vector3f v_color = r_cube_face.f_Get(x, y);
			v_color = v_color.v_Min(Vector3f(1, 1, 1)).v_Max(Vector3f(0, 0, 0)) * 255;
			// clamp to [0, 1], scale to unsigned byte range

			Vector3f v_r_diff(CVarcoeffED::v_Error((int)v_color.f_X(),
				v_color.f_X() - (int)v_color.f_X()));
			Vector3f v_g_diff(CVarcoeffED::v_Error((int)v_color.f_Y(),
				v_color.f_Y() - (int)v_color.f_Y()));
			Vector3f v_b_diff(CVarcoeffED::v_Error((int)v_color.f_Z(),
				v_color.f_Z() - (int)v_color.f_Z()));
			const float f = 1.0f / 255;
			Vector3f v_right(v_r_diff.f_X() * f, v_g_diff.f_X() * f, v_b_diff.f_X() * f);
			Vector3f v_down_left(v_r_diff.f_Y() * f, v_g_diff.f_Y() * f, v_b_diff.f_Y() * f);
			Vector3f v_down(v_r_diff.f_Z() * f, v_g_diff.f_Z() * f, v_b_diff.f_Z() * f);
			if(y & 1) {
				if(x)
					r_cube_face.f_At(x - 1, y) += v_right;
				if(y + 1 < n_resolution) {
					if(x + 1 < n_resolution)
						r_cube_face.f_At(x + 1, y + 1) += v_down_left;
					r_cube_face.f_At(x, y + 1) += v_down;
				}
			} else {
				if(x + 1 < n_resolution)
					r_cube_face.f_At(x + 1, y) += v_right;
				if(y + 1 < n_resolution) {
					if(x)
						r_cube_face.f_At(x - 1, y + 1) += v_down_left;
					r_cube_face.f_At(x, y + 1) += v_down;
				}
			}
			// apply Ostroumhkov's dithering scheme
			// (alternates error diffusion direction at even / odd lines)

			p_raster[x] = 0xff000000 | ((int)v_color.f_Z() << 16) |
				((int)v_color.f_Y() << 8) | (int)v_color.f_X();
			// convert to RGBA
		}
	}
	// convert colors from float to RGBA8
}

bool Write_Images(CCubeSampler<Vector3f> &r_output_color, const char *p_s_filename)
{
	unsigned int *p_image_buffer;
	if(!(p_image_buffer = new unsigned int[r_output_color.n_Size() * r_output_color.n_Size()]))
		return false;
	// alloc buffer for texture data

	for(int i = 0; i < 6; ++ i) {
		Dither_Image(p_image_buffer, r_output_color.r_Image(i));
		// convert colors from float to RGBA8

		char p_s_side[5];
		sprintf(p_s_side, "%s%c", (i % 2)? "pos" : "neg", 'x' + i / 2);
		// create side name string

		std::string s_filename;
		s_filename.reserve(strlen(p_s_filename) + 5);
		if(s_filename.capacity() < strlen(p_s_filename) + 5)
			return false;
		s_filename.assign(p_s_filename);
		unsigned int n_pos;
		if((n_pos = s_filename.find("%s")) != std::string::npos) {
			s_filename.erase(n_pos, 2);
			s_filename.insert(n_pos, p_s_side);
		} else if((n_pos = s_filename.rfind(".")) != std::string::npos)
			s_filename.insert(n_pos, p_s_side);
		else
			s_filename += p_s_side;
		// create filename string

		if(strstr(p_s_filename, ".tga")) {
			if(!Save_TGA(s_filename.c_str(), r_output_color.n_Size(),
			   r_output_color.n_Size(), p_image_buffer)) {
				fprintf(stderr, "error: error writing image \'%s\' (disk full?)\n",
					s_filename.c_str());
				delete[] p_image_buffer;
				return false;
			}
		} else {
			if(!Save_BMP(s_filename.c_str(), r_output_color.n_Size(),
			   r_output_color.n_Size(), p_image_buffer)) {
				fprintf(stderr, "error: error writing image \'%s\' (disk full?)\n",
					s_filename.c_str());
				delete[] p_image_buffer;
				return false;
			}
		}
		// save image
	}
	// generate fixed-point textures

	delete[] p_image_buffer;
	// delete buffer

	return true;
}

GLuint n_Create_CubeMap(CCubeSampler<Vector3f> &r_output_color, bool b_create_mipmaps)
{
	if(!b_HaveExtension("GL_ARB_texture_cube_map"))
		return 0;
	// need this extension

	GLuint n_cubemap;
	glGenTextures(1, &n_cubemap);
	glEnable(GL_TEXTURE_CUBE_MAP_ARB);
	glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, n_cubemap);
	// create new texture handle and bind it

	const GLenum p_side_name_list[6] = {
		GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB,
		GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB,
		GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB,
		GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB,
		GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB,
		GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB
	};
	// names of cube-map sides

	if(b_create_mipmaps) {
		if(b_HaveExtension("GL_SGIS_generate_mipmap"))
			glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
		else
			b_create_mipmaps = false;
	}
	// prepare mip-map levels if required (let opengl calculate them, for now)

	unsigned int *p_image_buffer;
	if(!(p_image_buffer = new unsigned int[r_output_color.n_Size() * r_output_color.n_Size()])) {
		glDeleteTextures(1, &n_cubemap);
		return 0;
	}
	// alloc buffer for texture data

	for(int i = 0; i < 6; ++ i) {
		Dither_Image(p_image_buffer, r_output_color.r_Image(i));
		// convert colors from float to RGBA8

		glTexImage2D(p_side_name_list[i], 0, GL_RGBA, r_output_color.n_Size(),
			r_output_color.n_Size(), 0, GL_RGBA, GL_UNSIGNED_BYTE, p_image_buffer);
		// specify single face texture
	}
	// feed OpenGL with textures

	delete[] p_image_buffer;
	// delete buffer

	if(!b_create_mipmaps) {
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	} else {
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	// set proper texture filtering

	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	// clamp to edge is important!

	if(glGetError() != GL_NO_ERROR) {
		glDeleteTextures(1, &n_cubemap);
		return 0;
	}

	return n_cubemap;
}

/*
 *	class CApplyOnTerrain
 *		- simple adaptor object that applies operator on terrain
 */
class CApplyOnTerrain {
protected:
	CHeightField &m_r_heightmap;

public:
	/*
	 *	inline CApplyOnTerrain::CApplyOnTerrain(CHeightField &r_heightmap)
	 *		- default constructor, r_heightmap specifies the terrain
	 */
	inline CApplyOnTerrain(CHeightField &r_heightmap)
		:m_r_heightmap(r_heightmap)
	{}

	/*
	 *	inline void CApplyOnTerrain::operator ()(
	 *		const CTaskDescription::TTerrainDescription::TTerrainOp oper)
	 *		- applies operator oper on terrain passed to constructor
	 */
	inline void operator ()(const CTaskDescription::TTerrainDescription::TTerrainOp oper)
	{
		oper.Apply(m_r_heightmap);
	}
};

void PrintHelp()
{
	printf("synopsis:\nskybox_gen -i <taskfile> [-o <image%%s.tga>]"
		" [-v|--verbose] [-s|--see-results] [-h|--help]\n\n"
		"where taskfile is file with task description,\n"
		"image%%s.tga is name of skybox images (%%s is needed to fill-in name of cubemap face),\n"
		"-v or --verbose displays verbose info (which is useful because raytracing takes some\n"
		"time and watching blank console tends to be excrutiatingly boring),\n"
		"-s or --see-results launches GLUT window to see the results and\n"
		"-h or --help displays this screen\n");
}

int main(int n_arg_num, const char **p_arg_list)
{
	const char *p_s_task_file = 0;
	const char *p_s_output_file = 0;
	bool b_verbose = false;
	bool b_view_results = false;

	if(n_arg_num == 1) {
		PrintHelp();
		return -1;
	}

	for(int i = 1; i < n_arg_num; ++ i) {
		if(!strcmp(p_arg_list[i], "-i")) {
			if(i + 1 == n_arg_num) {
				fprintf(stderr, "error: argument \'%s\' is supposed"
					" to be followed by value\n", p_arg_list[i]);
				return -1;
			}
			p_s_task_file = p_arg_list[++ i];
		} else if(!strcmp(p_arg_list[i], "-o")) {
			if(i + 1 == n_arg_num) {
				fprintf(stderr, "error: argument \'%s\' is supposed"
					" to be followed by value\n", p_arg_list[i]);
				return -1;
			}
			p_s_output_file = p_arg_list[++ i];
		} else if(!strcmp(p_arg_list[i], "-v") || !strcmp(p_arg_list[i], "--verbose")) {
			b_verbose = true;
		} else if(!strcmp(p_arg_list[i], "-s") || !strcmp(p_arg_list[i], "--see-results")) {
			b_view_results = true;
		} else if(!strcmp(p_arg_list[i], "-h") || !strcmp(p_arg_list[i], "--help")) {
			PrintHelp();
			return 0;
		} else {
			fprintf(stderr, "error: unknown argument: \'%s\'\n", p_arg_list[i]);
			return -1;
		}
	}
	// parse cmdline params

	if(!p_s_task_file) {
		fprintf(stderr, "error: no task file specified\n");
		return -1;
	}
	if(!p_s_output_file && !b_view_results)
		fprintf(stderr, "warning: no results will be displayed nor saved ... what good is it?");
	// check params

	if(b_verbose)
		printf("\t\t\t\t=== Skybox raytracer ===\n\n");

	if(b_verbose)
		printf("parsing task description ...\n");
	CTaskDescription task;
	int n_line, n_column;
	const char *p_s_err_str = "lexical error";
	if(!task.Parse(p_s_task_file, p_s_err_str, n_line, n_column)) {
		if(_access(p_s_task_file, R_OK)) {
			fprintf(stderr, "error: file \'%s\' can't be opened for reading\n", p_s_task_file);
			return -1;
		}
		fprintf(stderr, "error in task description on line %d, column %d: %s\n",
			n_line + 1, n_column + 1, p_s_err_str);
		return -1;
	}
	// parser test

	if(b_verbose)
		printf("building heightmap ...\n");
	CHeightField heightmap(task.r_Terrain().n_width, task.r_Terrain().n_height);
	//const float f_quad_size = task.r_Terrain().f_tile_size;
	if(!heightmap.b_Status()) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	// alloc heightmap

	std::for_each(task.r_Terrain().op_list.begin(),
		task.r_Terrain().op_list.end(), CApplyOnTerrain(heightmap));
	// apply operators on terrain

	CCubeSampler<Vector3f> output_color(task.r_Output().n_texture_size);
	if(!output_color.b_Status()) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	// alloc output cubemap

	long n_time_start = clock();
	if(!CSkyBoxRaytracer::Raytrace_Skybox(task, heightmap, output_color, b_verbose)) {
		fprintf(stderr, "error: error rendering cubemap\n");
		return -1;
	}
	float f_time = (clock() - n_time_start) / (float)CLOCKS_PER_SEC;
	if(b_verbose)
		printf("\t- it took %f sec\n", f_time);
	// raytrace cubemap, after mesh is done so it can be raytraced

	if(p_s_output_file) {
		if(b_verbose)
			printf("writing output images ...\n");
		if(!strstr(p_s_output_file, "%s")) {
			fprintf(stderr, "warning: output filename doesn't contain \'%%s\' tag.\n"
				"cubemap face name will be inserted before file extension.");
		}
		if(!Write_Images(output_color, p_s_output_file)) {
			fprintf(stderr, "error: error writing output images\n");
			return -1;
		}
	}
	// write output images

	if(!b_view_results)
		return 0;
	// we're done in case we don't want to see

	if(b_verbose)
		printf("launching GLUT window ...\n");

	glutInit(&n_arg_num, (char **)p_arg_list);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowSize(n_width, n_height);
	glutCreateWindow("cube-map raytracer");
	// init OpenGL using GLUT

	if(!n_Create_CubeMap(output_color, true)) {
		fprintf(stderr, "error: error creating cubemap\n");
		if(!b_HaveExtension("GL_ARB_texture_cube_map"))
			fprintf(stderr, "error: GL_ARB_texture_cube_map not supported on your hardware\n");
		return -1;
	}
	// create cubemap we could admire (don't save texture id,
	// texture will simply remain bound until end of the program)

	glutDisplayFunc(onDisplay);
	glutReshapeFunc(onResize);
	glutMouseFunc(onMouseClick);
	glutMotionFunc(onMouseMotion);
	glutPassiveMotionFunc(onPassiveMouseMotion);
	// register our callbacks

	glutMainLoop();
	// begin rendering

	return 0;
}
