/*
								+---------------------------------+
								|                                 |
								| *** Simple vector templates *** |
								|                                 |
								|   Design by Lukas Polok 2007    |
								|   Implemented by Vit Kolar      |
								|                                 |
								|            Vector.h             |
								|                                 |
								+---------------------------------+
*/

#ifndef __VECTOR_INCLUDED
#define __VECTOR_INCLUDED

/*
 *	template <class CScalarType, const int n_dimension>
 *	class CVector_nD
 *		- generic n-dimensional vector template
 *
 *		- zadn
 *			- ablona pro obecn n-rozmrn vektor
 *			- jsou implementovny jen obecn operace, bli operace pro jednotliv
 *			  dimenze potom budou ve specializovanch zddnch tdch
 *			- poadovan operace jsou + - * / += -= *= /= (vektor s vektorem nebo vektor
 *			  s cislem), unarni -, min / max (component-wise a nebo unarni), suma, delka^2,
 *			  delka, skalarni soucin, normalizace, normalizace na danou delku
 *			- dodrovat konzistentn komentovn
 *
 */
template <class CScalarType, const int n_dimension>
class CVector_nD {
//protected:
	CScalarType m_p_coord[n_dimension];

public:
	/*
	 *	inline CVector_nD()
	 *		- default constructor, have no effect
	 */
	inline CVector_nD()
	{}

	/*
	 *	inline CVector_nD(CScalarType f_x)
	 *		- constructor setting first coordinate to
	 *		  f_x and others (if present) to 0
	 */
	inline CVector_nD(CScalarType f_x)
	{
		_ASSERTE(n_dimension > 0);
		m_p_coord[0] = f_x;
		for(int i = 1; i < n_dimension; ++ i)
			m_p_coord[i] = 0;
	}

	/*
	 *	inline CVector_nD(CScalarType f_x, CScalarType f_y)
	 *		- constructor setting first two coordinates to
	 *		  f_x, f_y and others (if present) to 0
	 */
	inline CVector_nD(CScalarType f_x, CScalarType f_y)
	{
		_ASSERTE(n_dimension > 1);
		m_p_coord[0] = f_x;
		m_p_coord[1] = f_y;
		for(int i = 2; i < n_dimension; ++ i)
			m_p_coord[i] = 0;
	}

	/*
	 *	inline CVector_nD(CScalarType f_x, CScalarType f_y, CScalarType f_z)
	 *		- constructor setting first three coordinates to
	 *		  f_x, f_y, f_z and others (if present) to 0
	 */
	inline CVector_nD(CScalarType f_x, CScalarType f_y, CScalarType f_z)
	{
		_ASSERTE(n_dimension > 2);
		m_p_coord[0] = f_x;
		m_p_coord[1] = f_y;
		m_p_coord[2] = f_z;
		for(int i = 3; i < n_dimension; ++ i)
			m_p_coord[i] = 0;
	}

	/*
	 *	inline CVector_nD(CScalarType f_x, CScalarType f_y, CScalarType f_z, CScalarType f_w)
	 *		- constructor setting first four coordinates to
	 *		  f_x, f_y, f_z, f_w and others (if present) to 0
	 */
	inline CVector_nD(CScalarType f_x, CScalarType f_y, CScalarType f_z, CScalarType f_w)
	{
		_ASSERTE(n_dimension > 3);
		m_p_coord[0] = f_x;
		m_p_coord[1] = f_y;
		m_p_coord[2] = f_z;
		m_p_coord[3] = f_w;
		for(int i = 4; i < n_dimension; ++ i)
			m_p_coord[i] = 0;
	}

	/*
	 *	inline CVector_nD(const CVector_nD &r_v_vec)
	 *		- copy-constructor
	 */
	inline CVector_nD(const CVector_nD &r_v_vec)
	{
		for(int i = 0; i < n_dimension; ++ i)
			m_p_coord[i] = r_v_vec.m_p_coord[i];
	}

	/*
	 *	inline const CVector_nD &operator =(const CVector_nD &r_v_vec)
	 *		- operator = for vectors
	 *		- returns const reference to this vector
	 */
	inline const CVector_nD &operator =(const CVector_nD &r_v_vec)
	{
		for(int i = 0; i < n_dimension; ++ i)
			m_p_coord[i] = r_v_vec.m_p_coord[i];
		return *this;
	}

	/*
	 *	inline CScalarType operator [](int i) const
	 *		- returns i-th element of vector
	 */
	inline CScalarType operator [](int i) const
	{
		_ASSERTE(i >= 0 && i < n_dimension);
		return m_p_coord[i];
	}

	/*
	 *	inline CScalarType &operator [](int i)
	 *		- returns reference to i-th element of vector
	 */
	inline CScalarType &operator [](int i)
	{
		_ASSERTE(i >= 0 && i < n_dimension);
		return m_p_coord[i];
	}

	/*
	 *	inline CVector_nD operator *(CScalarType f_scale) const
	 *		- multiplication operator for scaling vector components by number
	 */
	inline CVector_nD operator *(CScalarType f_scale) const
	{
		CVector_nD v_out;
		for(int i = 0; i < n_dimension; ++ i)
			v_out.m_p_coord[i] = m_p_coord[i] * f_scale;
		return v_out;
	}

	/*
	 *	inline const CVector_nD &operator *=(CScalarType f_scale)
	 *		- multiplication operator for scaling vector components by number
	 */
	inline const CVector_nD &operator *=(CScalarType f_scale)
	{
		for(int i = 0; i < n_dimension; ++ i)
			m_p_coord[i] *= f_scale;
		return *this;
	}

	/*
	 *	inline CVector_nD operator *(const CVector_nD &r_v_vec) const
	 *		- multiplication operator for component-wise multiplication of this vector by r_v_vec
	 */
	inline CVector_nD operator *(const CVector_nD &r_v_vec) const
	{
		CVector_nD v_out;
		for(int i = 0; i < n_dimension; ++ i)
			v_out.m_p_coord[i] = m_p_coord[i] * r_v_vec.m_p_coord[i];
		return v_out;
	}

	/*
	 *	inline const CVector_nD &operator *=(const CVector_nD &r_v_vec)
	 *		- multiplication operator for component-wise multiplication of this vector by r_v_vec
	 *		- returns const reference to this vector
	 */
	inline const CVector_nD &operator *=(const CVector_nD &r_v_vec)
	{
		for(int i = 0; i < n_dimension; ++ i)
			m_p_coord[i] *= r_v_vec.m_p_coord[i];
		return *this;
	}

	/*
	 *	inline CVector_nD operator /(CScalarType f_scale) const
	 *		- division operator for scaling vector components by number
	 */
	inline CVector_nD operator /(CScalarType f_scale) const
	{
		f_scale = 1 / f_scale;
		CVector_nD v_out;
		for(int i = 0; i < n_dimension; ++ i)
			v_out.m_p_coord[i] = m_p_coord[i] * f_scale;
		return v_out;
	}

	/*
	 *	inline const CVector_nD &operator /=(CScalarType f_scale)
	 *		- division operator for scaling vector components by number
	 */
	inline const CVector_nD &operator /=(CScalarType f_scale)
	{
		f_scale = 1 / f_scale;
		for(int i = 0; i < n_dimension; ++ i)
			m_p_coord[i] *= f_scale;
		return *this;
	}

	/*
	 *	inline CVector_nD operator /(const CVector_nD &r_v_vec) const
	 *		- division operator for component-wise multiplication of this vector by r_v_vec
	 */
	inline CVector_nD operator /(const CVector_nD &r_v_vec) const
	{
		CVector_nD v_out;
		for(int i = 0; i < n_dimension; ++ i)
			v_out.m_p_coord[i] = m_p_coord[i] / r_v_vec.m_p_coord[i];
		return v_out;
	}

	/*
	 *	inline const CVector_nD &operator /=(const CVector_nD &r_v_vec)
	 *		- division operator for component-wise multiplication of this vector by r_v_vec
	 *		- returns const reference to this vector
	 */
	inline const CVector_nD &operator /=(const CVector_nD &r_v_vec)
	{
		for(int i = 0; i < n_dimension; ++ i)
			m_p_coord[i] /= r_v_vec.m_p_coord[i];
		return *this;
	}

	/*
	 *	inline CVector_nD operator +(const CVector_nD &r_v_vec) const
	 *		- addition operator for component-wise addition of this vector with r_v_vec
	 */
	inline CVector_nD operator +(const CVector_nD &r_v_vec) const
	{
		CVector_nD v_out;
		for(int i = 0; i < n_dimension; ++ i)
			v_out.m_p_coord[i] = m_p_coord[i] + r_v_vec.m_p_coord[i];
		return v_out;
	}

	/*
	 *	inline CVector_nD operator +=(const CVector_nD &r_v_vec)
	 *		- addition operator for component-wise addition of this vector with r_v_vec
	 *		- returns const reference to this vector
	 */
	inline const CVector_nD &operator +=(const CVector_nD &r_v_vec)
	{
		for(int i = 0; i < n_dimension; ++ i)
			m_p_coord[i] += r_v_vec.m_p_coord[i];
		return *this;
	}

	/*
	 *	inline CVector_nD operator -(const CVector_nD &r_v_vec) const
	 *		- addition operator for component-wise addition of this vector with r_v_vec
	 */
	inline CVector_nD operator -(const CVector_nD &r_v_vec) const
	{
		CVector_nD v_out;
		for(int i = 0; i < n_dimension; ++ i)
			v_out.m_p_coord[i] = m_p_coord[i] - r_v_vec.m_p_coord[i];
		return v_out;
	}

	/*
	 *	inline CVector_nD operator -() const
	 *		- unary minus operator
	 */
	inline CVector_nD operator -() const
	{
		CVector_nD v_out;
		for(int i = 0; i < n_dimension; ++ i)
			v_out.m_p_coord[i] = -m_p_coord[i];
		return v_out;
	}

	/*
	 *	inline CVector_nD operator -=(const CVector_nD &r_v_vec)
	 *		- addition operator for component-wise addition of this vector with r_v_vec
	 *		- returns const reference to this vector
	 */
	inline const CVector_nD &operator -=(const CVector_nD &r_v_vec)
	{
		for(int i = 0; i < n_dimension; ++ i)
			m_p_coord[i] -= r_v_vec.m_p_coord[i];
		return *this;
	}

	/*
	 *	inline CScalarType f_Dot(const CVector_nD &r_v_vec) const
	 *		- returns dot product of this vector with r_v_vec
	 */
	inline CScalarType f_Dot(const CVector_nD &r_v_vec) const
	{
		CScalarType f_result = 0;
		for(int i = 0; i < n_dimension; ++ i)
			f_result += m_p_coord[i] * r_v_vec.m_p_coord[i];
		return f_result;
	}

	/*
	 *	inline CScalarType f_Length2() const
	 *		- returns squared length of this vector
	 */
	inline CScalarType f_Length2() const
	{
		return f_Dot(*this);
	}

	/*
	 *	inline CScalarType f_Length() const
	 *		- returns length of this vector
	 */
	inline CScalarType f_Length() const
	{
		return (CScalarType)sqrt(f_Dot(*this));
	}

	/*
	 *	inline void Normalize()
	 *		- creates unit vector
	 *		- note in case vector is equal to O
	 *		  (zero vector), this operation has no effect
	 */
	inline void Normalize()
	{
		CScalarType f_length = f_Length();
		if(f_length != 0) {
			f_length = 1 / f_length;
			for(int i = 0; i < n_dimension; ++ i)
				m_p_coord[i] *= f_length;
		}
		// normalize
	}

	/*
	 *	inline void Normalize(CScalarType f_desired_length)
	 *		- creates vector of length f_desired_length
	 *		- note in case vector is equal to O
	 *		  (zero vector), this operation has no effect
	 */
	inline void Normalize(CScalarType f_desired_length)
	{
		CScalarType f_length = f_Length();
		if(f_length != 0) {
			f_length = f_desired_length / f_length;
			for(int i = 0; i < n_dimension; ++ i)
				m_p_coord[i] *= f_length;
		}
		// normalize to f_desired_length
	}

	/*
	 *	inline CScalarType f_Sum() const
	 *		- returns sum of vector components
	 */
	inline CScalarType f_Sum() const
	{
		CScalarType f_result = 0;
		for(int i = 0; i < n_dimension; ++ i)
			f_result += m_p_coord[i];
		return f_result;
	}

	/*
	 *	inline CScalarType f_Min() const
	 *		- returns value of minimal component
	 */
	inline CScalarType f_Min() const
	{
		CScalarType f_result = m_p_coord[0];
		for(int i = 1; i < n_dimension; ++ i) {
			if(f_result > m_p_coord[i])
				f_result = m_p_coord[i];
		}
		return f_result;
	}

	/*
	 *	inline CScalarType f_Max() const
	 *		- returns value of maximal component
	 */
	inline CScalarType f_Max() const
	{
		CScalarType f_result = m_p_coord[0];
		for(int i = 1; i < n_dimension; ++ i) {
			if(f_result < m_p_coord[i])
				f_result = m_p_coord[i];
		}
		return f_result;
	}

	/*
	 *	inline CVector_nD v_Min(const CVector_nD &r_v_vec) const
	 *		- returns component-wise minimum of two vectors
	 */
	inline CVector_nD v_Min(const CVector_nD &r_v_vec) const
	{
		CVector_nD v_out;
		for(int i = 0; i < n_dimension; ++ i) {
			v_out.m_p_coord[i] = (m_p_coord[i] < r_v_vec.m_p_coord[i])?
				m_p_coord[i] : r_v_vec.m_p_coord[i];
		}
		return v_out;
	}

	/*
	 *	inline CVector_nD v_Max(const CVector_nD &r_v_vec) const
	 *		- returns component-wise maximum of two vectors
	 */
	inline CVector_nD v_Max(const CVector_nD &r_v_vec) const
	{
		CVector_nD v_out;
		for(int i = 0; i < n_dimension; ++ i) {
			v_out.m_p_coord[i] = (m_p_coord[i] < r_v_vec.m_p_coord[i])?
				r_v_vec.m_p_coord[i] : m_p_coord[i];
		}
		return v_out;
	}
};

/*
 *	template <class CScalarType>
 *	class Vector2
 *		- extension of generic 2D vector template to specialized 2D vector template
 *
 *		- zadn
 *			- implementovat potebn konstruktory a zpstupnit jednotliv souadnice
 *			  pomoc funkc f_X(), f_Y()
 *			- okomentovat ve stejnm stylu jako byla ablona CVector_nD
 *
 */
template <class CScalarType>
class Vector2 : public CVector_nD<CScalarType, 2> {
public:
	/*
	 *	inline Vector2()
	 *		- default constructor; has no effect
	 */
	inline Vector2()
	{}

	/*
	 *	inline Vector2(CScalarType f_x)
	 *		- constructor setting x to f_x and y to zero
	 */
	inline Vector2(CScalarType f_x)
		:CVector_nD<CScalarType, 2>(f_x)
	{}

	/*
	 *	inline Vector2(CScalarType f_x, CScalarType f_y)
	 *		- constructor setting x to f_x and y to f_y
	 */
	inline Vector2(CScalarType f_x, CScalarType f_y)
		:CVector_nD<CScalarType, 2>(f_x, f_y)
	{}

	/*
	 *	inline Vector2(const CVector_nD<CScalarType, 2> &r_v_vec)
	 *		- copy-constructor
	 */
	inline Vector2(const CVector_nD<CScalarType, 2> &r_v_vec)
		:CVector_nD<CScalarType, 2>(r_v_vec)
	{}

	/*
	 *	inline const Vector2 &operator =(const CVector_nD<CScalarType, 2> &r_v_vec)
	 *		- copy-operator
	 *		- returns const reference to this vector
	 */
	inline const Vector2 &operator =(const CVector_nD<CScalarType, 2> &r_v_vec)
	{
		for(int i = 0; i < 2; ++ i)
			(*this)[i] = r_v_vec[i];
		return *this;
	}

	/*
	 *	inline CScalarType f_X() const
	 *		- returns the x coordinate of this vector
	 */
	inline CScalarType f_X() const
	{
		return (*this)[0];
	}

	/*
	 *	inline CScalarType f_Y() const
	 *		- returns the y coordinate of this vector
	 */
	inline CScalarType f_Y() const
	{
		return (*this)[1];
	}

	/*
	 *	inline CScalarType &f_X() const
	 *		- returns reference to the x coordinate of this vector
	 */
	inline CScalarType &f_X()
	{
		return (*this)[0];
	}

	/*
	 *	inline CScalarType &f_Y() const
	 *		- returns reference to the x coordinate of this vector
	 */
	inline CScalarType &f_Y()
	{
		return (*this)[1];
	}
};

/*
 *	template <class CScalarType>
 *	class Vector3
 *		- extension of generic 3D vector template to specialized 3D vector template
 *
 *		- zadn
 *			- implementovat potebn konstruktory a zpstupnit jednotliv souadnice
 *			  pomoc funkc f_X(), f_Y(), f_Z(), implementovat vektorov souin,
 *			  zrcadlen vektoru podle normly a Gram-Schmidt ortogonalizaci
 *			- okomentovat ve stejnm stylu jako byla ablona CVector_nD
 *
 */
template <class CScalarType>
class Vector3 : public CVector_nD<CScalarType, 3> {
public:
	/*
	 *	inline Vector3()
	 *		- default constructor; has no effect
	 */
	inline Vector3()
	{}

	/*
	 *	inline Vector3(CScalarType f_x)
	 *		- constructor setting vector coordinates to (f_x, 0, 0)
	 */
	inline Vector3(CScalarType f_x)
		:CVector_nD<CScalarType, 3>(f_x)
	{}

	/*
	 *	inline Vector3(CScalarType f_x, CScalarType f_y)
	 *		- constructor setting vector coordinates to (f_x, f_y, 0)
	 */
	inline Vector3(CScalarType f_x, CScalarType f_y)
		:CVector_nD<CScalarType, 3>(f_x, f_y)
	{}

	/*
	 *	inline Vector3(CScalarType f_x, CScalarType f_y, CScalarType f_z)
	 *		- constructor setting vector coordinates to (f_x, f_y, f_z)
	 */
	inline Vector3(CScalarType f_x, CScalarType f_y, CScalarType f_z)
		:CVector_nD<CScalarType, 3>(f_x, f_y, f_z)
	{}

	/*
	 *	inline Vector3(const CVector_nD<CScalarType, 3> &r_v_vec)
	 *		- copy-constructor
	 */
	inline Vector3(const CVector_nD<CScalarType, 3> &r_v_vec)
		:CVector_nD<CScalarType, 3>(r_v_vec)
	{}

	/*
	 *	inline const Vector3 &operator =(const CVector_nD<CScalarType, 3> &r_v_vec)
	 *		- copy-operator
	 *		- returns const reference to this vector
	 */
	inline const Vector3 &operator =(const CVector_nD<CScalarType, 3> &r_v_vec)
	{
		for(int i = 0; i < 3; ++ i)
			(*this)[i] = r_v_vec[i];
		return *this;
	}

	/*
	 *	inline CScalarType f_X() const
	 *		- returns the x coordinate of this vector
	 */
	inline CScalarType f_X() const
	{
		return (*this)[0];
	}

	/*
	 *	inline CScalarType f_Y() const
	 *		- returns the y coordinate of this vector
	 */
	inline CScalarType f_Y() const
	{
		return (*this)[1];
	}

	/*
	 *	inline CScalarType f_Z() const
	 *		- returns the z coordinate of this vector
	 */
	inline CScalarType f_Z() const
	{
		return (*this)[2];
	}

	/*
	 *	inline CScalarType &f_X() const
	 *		- returns reference to the x coordinate of this vector
	 */
	inline CScalarType &f_X()
	{
		return (*this)[0];
	}

	/*
	 *	inline CScalarType &f_Y() const
	 *		- returns reference to the y coordinate of this vector
	 */
	inline CScalarType &f_Y()
	{
		return (*this)[1];
	}

	/*
	 *	inline CScalarType &f_Z() const
	 *		- returns reference to the z coordinate of this vector
	 */
	inline CScalarType &f_Z()
	{
		return (*this)[2];
	}

	/*
	 *	Vector3 v_Cross(const Vector3 &r_v_vec) const
	 *		- returns cross product of this vector with r_v_vec
	 */
	Vector3 v_Cross(const Vector3 &r_v_vec) const
	{
		return Vector3(f_Y() * r_v_vec.f_Z() - f_Z() * r_v_vec.f_Y(),
						f_Z() * r_v_vec.f_X() - f_X() * r_v_vec.f_Z(),
						f_X() * r_v_vec.f_Y() - f_Y() * r_v_vec.f_X());
	}

	/*
	 *	Vector3 v_Reflect(const Vector3 &r_v_vec) const
	 *		- returns vector reflected arround r_v_vec, which must be unit length
	 *		- returned vector has length equal to length of this vector
	 */
	Vector3 v_Reflect(const Vector3 &r_v_vec) const
	{
		return r_v_vec * (f_Dot(r_v_vec) * 2) - *this;
	}

	/*
	 *	Vector3 v_Orthogonal(const Vector3 &r_v_vec) const
	 *		- returns vector orthogonal to r_v_vec, which must be unit length,
	 *		  and lying in plane given by r_v_vec and this vector
	 */
	Vector3 v_Orthogonal(const Vector3 &r_v_vec) const
	{
		return *this - r_v_vec * f_Dot(r_v_vec);
	}
};

/*
 *	template <class CScalarType>
 *	class Vector4
 *		- extension of generic 4D vector template to specialized 4D vector template
 *
 *		- zadn
 *			- implementovat potebn konstruktory a zpstupnit jednotliv souadnice
 *			  pomoc funkc f_X(), f_Y(), f_Z(), f_W()
 *			- okomentovat ve stejnm stylu jako byla ablona CVector_nD
 *
 */
template <class CScalarType>
class Vector4 : public CVector_nD<CScalarType, 4> {
public:
	/*
	 *	inline Vector4()
	 *		- default constructor; has no effect
	 */
	inline Vector4()
	{}

	/*
	 *	inline Vector4(CScalarType f_x)
	 *		- constructor setting vector coordinates to (f_x, 0, 0, 1)
	 */
	inline Vector4(CScalarType f_x)
		:CVector_nD<CScalarType, 4>(f_x, 0, 0, 1)
	{}

	/*
	 *	inline Vector4(CScalarType f_x, CScalarType f_y)
	 *		- constructor setting vector coordinates to (f_x, f_y, 0, 1)
	 */
	inline Vector4(CScalarType f_x, CScalarType f_y)
		:CVector_nD<CScalarType, 4>(f_x, f_y, 0, 1)
	{}

	/*
	 *	inline Vector4(CScalarType f_x, CScalarType f_y, CScalarType f_z)
	 *		- constructor setting vector coordinates to (f_x, f_y, f_z, 1)
	 */
	inline Vector4(CScalarType f_x, CScalarType f_y, CScalarType f_z)
		:CVector_nD<CScalarType, 4>(f_x, f_y, f_z, 1)
	{}

	/*
	 *	inline Vector4(CScalarType f_x, CScalarType f_y, CScalarType f_z, CScalarType f_w)
	 *		- constructor setting vector coordinates to (f_x, f_y, f_z, f_w)
	 */
	inline Vector4(CScalarType f_x, CScalarType f_y, CScalarType f_z, CScalarType f_w)
		:CVector_nD<CScalarType, 4>(f_x, f_y, f_z, f_w)
	{}

	/*
	 *	inline Vector4(const CVector_nD<CScalarType, 4> &r_v_vec)
	 *		- copy-constructor
	 */
	inline Vector4(const CVector_nD<CScalarType, 4> &r_v_vec)
		:CVector_nD<CScalarType, 4>(r_v_vec)
	{}

	/*
	 *	inline const Vector4 &operator =(const CVector_nD<CScalarType, 4> &r_v_vec)
	 *		- copy-operator
	 *		- returns const reference to this vector
	 */
	inline const Vector4 &operator =(const CVector_nD<CScalarType, 4> &r_v_vec)
	{
		for(int i = 0; i < 4; ++ i)
			(*this)[i] = r_v_vec[i];
		return *this;
	}

	/*
	 *	inline CScalarType f_X() const
	 *		- returns the x coordinate of this vector
	 */
	inline CScalarType f_X() const
	{
		return (*this)[0];
	}

	/*
	 *	inline CScalarType f_Y() const
	 *		- returns the y coordinate of this vector
	 */
	inline CScalarType f_Y() const
	{
		return (*this)[1];
	}

	/*
	 *	inline CScalarType f_Z() const
	 *		- returns the z coordinate of this vector
	 */
	inline CScalarType f_Z() const
	{
		return (*this)[2];
	}

	/*
	 *	inline CScalarType f_W() const
	 *		- returns the w coordinate of this vector
	 */
	inline CScalarType f_W() const
	{
		return (*this)[3];
	}

	/*
	 *	inline CScalarType &f_X() const
	 *		- returns reference to the x coordinate of this vector
	 */
	inline CScalarType &f_X()
	{
		return (*this)[0];
	}

	/*
	 *	inline CScalarType &f_Y() const
	 *		- returns reference to the y coordinate of this vector
	 */
	inline CScalarType &f_Y()
	{
		return (*this)[1];
	}

	/*
	 *	inline CScalarType &f_Z() const
	 *		- returns reference to the z coordinate of this vector
	 */
	inline CScalarType &f_Z()
	{
		return (*this)[2];
	}

	/*
	 *	inline CScalarType &f_W() const
	 *		- returns reference to the w coordinate of this vector
	 */
	inline CScalarType &f_W()
	{
		return (*this)[3];
	}
};

typedef Vector2<float> Vector2f;
typedef Vector3<float> Vector3f;
typedef Vector4<float> Vector4f;

#endif //__VECTOR_INCLUDED
