/*
								+----------------------------------+
								|                                  |
								| ***  Poisson disc generator  *** |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|           Poisson.cpp            |
								|                                  |
								+----------------------------------+
*/

#ifdef WIN32
#include <crtdbg.h>
#else
#include <assert.h>
#define _ASSERTE(x) assert(x)
#endif
#include <vector>
#include <algorithm>
#include <functional>
#include <math.h>
#include "MiniVector.h"
#include "RandGen.h"
#include "Poisson.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif

const float f_pi = 3.1415926535897932384626433832795028841971697510f;

/*
 *								=== TDisc ===
 */

/*
 *	TDisc::TDisc(const Vector2f &r_v_center, float _f_radius)
 *		- default constructor
 *		- creates disc with center r_v_center and radius _f_radius
 */
TDisc::TDisc(const Vector2f &r_v_center, float _f_radius)
	:n_interval_num(1), f_radius(_f_radius), v_center(r_v_center)
{
	p_arc_angle[0] = TInterval(0, 2 * f_pi);
}

/*
 *	static float TDisc::f_Angle(const Vector2f &r_v_vector)
 *		- returns angle r_v_vector is pointing, returned angle is 0 to 2pi radians
 *		- r_v_vector has to be unit-length
 */
float TDisc::f_Angle(const Vector2f &r_v_vector)
{
	return (float)atan(r_v_vector.f_Y() / r_v_vector.f_X()) +
		((r_v_vector.f_X() < 0)? f_pi : ((r_v_vector.f_Y() < 0)? 2 * f_pi : 0));
}

/*
 *	bool TDisc::Subtract(const TDisc &r_t_disc)
 *		- subtracts r_t_disc disc's area from arc of this disc
 *		- r_t_disc is the other disc
 *		- operation is not commutative
 *		- returns true on disc collision, false on no collision
 */
bool TDisc::Subtract(const TDisc &r_t_disc)
{
	_ASSERTE(r_t_disc.f_radius == f_radius);

	Vector2f v_diff = r_t_disc.v_center - v_center;
	float f_len = v_diff.f_Length();

	if(f_len >= 2 * f_radius)
		return false;
	// discs miss each other

	_ASSERTE(f_len > 0);
	// do not handle situation when discs exactly overlap

	float f_alpha = (float)acos(f_len / (2 * f_radius));
	float f_angle = f_Angle(v_diff * (1.0f / f_len));
	// calculate incident angle and intersection disc (angular) length

	float f_angle0 = f_angle - f_alpha;
	float f_angle1 = f_angle + f_alpha;
	// calculate angle interval

	SubtractAngleInterval(f_angle0, f_angle1);
	// subtract angular range from this disc

	/*static int nn = 1;
	if(n_interval_num > nn) {
		nn = n_interval_num;
		printf("new maximum: %d\n", n_interval_num);
	}*/ // verifying maximal number of intervals theory

	return true;
}

/*
 *	void TDisc::SubtractAngleInterval(float f_angle0, float f_angle1)
 *		- substracts angular interval f_angle0 to f_angle1 from disc's arc
 *		- angles must be ordered so f_anle0 is less than f_angle1
 *		- note the disc structure is able to describe just one continuous interval,
 *		  in case more intervals originate by subtraction, the longest one is stored
 */
void TDisc::SubtractAngleInterval(float f_angle0, float f_angle1)
{
	while(f_angle0 > 2 * f_pi) {
		f_angle0 -= 2 * f_pi;
		f_angle1 -= 2 * f_pi;
	}
	while(f_angle0 < 0) {
		f_angle0 += 2 * f_pi;
		f_angle1 += 2 * f_pi;
	}
	// make sure angles are in plausible intervals

	for(;;) {
		float f_max = (f_angle1 <= 2 * f_pi)? f_angle1 : 2 * f_pi;
		TInterval t_sub_interval(f_angle0, f_max);

		TInterval p_new_arcs[max_Interval_Num];
		int n_arc_num = 0;

		for(int i = 0, n = n_interval_num; i < n; ++ i) {
			if(p_arc_angle[i].b_Disjunct(t_sub_interval)) {
				p_new_arcs[n_arc_num] = p_arc_angle[i];
				if(++ n_arc_num == max_Interval_Num)
					break;
				continue;
			}
			// intervals do not intersect

			{
				TInterval t_complement;
				if(!(t_complement = p_arc_angle[i].t_LeftComplement(t_sub_interval)).b_Empty()) {
					p_new_arcs[n_arc_num] = t_complement;
					if(++ n_arc_num == max_Interval_Num)
						break;
				}
			}
			{
				TInterval t_complement;
				if(!(t_complement = p_arc_angle[i].t_RightComplement(t_sub_interval)).b_Empty()) {
					p_new_arcs[n_arc_num] = t_complement;
					if(++ n_arc_num == max_Interval_Num)
						break;
				}
			}
			// subtract intervals
		}
		// subtract intervals

		n_interval_num = n_arc_num;
		memcpy(p_arc_angle, p_new_arcs, sizeof(TInterval) * n_arc_num);
		// copy intervals back (presumably cheaper than redistributing)

		if(f_angle1 <= 2 * f_pi)
			break;
		f_angle0 = 0;
		f_angle1 -= 2 * f_pi;
	}
	// subtract intervals, subdivide interval going over 2pi
}

/*
 *	float TDisc::f_FreeArc_Length() const
 *		- returns angular length of free arc (in radians)
 */
float TDisc::f_FreeArc_Length() const
{
	float f_result = 0;
	for(int i = 0, n = n_interval_num; i < n; ++ i) {
		if(!p_arc_angle[i].b_Empty())
			f_result += p_arc_angle[i].f_max - p_arc_angle[i].f_min;
	}
	return f_result;
}

/*
 *	Vector2f TDisc::v_Point(float f_angle)
 *		- returns point lying at free arc at position f_angle
 *		  (meassured on free arc segments only)
 *		- in case there is less than f_angle free arc, returns center point to mark error
 */
Vector2f TDisc::v_Point(float f_angle)
{
	for(int i = 0, n = n_interval_num; i < n; ++ i) {
		if(!p_arc_angle[i].b_Empty()) {
			float f_arc_len = p_arc_angle[i].f_max - p_arc_angle[i].f_min;
			if(f_angle < f_arc_len) {
				_ASSERTE(f_angle >= 0);
				f_angle += p_arc_angle[i].f_min;
				return v_center + Vector2f(cos(f_angle), sin(f_angle)) * f_radius;
			}
			f_angle -= f_arc_len;
		}
	}
	return v_center; // return center to mark error
}

/*
 *								=== ~TDisc ===
 */

/*
 *								=== CRepeatable_PDSet ===
 */

class CRepeatable_PDSet::CSumArcLengths {
protected:
	float m_f_sum;

public:
	inline CSumArcLengths()
		:m_f_sum(0)
	{}

	inline void operator ()(const TDisc &r_t_disc)
	{
		m_f_sum += r_t_disc.f_FreeArc_Length();
	}

	inline operator float() const
	{
		return m_f_sum;
	}
};

class CRepeatable_PDSet::CFindPointOnArc {
protected:
	float &m_r_f_arc_angle;

public:
	inline CFindPointOnArc(float &r_f_arc_angle)
		:m_r_f_arc_angle(r_f_arc_angle)
	{}

	inline bool operator ()(const TDisc &r_t_disc)
	{
		float f_arc_len = r_t_disc.f_FreeArc_Length();
		if(m_r_f_arc_angle < f_arc_len)
			return true;
		m_r_f_arc_angle -= f_arc_len;
		return false;
	}
};

/*
 *	CRepeatable_PDSet::CRepeatable_PDSet(float f_disc_radius, CMinRandomGenerator &r_generator)
 *		- default constructor
 *		- fills square 1 per 1 units with discs of radius f_disc_radius units
 *		  (square is in first quadrant and one of it's corners is identical with origin)
 *		- r_generator is pseudorandom number generator
 *		- call b_Status() to determine wheter constructor succeeded or not
 *		- note this is slow for big sets because there's no spatial subdivision
 *		  structure employed to accelerate
 */
CRepeatable_PDSet::CRepeatable_PDSet(float f_disc_radius, CMinRandomGenerator &r_generator)
{
	m_b_status = Create(f_disc_radius, r_generator);
}

/*
 *	bool CRepeatable_PDSet::b_Status() const
 *		- if constructor succeeded returns true, otherwise returns false
 */
bool CRepeatable_PDSet::b_Status() const
{
	return m_b_status;
}

/*
 *	bool CRepeatable_PDSet::Get(std::vector<TDisc> &r_disc_list) const
 *		- copies generated set to r_disc_list
 *		- returns true on success, false on failure
 */
bool CRepeatable_PDSet::Get(std::vector<TDisc> &r_disc_list) const
{
	r_disc_list.clear();
	r_disc_list.reserve(m_disc_list.size());
	if(r_disc_list.capacity() < m_disc_list.size())
		return false;
	r_disc_list.insert(r_disc_list.begin(), m_disc_list.begin(), m_disc_list.end());
	return true;
}

/*
 *	bool CRepeatable_PDSet::Create(float f_disc_radius, CMinRandomGenerator &r_generator)
 *		- fills square 1 per 1 units with discs of radius f_disc_radius units
 *		  (square is in first quadrant and one of it's corners is identical with origin)
 *		- discs are stored in m_disc_list
 *		- r_generator is pseudorandom number generator
 *		- returns true on success, false on failure
 *		- note this is slow for big sets because there's no spatial subdivision
 *		  structure employed to accelerate
 */
bool CRepeatable_PDSet::Create(float f_disc_radius, CMinRandomGenerator &r_generator)
{
	std::vector<TDisc> mirror_disc_list;
	// side-mirrored discs to assure repeatability of generated set

	for(bool b_first_pass = true;; b_first_pass = false) {
		Vector2f v_new_center;
		if(b_first_pass) {
			v_new_center = Vector2f(r_generator.f_Rand(), r_generator.f_Rand());
			// randomly generate position of the first disc
		} else {
			float f_free_arc_sum = std::for_each(m_disc_list.begin(),
				m_disc_list.end(), CSumArcLengths());
			if(f_free_arc_sum < 1e-3f)
				return true;
			// there's no free arc to be occupied

			float f_free_arc_pos = f_free_arc_sum * r_generator.f_Rand();
			// get angular position on free arc where new point should be generated

			std::vector<TDisc>::iterator p_disc_iter = std::find_if(m_disc_list.begin(),
				m_disc_list.end(), CFindPointOnArc(f_free_arc_pos));
			if(p_disc_iter == m_disc_list.end())
				return false;
			// find the disc, f_free_arc_pos is modified so it now points to position
			// on the disc

			_ASSERTE(f_free_arc_pos >= 0 && f_free_arc_pos <=
				(*p_disc_iter).f_FreeArc_Length());
			v_new_center = (*p_disc_iter).v_Point(f_free_arc_pos);
			// find a new point on the disc
		}

		TDisc t_new_disc(v_new_center, f_disc_radius);
		// new disc

		int n_mirror_disc_num = 0;
		for(int n_dy = -1; n_dy < 2; ++ n_dy) {
			float f_new_y = v_new_center.f_Y() + n_dy;
			if(f_new_y <= -f_disc_radius || f_new_y >= 1 + f_disc_radius)
				continue;
			for(int n_dx = -1; n_dx < 2; ++ n_dx) {
				if(!n_dx && !n_dy)
					continue;
				Vector2f v_center(v_new_center.f_X() + n_dx, f_new_y);
				if(v_center.f_X() > -f_disc_radius && v_center.f_X() < 1 + f_disc_radius) {
					if(mirror_disc_list.capacity() == mirror_disc_list.size()) {
						mirror_disc_list.reserve((mirror_disc_list.size())?
							mirror_disc_list.size() * 2 : 1);
						if(mirror_disc_list.capacity() == mirror_disc_list.size())
							return false;
					}
					mirror_disc_list.push_back(TDisc(v_center, f_disc_radius));
					++ n_mirror_disc_num;
				}
			}
		}
		// create several more discs to subtract from existing
		// discs to mimic wrap-arround behavior

		if(v_new_center.f_X() < f_disc_radius) {
			t_new_disc.Subtract(TDisc(Vector2f(-v_new_center.f_X(),
				v_new_center.f_Y()), f_disc_radius));
		}
		if(v_new_center.f_X() > 1 - f_disc_radius) {
			t_new_disc.Subtract(TDisc(Vector2f(2 - v_new_center.f_X(),
				v_new_center.f_Y()), f_disc_radius));
		}
		if(v_new_center.f_Y() < f_disc_radius) {
			t_new_disc.Subtract(TDisc(Vector2f(v_new_center.f_X(),
				-v_new_center.f_Y()), f_disc_radius));
		}
		if(v_new_center.f_Y() > 1 - f_disc_radius) {
			t_new_disc.Subtract(TDisc(Vector2f(v_new_center.f_X(),
				2 - v_new_center.f_Y()), f_disc_radius));
		}
		// subtract borders (do it better way?)

		for(int j = 0, n = mirror_disc_list.size(); j < n; ++ j)
			t_new_disc.Subtract(mirror_disc_list[j]);
		// subtract disc's mirrors from it

		for(int i = 0, n = m_disc_list.size(); i < n; ++ i) {
			for(int j = mirror_disc_list.size() - n_mirror_disc_num,
			   n = mirror_disc_list.size(); j < n; ++ j)
				m_disc_list[i].Subtract(mirror_disc_list[j]);
			m_disc_list[i].Subtract(t_new_disc);
			t_new_disc.Subtract(m_disc_list[i]);
		}
		// subtract discs from each other

		if(m_disc_list.capacity() == m_disc_list.size()) {
			m_disc_list.reserve((m_disc_list.size())? m_disc_list.size() * 2 : 1);
			if(m_disc_list.capacity() == m_disc_list.size())
				return false;
		}
		m_disc_list.push_back(t_new_disc);
		// add new disc
	}

	return false; // unreachable code, only to stop compiler from complaining
}

/*
 *								=== ~CRepeatable_PDSet ===
 */

/*
 *								=== CSamplePattern ===
 */

/*
 *	CSamplePattern::CSamplePattern(int n_point_num, int n_tolerance, CRandomGenerator &generator)
 *		- generates poisson disc sample pattern with n_point_num samples
 *		- n_tolerance is allowed tolerance for sample count (0 will do for counts below 10000)
 *		- generator is pseudo-random number generator (pre-set seed to get different results)
 *		- all samples lie inside square (0, 0), (1, 0), (1, 1), (0, 1)
 *		  and can be tiled along both axes
 *		- note in case constructor failed, number of generated points is zero
 */
CSamplePattern::CSamplePattern(int n_point_num, int n_tolerance, CRandomGenerator &generator)
{
	std::vector<TDisc> disc_list;
	if(GenerateDiscs(n_point_num, n_tolerance, generator, disc_list)) {
		m_sample_list.reserve(disc_list.size());
		if(m_sample_list.capacity() < disc_list.size())
			return;
		// make sure there's enough space for points

		//m_sample_list.insert(m_sample_list.begin(), disc_list.begin(), disc_list.end());
		// older compilers don't support that

		std::insert_iterator<std::vector<Vector2f> >
			p_insert_iter(m_sample_list, m_sample_list.begin());
		std::copy(disc_list.begin(), disc_list.end(), p_insert_iter);
		// this should do the trick
	}
}

/*
 *	void CSamplePattern::Center()
 *		- shifts center of bounding box of all samples to origin
 *		- note bounding box is used instead of averaged samples position because
 *		  samples are evenly distributed and so the difference should be minimal
 */
void CSamplePattern::Center()
{
	if(m_sample_list.empty())
		return;

	Vector2f v_min(m_sample_list[0]), v_max(m_sample_list[0]);
	for(int i = 0, n = m_sample_list.size(); i < n; ++ i) {
		for(int j = 0; j < 2; ++ j) {
			float f = m_sample_list[i][j];
			v_min[j] = (v_min[j] < f)? v_min[j] : f;
			v_max[j] = (v_max[j] > f)? v_max[j] : f;
		}
	}
	// determine bounding box

	Vector2f v_shift = (v_min + v_max) * -.5f;
	// determine shift

	for(int i = 0, n = m_sample_list.size(); i < n; ++ i)
		m_sample_list[i] += v_shift;
	// shift samples
}

/*
 *	void CSamplePattern::Scale(float f_scale)
 *		- scales all sample positions by factor f_scale
 */
void CSamplePattern::Scale(float f_scale)
{
	for(int i = 0, n = m_sample_list.size(); i < n; ++ i)
		m_sample_list[i] *= f_scale;
	// scale samples
}

/*
 *	bool CSamplePattern::GenerateDiscs(int n_point_num, int n_tolerance,
 *		CRandomGenerator &r_generator, std::vector<TDisc> &r_set)
 *		- generates n_point_num discs (with tolerance n_tolerance discs)
 *		  using pseudo-random number generator r_generator
 *		- r_set is filled with points
 *		- returns true on success, false on failure
 */
bool CSamplePattern::GenerateDiscs(int n_point_num, int n_tolerance,
	CRandomGenerator &r_generator, std::vector<TDisc> &r_set)
{
	float f_size = 1;
	for(float f_step = .5f; f_step > 1e-10f; f_step *= .5f) {
		r_generator.Reset();
		CRepeatable_PDSet pd_set(f_size, r_generator);
		if(!pd_set.b_Status() || !pd_set.Get(r_set))
			return false;
		int n_cur_point_num = r_set.size();
		// generate set of points

		if(abs(n_cur_point_num - n_point_num) <= n_tolerance)
			return true;
		else if(n_cur_point_num > n_point_num)
			f_size += f_step;
		else
			f_size -= f_step;
		// decide next action
	}
	// try to generate set with required number of points by altering disc radius

	r_generator.Reset();
	for(int i = 0; i < 100; ++ i) {
		CRepeatable_PDSet pd_set(f_size, r_generator);
		if(!pd_set.Get(r_set))
			return false;
		int n_cur_point_num = r_set.size();
		// generate set of points using the same generator
		// several times (effectively changing seed), try luck

		if(abs(n_cur_point_num - n_point_num) <= n_tolerance)
			return true;
	}
	// try luck in case it's hopeless (can't balance
	// number of points just by setting disc radius)

	return false;
}

/*
 *								=== ~CSamplePattern ===
 */
