/*
								+----------------------------------+
								|                                  |
								| ***  Poisson disc generator  *** |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|            Poisson.h             |
								|                                  |
								+----------------------------------+
*/

#ifndef __POISSON_GENERATOR_INCLUDED
#define __POISSON_GENERATOR_INCLUDED

extern const float f_pi;

/*
 *	struct TInterval
 *		- simple numeric interval structure
 */
struct TInterval {
	float f_min, f_max;

	/*
	 *	inline TInterval::TInterval(float _f_min = 0, float _f_max = 2 * f_pi)
	 *		- default constructor, parameters are interval end points
	 */
	inline TInterval(float _f_min = 0, float _f_max = 2 * f_pi)
		:f_min(_f_min), f_max(_f_max)
	{}

	/*
	 *	inline bool TInterval::b_Disjunct(const TInterval &r_interval) const
	 *		- returns true in case this interval and r_interval are disjunct, otherwise false
	 */
	inline bool b_Disjunct(const TInterval &r_interval) const
	{
		return f_max < r_interval.f_min || f_min > r_interval.f_max;
	}

	/*
	 *	inline bool TInterval::b_Empty() const
	 *		- returns true in case interval is empty, otherwise false
	 */
	inline bool b_Empty() const
	{
		return f_min >= f_max;
	}

	/*
	 *	inline TInterval TInterval::t_LeftComplement(const TInterval &r_interval) const
	 *		- returns left part of r_interval's complement in this interval (this - r_interval)
	 *		- note this returns empty interval if this interval is contained in r_interval
	 *		  or if r_interval.f_min < f_min
	 */
	inline TInterval t_LeftComplement(const TInterval &r_interval) const
	{
		return TInterval(f_min, (f_max < r_interval.f_min)? f_max : r_interval.f_min);
	}

	/*
	 *	inline TInterval TInterval::t_RightComplement(const TInterval &r_interval) const
	 *		- returns right part of r_interval's complement in this interval (this - r_interval)
	 *		- note this returns empty interval if this interval is contained in r_interval
	 *		  or if r_interval.f_max > f_max
	 */
	inline TInterval t_RightComplement(const TInterval &r_interval) const
	{
		return TInterval((f_min > r_interval.f_max)? f_min : r_interval.f_max, f_max);
	}
};

/*
 *	struct TDisc
 *		- disc structure; contains center position, radius and free boundary arc
 *		- area of other discs can be substracted from free boundary arc and that's
 *		  about the idea of the poisson distribution generator algorithm
 */
struct TDisc {
	enum {
		max_Interval_Num = 6 // there can be up to 6 discs on radius of another disc
	};

	TInterval p_arc_angle[max_Interval_Num];
	int n_interval_num;
	float f_radius; // radius
	Vector2f v_center; // center

	/*
	 *	inline TDisc()
	 *		- empty constructor, has no effect
	 */
	inline TDisc()
	{}

	/*
	 *	TDisc::TDisc(const Vector2f &r_v_center, float _f_radius)
	 *		- default constructor
	 *		- creates disc with center r_v_center and radius _f_radius
	 */
	TDisc(const Vector2f &r_v_center, float _f_radius);

	/*
	 *	static float TDisc::f_Angle(const Vector2f &r_v_vector)
	 *		- returns angle r_v_vector is pointing, returned angle is 0 to 2pi radians
	 *		- r_v_vector has to be unit-length
	 */
	static float f_Angle(const Vector2f &r_v_vector);

	/*
	 *	bool TDisc::Subtract(const TDisc &r_t_disc)
	 *		- subtracts r_t_disc disc's area from arc of this disc
	 *		- r_t_disc is the other disc
	 *		- operation is not commutative
	 *		- returns true on disc collision, false on no collision
	 */
	bool Subtract(const TDisc &r_t_disc);

	/*
	 *	void TDisc::SubtractAngleInterval(float f_angle0, float f_angle1)
	 *		- substracts angular interval f_angle0 to f_angle1 from disc's arc
	 *		- angles must be ordered so f_anle0 is less than f_angle1
	 *		- note the disc structure is able to describe just one continuous interval,
	 *		  in case more intervals originate by subtraction, the longest one is stored
	 */
	void SubtractAngleInterval(float f_angle0, float f_angle1);

	/*
	 *	float TDisc::f_FreeArc_Length() const
	 *		- returns angular length of free arc (in radians)
	 */
	float f_FreeArc_Length() const;

	/*
	 *	Vector2f TDisc::v_Point(float f_angle)
	 *		- returns point lying at free arc at position f_angle
	 *		  (meassured on free arc segments only)
	 *		- in case there is less than f_angle free arc, returns center point to mark error
	 */
	Vector2f v_Point(float f_angle);

	/*
	 *	inline operator TDisc::Vector2f() const
	 *		- returns disc center
	 */
	inline operator Vector2f() const
	{
		return v_center;
	}
};

/*
 *	class CRepeatable_PDSet
 *		- repeatable poisson-disc set
 */
class CRepeatable_PDSet {
protected:
	std::vector<TDisc> m_disc_list;
	bool m_b_status;

	class CSumArcLengths;
	class CFindPointOnArc;

public:
	/*
	 *	CRepeatable_PDSet::CRepeatable_PDSet(float f_disc_radius, CMinRandomGenerator &r_generator)
	 *		- default constructor
	 *		- fills square 1 per 1 units with discs of radius f_disc_radius units
	 *		  (square is in first quadrant and one of it's corners is identical with origin)
	 *		- r_generator is pseudorandom number generator
	 *		- call b_Status() to determine wheter constructor succeeded or not
	 *		- note this is slow for big sets because there's no spatial subdivision
	 *		  structure employed to accelerate
	 */
	CRepeatable_PDSet(float f_disc_radius, CMinRandomGenerator &r_generator);

	/*
	 *	bool CRepeatable_PDSet::b_Status() const
	 *		- if constructor succeeded returns true, otherwise returns false
	 */
	bool b_Status() const;

	/*
	 *	bool CRepeatable_PDSet::Get(std::vector<TDisc> &r_disc_list) const
	 *		- copies generated set to r_disc_list
	 *		- returns true on success, false on failure
	 */
	bool Get(std::vector<TDisc> &r_disc_list) const;

protected:
	bool Create(float f_disc_radius, CMinRandomGenerator &r_generator);
};

/*
 *	class CSamplePattern
 *		- poisson disc sample pattern for supersampling
 */
class CSamplePattern {
protected:
	std::vector<Vector2f> m_sample_list;

public:
	/*
	 *	CSamplePattern::CSamplePattern(int n_point_num,
	 *		int n_tolerance, CRandomGenerator &generator)
	 *		- generates poisson disc sample pattern with n_point_num samples
	 *		- n_tolerance is allowed tolerance for sample count (0 will do for counts below 10000)
	 *		- generator is pseudo-random number generator (pre-set seed to get different results)
	 *		- all samples lie inside square (0, 0), (1, 0), (1, 1), (0, 1)
	 *		  and can be tiled along both axes
	 *		- note in case constructor failed, number of generated points is zero
	 */
	CSamplePattern(int n_point_num, int n_tolerance, CRandomGenerator &generator);

	/*
	 *	void CSamplePattern::Center()
	 *		- shifts center of bounding box of all samples to origin
	 *		- note bounding box is used instead of averaged samples position because
	 *		  samples are evenly distributed and so the difference should be minimal
	 */
	void Center();

	/*
	 *	void CSamplePattern::Scale(float f_scale)
	 *		- scales all sample positions by factor f_scale
	 */
	void Scale(float f_scale);

	/*
	 *	inline int CSamplePattern::n_Sample_Num() const
	 *		- returns number of sample points
	 *		- note if constructor failed this will return zero
	 */
	inline int n_Sample_Num() const
	{
		return m_sample_list.size();
	}

	/*
	 *	inline const Vector2f &CSamplePattern::v_Sample(int n_index) const
	 *		- returns sample point with zero-based index n_index
	 */
	inline const Vector2f &v_Sample(int n_index) const
	{
		return m_sample_list[n_index];
	}

protected:
	bool GenerateDiscs(int n_point_num, int n_tolerance,
		CRandomGenerator &r_generator, std::vector<TDisc> &r_set);
};

#endif //__POISSON_GENERATOR_INCLUDED
