/*
								+----------------------------------+
								|                                  |
								|   ***   Skybox raytracer   ***   |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|          Raytracer.cpp           |
								|                                  |
								+----------------------------------+
*/

#ifdef WIN32
#include <crtdbg.h>
#else
#include <assert.h>
#define _ASSERTE(x) assert(x)
#endif
#include <math.h>
#include <vector>
#include <algorithm>
#include <functional>
#include "Raytracer.h"

/*
 *								=== CRayCaster ===
 */

unsigned int CRayCaster::m_n_ray_id = (unsigned int)-1;

/*
 *	CRayCaster::CRayCaster(CTerrainMesh *p_mesh, float f_far = 1e10f)
 *		- default constructor
 *		- f_far is maximal intersection distance
 *		- p_mesh is mesh being raytraced
 */
CRayCaster::CRayCaster(CTerrainMesh *p_mesh, float f_far)
	:m_p_triangle(0), m_p_mesh(p_mesh), m_f_t(f_far)
{
	if(!(++ m_n_ray_id))
		p_mesh->ResetEnvelopes(0);
	// reset envelopes if neccessary
}

/*
 *	bool CRayCaster::operator ()(Vector3f v_org, Vector3f v_dir, const int *p_faces)
 *		- performs raycasting; v_org and v_dir are ray origin and direction,
 *		  p_faces are indices of mesh faces that are inside the current sector
 *		- returns true on hit (to end raycasting), returns false on no hit (to continue)
 */
bool CRayCaster::operator ()(Vector3f v_org, Vector3f v_dir, const int *p_faces)
{
	TTriangle *p_tri_base = m_p_mesh->p_Triangle();
	for(const int *p_face = p_faces + 1,
	   *p_end = p_faces + 1 + *p_faces; p_face != p_end; ++ p_face) {
		TTriangle *p_candidate = p_tri_base + *p_face;

		if(p_candidate->n_envelope == m_n_ray_id)
			continue;
		p_candidate->n_envelope = m_n_ray_id;
		// avoid raytracing single face multiple times (saves about 25% time spent raytracing)

		float f_t, f_u, f_v;
		if(p_candidate->Cast_Ray(v_org, v_dir, f_t, f_u, f_v) && f_t > 0 && f_t < m_f_t) {
			m_f_t = f_t;
			m_f_u = f_u;
			m_f_v = f_v;
			m_p_triangle = p_candidate;
			return true; // low probability of glitches with terrain
		}
	}
	// go trough node's faces

	return false;
}

/*
 *	Vector3f CRayCaster::v_Normal() const
 *		- interpolates intersection normal from triangle vertex normals
 *		- this must not be called if no hit occured
 */
Vector3f CRayCaster::v_Normal() const
{
	_ASSERTE(m_p_triangle);
	Vector3f v_norm0 = m_p_triangle->p_vertex[0]->v_normal;
	Vector3f v_edge1 = m_p_triangle->p_vertex[1]->v_normal - v_norm0;
	Vector3f v_edge2 = m_p_triangle->p_vertex[2]->v_normal - v_norm0;
	Vector3f v_normal = v_norm0 + v_edge1 * m_f_u + v_edge2 * m_f_v;
	v_normal.Normalize();
	// although weights have sum 1, normal was most likely skewed
	return v_normal;
}

/*
 *								=== ~CRayCaster ===
 */

/*
 *								=== CAOSampling ===
 */

/*
 *	CAOSampling::CAOSampling(int n_min_size)
 *		- default constructor
 *		- creates a new ambient occlusion sampling pattern
 *		  with minimal size n_min_size samples
 */
CAOSampling::CAOSampling(int n_min_size)
{
	if(!CreateSampleVectors(n_min_size))
		m_vec_list.clear();
}

/*
 *	bool CAOSampling::CreateSampleVectors(int n_min_size)
 *		- creates at least n_min_size sample vectors
 *		- returns true on success, false on failure
 */
bool CAOSampling::CreateSampleVectors(int n_min_size)
{
	for(int n_sample_num = n_min_size;; ++ n_sample_num) {
        CStdLibGenerator rand_gen(123456);
		CSamplePattern sample_pattern(n_sample_num, 0, rand_gen);
		sample_pattern.Center();
		sample_pattern.Scale(2);
		if(sample_pattern.n_Sample_Num() != n_sample_num)
			return false;
		// create sample pattern

		m_vec_list.erase(m_vec_list.begin(), m_vec_list.end());
		m_vec_list.reserve(n_sample_num);
		if(m_vec_list.capacity() < (unsigned)n_sample_num)
			return false;
		float f_weight_sum = 0;
		for(int i = 0; i < n_sample_num; ++ i) {
			Vector2f v_sam = sample_pattern.v_Sample(i);
			if(v_sam.f_Length() >= 1)
				continue;
			// select only vectors inside unit circle

			float f_y = (float)sqrt(1 - v_sam.f_Length());
			// determine sample y (making sample distribution
			// a little bit side-centric, but nevermind)

			f_weight_sum += f_y;
			m_vec_list.push_back(TSampleVector(Vector3f(v_sam.f_X(), f_y, v_sam.f_Y()), f_y));
		}
		// discard samples outside unit circle, place remaining samples onto unit sphere

		if(m_vec_list.size() < (unsigned)n_min_size)
			continue;
		// not enough of 'em

		f_weight_sum = 1.0f / f_weight_sum;
		for(int i = 0, n = m_vec_list.size(); i < n; ++ i)
			m_vec_list[i].f_weight *= f_weight_sum;
		// normalize weights

		return true;
	}
}

/*
 *								=== ~CAOSampling ===
 */
