/*
								+----------------------------------+
								|                                  |
								|   ***   Skybox raytracer   ***   |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|           Raytracer.h            |
								|                                  |
								+----------------------------------+
*/

#ifndef __SKYBOX_RAYTRACER_INCLUDED
#define __SKYBOX_RAYTRACER_INCLUDED

#include "MiniVector.h"
#include "Terrain.h"
#include "TerrainMesh.h"
#include "RandGen.h"
#include "Poisson.h"
#include "PNoise.h"
#include "Task.h"
#include "Atmosphere.h"
#include "ImageCube.h"
#include "Color.h"

/*
 *	class CRayCaster
 *		- simple raycaster class
 *		- this processes callbacks from uniform grid to accelerate raytracing
 */
class CRayCaster {
protected:
	static unsigned int m_n_ray_id;
	const TTriangle *m_p_triangle;
	CTerrainMesh *m_p_mesh;
	float m_f_t, m_f_u, m_f_v;

public:
	/*
	 *	CRayCaster::CRayCaster(CTerrainMesh *p_mesh, float f_far = 1e10f)
	 *		- default constructor
	 *		- f_far is maximal intersection distance
	 *		- p_mesh is mesh being raytraced
	 */
	CRayCaster(CTerrainMesh *p_mesh, float f_far = 1e10f);

	/*
	 *	CRayCaster::CRayCaster(const CRayCaster &r_t_raytracer)
	 *		- copy-constructor
	 */
	inline CRayCaster(const CRayCaster &r_t_raytracer)
		:m_p_triangle(r_t_raytracer.m_p_triangle), m_p_mesh(r_t_raytracer.m_p_mesh),
		m_f_t(r_t_raytracer.m_f_t), m_f_u(r_t_raytracer.m_f_u), m_f_v(r_t_raytracer.m_f_v)
	{}

	/*
	 *	bool CRayCaster::operator ()(Vector3f v_org, Vector3f v_dir, const int *p_faces)
	 *		- performs raycasting; v_org and v_dir are ray origin and direction,
	 *		  p_faces are indices of mesh faces that are inside the current sector
	 *		- returns true on hit (to end raycasting), returns false on no hit (to continue)
	 */
	bool operator ()(Vector3f v_org, Vector3f v_dir, const int *p_faces);

	/*
	 *	inline bool CRayCaster::b_Hit() const
	 *		- returns true in case hit occured, otherwise returns false
	 */
	inline bool b_Hit() const
	{
		return m_p_triangle != 0;
	}

	/*
	 *	inline const TTriangle *CRayCaster::p_Triangle() const
	 *		- returns const pointer to triangle hit, or 0 if no triangle was hit
	 */
	inline const TTriangle *p_Triangle() const
	{
		return m_p_triangle;
	}

	/*
	 *	inline float CRayCaster::f_T() const
	 *		- returns distance of intersection along the ray or maximal
	 *		  intersection distance limit passed to constructor if no triangle was hit
	 */
	inline float f_T() const
	{
		return m_f_t;
	}

	/*
	 *	inline float CRayCaster::f_U() const
	 *		- return u (barycentric coordinate of triangle hit in it's barycentric space)
	 *		- note result is undefined in case no triangle was hit
	 */
	inline float f_U() const
	{
		return m_f_u;
	}

	/*
	 *	inline float CRayCaster::f_V() const
	 *		- return v (barycentric coordinate of triangle hit in it's barycentric space)
	 *		- note result is undefined in case no triangle was hit
	 */
	inline float f_V() const
	{
		return m_f_v;
	}

	/*
	 *	Vector3f CRayCaster::v_Normal() const
	 *		- interpolates intersection normal from triangle vertex normals
	 *		- this must not be called if no hit occured
	 */
	Vector3f v_Normal() const;
};

/*
 *	class CFogFunction
 *		- exponential, altitude-attenuated fog function (input to the integrator)
 */
class CFogFunction {
protected:
	Vector3f m_v_org, m_v_step;
	float m_f_fog_base, m_f_inv_fog_height, m_f_fog_thickness;

public:
	/*
	 *	inline CFogFunction::CFogFunction(Vector3f v_org, Vector3f v_step,
	 *		float f_fog_base, float f_fog_height, float f_fog_thickness)
	 *		- default constructor
	 *		- v_org and v_step defines line of sight fog is going to be integrated on
	 *		- f_fog_base is altitude where fog density is 1
	 *		- f_fog_height is difference of altitudes where densities are 1 and e^-1
	 *		- f_fog_thickness is multiplicative fog thickness
	 */
	inline CFogFunction(Vector3f v_org, Vector3f v_step,
		float f_fog_base, float f_fog_height, float f_fog_thickness)
		:m_v_org(v_org), m_v_step(v_step), m_f_fog_base(f_fog_base),
		m_f_inv_fog_height(1.0f / f_fog_height), m_f_fog_thickness(f_fog_thickness)
	{}

	/*
	 *	inline float CFogFunction::operator ()(float f_t) const
	 *		- returns fog density at f_t along the ray
	 */
	inline float operator ()(float f_t) const
	{
		float f_y = m_v_org.f_Y() + m_v_step.f_Y() * f_t;
		float f_fog = (float)exp(-(f_y - m_f_fog_base) * m_f_inv_fog_height) * m_f_fog_thickness;
		return (f_fog > 0)? f_fog : 0;
	}
};

/*
 *	class CAOSampling
 *		- ambient-occlusion hemi-spherical sampling pattern
 */
class CAOSampling {
public:
	/*
	 *	struct CAOSampling::TSampleVector
	 *		- single sample vector for ambient occlusion sampling
	 */
	struct TSampleVector {
		Vector3f v_direction;
		float f_weight;

		/*
		 *	inline CAOSampling::TSampleVector::TSampleVector(Vector3f v_vec, float f_w)
		 *		- default constructor; sets sample direction to v_vec and sample weight to f_w
		 */
		inline TSampleVector(Vector3f v_vec, float f_w)
			:v_direction(v_vec), f_weight(f_w)
		{}
	};

protected:
	std::vector<TSampleVector> m_vec_list;

public:
	/*
	 *	CAOSampling::CAOSampling(int n_min_size)
	 *		- default constructor
	 *		- creates a new ambient occlusion sampling pattern
	 *		  with minimal size n_min_size samples
	 */
	CAOSampling(int n_min_size);

	/*
	 *	inline int CAOSampling::n_Sample_Num() const
	 *		- returns number of samples
	 *		- returns 0 in case there was not enough memory for samples
	 */
	inline int n_Sample_Num() const
	{
		return m_vec_list.size();
	}

	/*
	 *	inline const CAOSampling::TSampleVector &CAOSampling::t_Sample(int n_index) const
	 *		- return sample with zero based index n_index
	 */
	inline const TSampleVector &t_Sample(int n_index) const
	{
		return m_vec_list[n_index];
	}

protected:
	bool CreateSampleVectors(int n_min_size);
};

/*
 *	class CSunIntensityTransfer
 *		- sun intensity transfer function
 *		- enables for processing sun photometric data
 */
class CSunIntensityTransfer {
protected:
	const CAtmosphere::CIntensity &m_r_atmosphere_func;
	CSunSpectrum m_sun_spectrum;

public:
	/*
	 *	inline CSunIntensityTransfer::CSunIntensityTransfer(
	 *		const CAtmosphere::CIntensity &r_atmosphere_func)
	 *		- default constructor
	 *		- r_atmosphere_func is atmosphere light transfer function
	 */
	inline CSunIntensityTransfer(const CAtmosphere::CIntensity &r_atmosphere_func)
		:m_r_atmosphere_func(r_atmosphere_func)
	{}

	/*
	 *	inline double CSunIntensityTransfer::operator ()(float f_lambda) const
	 *		- returns intensity of light with wavelength f_lambda (in nanometers)
	 *		  reaching the observer (in W.m^-2.nm^-1)
	 */
	inline double operator ()(float f_lambda) const
	{
		return (double)m_sun_spectrum(f_lambda) * (double)m_r_atmosphere_func(f_lambda);
	}
};

/*
 *	struct TBasis
 *		- simple structure for basis definition for simple cubemap raytracing
 *		  (for each cubemap side there's basis allowing to transform from
 *		  raster coordinates to ray coordinates)
 */
struct TBasis {
	Vector3f v_s_step;
	Vector3f v_t_step;
	Vector3f v_org;

	/*
	 *	inline TBasis::TBasis(float f_step, Vector3f _v_s_dir, Vector3f _v_t_dir, Vector3f _v_org)
	 *		- default constructor
	 *		- _v_s_dir is direction of the s-coordinate axis (direction of box side edge)
	 *		- _v_t_dir is direction of the t-coordinate axis (direction of the other box side edge)
	 *		- _v_org is basis origin (position of corner of the box side)
	 */
	inline TBasis(float f_step, Vector3f _v_s_dir, Vector3f _v_t_dir, Vector3f _v_org)
		:v_s_step(_v_s_dir * f_step), v_t_step(_v_t_dir * f_step),
		v_org(_v_org + (_v_s_dir + _v_t_dir) * f_step * .5f)
	{}
};

/*
 *	class CSkyBoxRaytracer
 *		- skybox raytracing encapsulation class
 */
class CSkyBoxRaytracer {
public:
	/*
	 *	static bool CSkyBoxRaytracer::Raytrace_Skybox(const CTaskDescription &r_task,
	 *		const CHeightField &r_heightmap, CCubeSampler<Vector3f> &r_output_color,
	 *		bool b_verbose)
	 *		- this function converts r_heightmap to triangle mesh and raytraces it
	 *		  according to what r_task wants
	 *		- r_output_color is output and must be allocated to proper size, specified
	 *		  by r_task.r_Output().n_texture_size
	 *		- b_verbose enables verbose output about raytracing progress
	 *		- returns true on success, false on failure
	 */
	static bool Raytrace_Skybox(const CTaskDescription &r_task, const CHeightField &r_heightmap,
		CCubeSampler<Vector3f> &r_output_color, bool b_verbose)
	{
		if(b_verbose)
			printf("converting to mesh ...\n");
		CTerrainMesh *p_mesh;
		if(!(p_mesh = new CTerrainMesh(r_heightmap, r_task.r_Terrain().f_tile_size)) ||
		   !p_mesh->b_Status()) {
			if(p_mesh)
				delete p_mesh;
			fprintf(stderr, "error: error creating heightmap\n");
			return false;
		}
		// create terrain mesh

		if(b_verbose)
			printf("spatial subdivision ...\n");
		CUniformGrid *p_grid;
		if(!(p_grid = new CUniformGrid(p_mesh->n_Triangle_Num() * 10,
		   p_mesh->n_Triangle_Num(), p_mesh->p_Triangle())) || !p_grid->b_Status()) {
			if(p_grid)
				delete p_grid;
			delete p_mesh;
			return false;
		}
		// create uniform grid for raytracing acceleration

		int n_sky_resolution = (int)r_task.r_SkyColor().n_texture_size;
		CCubeSampler<Vector3f> sky_color(n_sky_resolution);
		if(!sky_color.b_Status() || // alloc sky-color cubemap
		   !Integrate_SkyColor(sky_color, r_task, b_verbose) || // create sky cubemap
		   !Raytrace_Terrain(r_output_color, r_task, sky_color, p_mesh, p_grid, b_verbose)) {
			delete p_grid;
			delete p_mesh;
			return false;
		}
		// raytrace terrain

		delete p_grid;
		delete p_mesh;
		// don't need those anymore

		return true;
	}

protected:
	static Vector3f v_Raytrace(Vector3f v_ray, const CTaskDescription &r_task/*, CAtmosphere &r_atm*/,
		const CCubeSampler<Vector3f> &r_sky_color, const CAOSampling &r_ao_samplers,
		CTerrainMesh *p_mesh, const CUniformGrid *p_grid)
	{
		v_ray.Normalize();
		// normalize ray

		CRayCaster intersection(p_grid->Traverse(Vector3f(0, 0, 0), v_ray, CRayCaster(p_mesh)));
		if(intersection.b_Hit()) {
			//const TTriangle *p_tri = intersection.p_Triangle();
			Vector3f v_normal = intersection.v_Normal();
			// lerp normal

			Vector3f v_pos(v_ray * intersection.f_T());
			// calculate pos

			const CTaskDescription::TSkyDescription &r_sky = r_task.r_Atmosphere();

			Vector3f v_color(0, 0, 0);
			{
				float f_sun_visibility = 1;
				Vector3f v_fakepos(v_pos);
				Vector3f v_offset = v_normal.v_Orthogonal(r_sky.v_sun_direction);
				v_offset.Normalize(r_task.r_Output().f_shadow_sample_offset);
				float f_sample_weight = 1.0f / r_task.r_Output().n_shadow_sample_num;
				for(int i = 0, n = r_task.r_Output().n_shadow_sample_num; i < n; ++ i) {
					if(p_grid->Traverse(v_fakepos, r_sky.v_sun_direction, CRayCaster(p_mesh)).b_Hit())
						f_sun_visibility -= f_sample_weight;
					v_fakepos += v_offset;
				}
				// soft shadows attempt

				if(f_sun_visibility) {
					float f_lambertian = v_normal.f_Dot(r_sky.v_sun_direction);
					if(f_lambertian > 0) {
						f_lambertian *= f_sun_visibility * r_task.r_Terrain().f_diffuse;
						v_color = Vector3f(f_lambertian, f_lambertian, f_lambertian);
					}
				}
				// account for direct sunlight
			}
			// calculate diffuse color

			{
				Vector3f v_tangent(v_normal.v_Orthogonal(Vector3f(0, 1, 0))); // will do on heightmap
				v_tangent.Normalize();
				Vector3f v_bitangent(v_tangent.v_Cross(v_normal));
				// form basis vectors

				Vector3f v_ambient(0, 0, 0);
				for(int i = 0, n = r_ao_samplers.n_Sample_Num(); i < n; ++ i) {
					Vector3f v_dir = r_ao_samplers.t_Sample(i).v_direction;
					v_dir = v_tangent * v_dir.f_X() + v_normal * v_dir.f_Y() +
						v_bitangent * v_dir.f_Z();
					// transform sampler direction to worldspace

					if(p_grid->Traverse(v_pos + v_normal, v_dir, CRayCaster(p_mesh)).b_Hit())
						continue;
					// shadowed

					v_color += r_sky_color.f_Sample(v_dir.f_X(), v_dir.f_Y(), v_dir.f_Z()) *
						r_ao_samplers.t_Sample(i).f_weight;
					// otherwise fetch ambient color there
				}
				// calculate ambient occlusion

				v_color += v_ambient * r_task.r_Terrain().f_ambient;
			}
			// calculate ambient color

			v_color *= r_task.r_Terrain().v_Color(v_pos, v_normal);
			// fetch texture color

			{
				const CTaskDescription::TTerrainDescription &r_terrain = r_task.r_Terrain();
				float f_length = v_pos.f_Length();
				float f_fog_coeff = f_Integrate_s(CFogFunction(Vector3f(0, 0, 0), v_pos / f_length,
					r_terrain.f_fog_base, r_terrain.f_fog_height, r_terrain.f_fog_thickness),
					0, f_length, 1e-3f);
				// integrate fog thickness

				//const CTaskDescription::THorizonSampling &r_horizon = r_task.r_HorizonSampling();
				Vector3f v_to_horizon(v_ray);
				float f_xz_distance = (float)sqrt(v_to_horizon.f_X() * v_to_horizon.f_X() +
					v_to_horizon.f_Z() * v_to_horizon.f_Z());
				float f_min_lean = r_task.r_HorizonSampling().f_min_y * f_xz_distance;
				float f_max_lean = r_task.r_HorizonSampling().f_max_y * f_xz_distance;
				if(v_to_horizon.f_Y() > f_max_lean)
					v_to_horizon.f_Y() = f_max_lean;
				else if(v_to_horizon.f_Y() < f_min_lean)
					v_to_horizon.f_Y() = f_min_lean;
				Vector3f v_horizon_color = r_sky_color.f_Sample(v_to_horizon.f_X(),
					v_to_horizon.f_Y(), v_to_horizon.f_Z());
				// sample horizon color

				if(f_fog_coeff < 0)
					f_fog_coeff = 0;
				else if(f_fog_coeff > 1)
					f_fog_coeff = 1;
				v_color += (v_horizon_color - v_color) * f_fog_coeff;
				// apply fog
			}
			// calculate fog (no shadows in fog, though)

			return v_color;
		} else
			return r_sky_color.f_Sample(v_ray.f_X(), v_ray.f_Y(), v_ray.f_Z());
		// raytrace using uniform grid
	}

	static bool Integrate_SkyColor(CCubeSampler<Vector3f> &r_sky_color,
		const CTaskDescription &r_task, bool b_verbose)
	{
		CColorSystem color_system(CColorSystem::Make_System(r_task.r_Output().n_color_system_name));
		// create color system for precise color rendering

		const CTaskDescription::TSkyDescription &r_sky = r_task.r_Atmosphere();
		CAtmosphere atmosphere(r_sky.f_planet_radius, r_sky.f_atmosphere_radius,
			r_sky.v_sun_direction, r_sky.f_u, r_sky.f_k_mie, r_sky.f_k_raileigh);
		// create atmosphere sim

		int n_resolution = (int)r_task.r_SkyColor().n_texture_size;
		_ASSERTE(r_sky_color.n_Size() == n_resolution);
		float f_step = 1.0f / n_resolution;
		const TBasis p_basis_list[6] = {
			TBasis(f_step, Vector3f(0, 0, 2), Vector3f(0, -2, 0), Vector3f(-1, 1, -1)),
			TBasis(f_step, Vector3f(0, 0, -2), Vector3f(0, -2, 0), Vector3f(1, 1, 1)),
			TBasis(f_step, Vector3f(2, 0, 0), Vector3f(0, 0, -2), Vector3f(-1, -1, 1)),
			TBasis(f_step, Vector3f(2, 0, 0), Vector3f(0, 0, 2), Vector3f(-1, 1, -1)),
			TBasis(f_step, Vector3f(-2, 0, 0), Vector3f(0, -2, 0), Vector3f(1, 1, -1)),
			TBasis(f_step, Vector3f(2, 0, 0), Vector3f(0, -2, 0), Vector3f(-1, 1, 1))
		};
		// calculate six cubeface bases

		float f_min = 1e10f, f_max = -1e10f;
		//float f_sky_texture_scale = (float)n_resolution / n_resolution;
		Vector3f v_observer(0, r_sky.f_planet_radius + r_sky.f_observer_altitude, 0);
		for(int i = 0, n_scanline_num = 6 * n_resolution; i < 6; ++ i) {
			Vector3f v_x_step = p_basis_list[i].v_s_step;
			Vector3f v_y_step = p_basis_list[i].v_t_step;
			Vector3f v_tar_y(p_basis_list[i].v_org);
			CImage2D<Vector3f> &r_cube_face = r_sky_color.r_Image(i);
			for(int y = 0; y < n_resolution; ++ y, v_tar_y += v_y_step) {
				Vector3f v_tar(v_tar_y);
				for(int x = 0; x < n_resolution; ++ x, v_tar += v_x_step) {
					Vector3f v_ray(v_tar);
					v_ray.Normalize();
					// ray

					if(v_ray.f_Y() > r_task.r_SkyColor().f_y_cutoff) {
						float f_t = 0;
						atmosphere.IntersectSphere(v_observer, -v_ray, r_sky.f_atmosphere_radius, f_t);
						Vector3f v_far(v_observer - v_ray * f_t);
						// calculate ray intersection with far side of atmosphere

						CAtmosphere::CIntensity intensity_func(atmosphere.Intensity(v_observer,
							v_far));
						if(!intensity_func.b_Status())
							return false;
						// create atmosphere light intensity transfer function

						/*float f_ri = atmosphere.f_Intensity(v_observer, v_far_side, f_r_wl);
						float f_gi = atmosphere.f_Intensity(v_observer, v_far_side, f_g_wl);
						float f_bi = atmosphere.f_Intensity(v_observer, v_far_side, f_b_wl);
						if(fabs((intensity_func(f_r_wl) - f_ri) / f_ri) > 1e-6 ||
						   fabs((intensity_func(f_g_wl) - f_gi) / f_gi) > 1e-6 ||
						   fabs((intensity_func(f_b_wl) - f_bi) / f_bi) > 1e-6)
							fprintf(stderr, "error: difference in intensity functions\n");*/
						// observed errors are below 1e-6th scale of magnitude (relative error)
						// and are considered merely different calculus errors

						double f_power;
						Vector3f v_xyz = CSpectrumToCIE::v_CIE(CSunIntensityTransfer(intensity_func),
							f_power);
						Vector3f v_color = color_system.v_GammaCorrect(
							color_system.v_Constrain(color_system.v_MapColor(v_xyz))) * f_power;
						// calculate sky color using precise sun radiation spectra

						r_cube_face.Set(x, y, v_color);
						// save color

						for(int n = 0; n < 3; ++ n) {
							if(f_min > v_color[n])
								f_min = v_color[n];
							if(f_max < v_color[n])
								f_max = v_color[n];
						}
						// minmax
					} else
						r_cube_face.Set(x, y, Vector3f(0, 0, 0)); // don't minmax this!
				}

				if(b_verbose) {
					int n_prev_progress = ((y + i * n_resolution) * 40) / n_scanline_num;
					int n_progress = ((y + 1 + i * n_resolution) * 40) / n_scanline_num;
					if(n_prev_progress != n_progress) {
						printf("\rintegrating skies %c%c [%-40s]", 'x' + i / 2, '-' - (2 * (i % 2)),
							"########################################" + (40 - n_progress));
					}
				}
				// draw progress-bar
			}
		}

		if(f_min < f_max) {
			float f_scale = 1.0f / (f_max - f_min);
			for(int i = 0; i < 6; ++ i) {
				CImage2D<Vector3f> &r_cube_face = r_sky_color.r_Image(i);
				for(int y = 0; y < n_resolution; ++ y) {
					for(int x = 0; x < n_resolution; ++ x)
						r_cube_face.Set(x, y, r_cube_face.f_Get(x, y) * f_scale);
				}
			}
		}
		// rescale sky colors to [x, 1] range (for better precision of following ops,
		// colors are later renormalized anyway)

		if(b_verbose)
			printf("\n");
		// progress-bar

		return true;
	}

	static bool Raytrace_Terrain(CCubeSampler<Vector3f> &r_output_color,
		const CTaskDescription &r_task, const CCubeSampler<Vector3f> &r_sky_color,
		CTerrainMesh *p_mesh, const CUniformGrid *p_grid, bool b_verbose)
	{
        CStdLibGenerator randgen(123456);
		CSamplePattern sample_pattern((int)r_task.r_Output().n_sample_num, 0, randgen);
		sample_pattern.Center();
		if(sample_pattern.n_Sample_Num() != r_task.r_Output().n_sample_num)
			return false;
		// create sampling pattern

		CAOSampling ao_samplers(r_task.r_Output().n_ao_sample_num);
		if(!ao_samplers.n_Sample_Num())
			return false;
		// create ambient occlusion sampling pattern

		const CTaskDescription::TSkyDescription &r_sky = r_task.r_Atmosphere();
		CAtmosphere atmosphere(r_sky.f_planet_radius, r_sky.f_atmosphere_radius,
			r_sky.v_sun_direction, r_sky.f_u, r_sky.f_k_mie, r_sky.f_k_raileigh);
		// create atmosphere sim

		int n_resolution = r_task.r_Output().n_texture_size;
		_ASSERTE(r_output_color.n_Size() == n_resolution);
		float f_step = 1.0f / n_resolution;
		const TBasis p_basis_list[6] = {
			TBasis(f_step, Vector3f(0, 0, 2), Vector3f(0, -2, 0), Vector3f(-1, 1, -1)),
			TBasis(f_step, Vector3f(0, 0, -2), Vector3f(0, -2, 0), Vector3f(1, 1, 1)),
			TBasis(f_step, Vector3f(2, 0, 0), Vector3f(0, 0, -2), Vector3f(-1, -1, 1)),
			TBasis(f_step, Vector3f(2, 0, 0), Vector3f(0, 0, 2), Vector3f(-1, 1, -1)),
			TBasis(f_step, Vector3f(-2, 0, 0), Vector3f(0, -2, 0), Vector3f(1, 1, -1)),
			TBasis(f_step, Vector3f(2, 0, 0), Vector3f(0, -2, 0), Vector3f(-1, 1, 1))
		};
		// calculate six cubeface bases

		float f_min = 1e10f, f_max = -1e10f;
		for(int i = 0, n_sample_num = r_task.r_Output().n_sample_num,
		   n_scanline_num = 6 * n_resolution; i < 6; ++ i) {
			Vector3f v_x_step = p_basis_list[i].v_s_step;
			Vector3f v_y_step = p_basis_list[i].v_t_step;
			Vector3f v_tar_y(p_basis_list[i].v_org);
			CImage2D<Vector3f> &r_cube_face = r_output_color.r_Image(i);
			for(int y = 0; y < n_resolution; ++ y, v_tar_y += v_y_step) {
				Vector3f v_tar(v_tar_y);
				for(int x = 0; x < n_resolution; ++ x, v_tar += v_x_step) {
					Vector3f v_color(0, 0, 0);
					for(int n = 0; n < n_sample_num; ++ n) {
						v_color += v_Raytrace(v_tar + v_x_step * sample_pattern.v_Sample(n).f_X() +
							v_y_step * sample_pattern.v_Sample(n).f_Y(),
							r_task, /*atmosphere,*/ r_sky_color, ao_samplers, p_mesh, p_grid);
					}
					v_color /= sample_pattern.n_Sample_Num();
					// multisample raytrace pixel (might do it several times and jitter)

					r_cube_face.Set(x, y, v_color);
					// store color inside cubemap

					for(int n = 0; n < 3; ++ n) {
						if(f_min > v_color[n])
							f_min = v_color[n];
						if(f_max < v_color[n])
							f_max = v_color[n];
					}
					// minmax
				}

				if(b_verbose) {
					int n_prev_progress = ((y + i * n_resolution) * 40) / n_scanline_num;
					int n_progress = ((y + 1 + i * n_resolution) * 40) / n_scanline_num;
					if(n_prev_progress != n_progress) {
						printf("\rraytracing ground %c%c [%-40s]", 'x' + i / 2, '-' - (2 * (i % 2)),
							"########################################" + (40 - n_progress));
					}
				}
				// draw progress-bar
			}
			// raytrace this cubemap face
		}
		// terrain raytracing loop

		if(b_verbose)
			printf("\n");
		// progress-bar

		if(r_task.r_Output().b_normalize || r_task.r_Output().f_gamma != 1 ||
		   r_task.r_Output().f_scale != 1 || r_task.r_Output().f_bias != 0) {
			//bool b_normalize = (f_max != f_min) && r_task.r_Output().b_normalize;
			float f_scale = 1.0f / (f_max - f_min);
			float f_gamma = r_task.r_Output().f_gamma;
			float f_extra_scale = r_task.r_Output().f_scale;
			float f_extra_bias = r_task.r_Output().f_bias;
			Vector3f v_sub(f_min, f_min, f_min);
			for(int i = 0; i < 6; ++ i) {
				CImage2D<Vector3f> &r_cube_face = r_output_color.r_Image(i);
				for(int y = 0; y < n_resolution; ++ y) {
					for(int x = 0; x < n_resolution; ++ x) {
						Vector3f v_color = r_cube_face.f_Get(x, y);
						// get raw output color

						if(r_task.r_Output().b_normalize) {
							v_color -= v_sub;
							v_color *= f_scale;
						}
						// normalize to [0, 1] range (enhance contrast)

						for(int n = 0; n < 3; ++ n)
							v_color[n] = pow(v_color[n], f_gamma) * f_extra_scale + f_extra_bias;
						// final color processing

						r_cube_face.Set(x, y, v_color);
					}
				}
			}
		}
		// final skybox color processing

		return true;
	}
};

#endif //__SKYBOX_RAYTRACER_INCLUDED
