/*
								+----------------------------------+
								|                                  |
								|   ***   Lexical analyzer   ***   |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|            Scanner.h             |
								|                                  |
								+----------------------------------+
*/

/*
 *	passed code revision at 2007-09-24
 *
 *	class CScanner made template class with parameter of character type to easily support
 *	simple 7-bit us-ascii parsing with stdlib functions
 *	fixed minor error in CScanner::GrowBuffer() in overflow check
 *	structures TTransition and TState were moved from CScanner to TScannerDrivingTable
 *	renamed CScanner::GetToken_T() to CScanner::GetToken()
 *	CScanner::CTokenEmit was made CScanner template parameter so it's possible
 *	to use functions instead of virtual function objects
 *	TScannerDrivingTable::TState and TScannerDrivingTable::TTransition were made CScanner
 *	template parameters so it's possible to use different versions of driving tables
 *	(like small us-ascii table instead of "default" unicode table)
 *
 */

#ifndef __SCANNER_INCLUDED
#define __SCANNER_INCLUDED

/*
 *	class CTokenEmit
 *		- example implementation of token emit class
 */
/*

class CTokenEmit {
public:
	/+
	 *	bool CTokenEmit::operator ()(const TCharType *p_s_buffer,
	 *		int n_regexp_id, int n_line, int n_column)
	 *		- called by scanner upon reading a token
	 *		- p_s_buffer contains string (TCharType because of unicode) with
	 *		  token literal, n_regexp_id is id of regular expression rule,
	 *		  n_line and n_column are position of token in input data
	 +/
	bool operator ()(const TCharType *p_s_buffer,
		int n_regexp_id, int n_line, int n_column);
};

*/

template <class TCharType, class TStateType, class TTransitionType, class CTokenEmit>
class CScanner {
public:
	/*
	 *	class CScanner::CStreamReader
	 *		- virtual stream reader delegate class
	 */
	class CStreamReader {
	public:
		/*
		 *	virtual bool CScanner::CStreamReader::ReadChar(int &r_n_char)
		 *		- read character and put it to r_n_char
		 *		- returns true on success or false on EOF
		 */
		virtual bool ReadChar(int &r_n_char) = 0;

		/*
		 *	virtual bool CScanner::CStreamReader::UnreadChar(int n_char)
		 *		- used for returning last read char back to be read next time
		 *		- it's guaranteed this won't be called before calling ReadChar()
		 *		  at least once
		 *		- this function is called maximally once until calling
		 *		  ReadChar again, n_char is equal to last char returned by ReadChar()
		 */
		virtual bool UnreadChar(int n_char) = 0;
	};

	/*
	 *	class CStringReader
	 *		- example implementation of simple CStreamReader intended
	 *		  to read null-terminated strings
	 */
	/*

	class CStringReader : public CStreamReader {
	protected:
		const char *m_p_s_string;

	public:
		CStringReader(const char *p_s_string)
			:m_p_s_string(p_s_string)
		{}

		virtual bool ReadChar(int &r_n_char)
		{
			if(*m_p_s_string) {
				r_n_char = *m_p_s_string ++;
				return true;
			}
			return false;
		}

		virtual bool UnreadChar(int n_char)
		{
			-- m_p_s_string;
			return true;
		}
	};

	*/

protected:
	const TStateType *m_p_state;
	CTokenEmit *m_p_token_emit;
	int m_n_token_emit_num;
	int m_n_cur_line, m_n_cur_column;
	TCharType *m_p_buffer, *m_p_buffer_ptr, *m_p_buffer_end;

public:
	/*
	 *	CScanner::CScanner(const TStateType *p_state,
	 *		CTokenEmit **p_token_emit, int n_token_emit_num)
	 *		- default constructor
	 *		- creates new instance of CScanner, accepting regular language given by
	 *		  deterministic finite automata specified as set of states, connected by
	 *		  transitions. accepting states have n_regexp_id set to index in CTokenEmit
	 *		  object array (out-of-bounds indices are not accepting)
	 *		- p_state contains finite state machine states
	 *		- p_token_emit contains n_token_emit_num CTokenEmit objects
	 *		- all arrays are only referenced, they have to remain allocated while
	 *		  using scanner
	 */
	CScanner(const TStateType *p_state,
		CTokenEmit *p_token_emit, int n_token_emit_num)
		:m_p_state(p_state),
		m_p_token_emit(p_token_emit), m_n_token_emit_num(n_token_emit_num),
		m_n_cur_line(0), m_n_cur_column(0),
		m_p_buffer(0), m_p_buffer_ptr(0), m_p_buffer_end(0)
	{}

	/*
	 *	bool CScanner::GetToken_V(CStreamReader &r_input)
	 *		- function for generating token from input stream
	 *		- r_input is object of class derived from CStreamReader
	 *		- returns true on success, false on failure (lexical error, not enough memory)
	 *		- note generated token must be returned by CTokenEmit object called by this function
	 *		  (in case of empty file function returns true but no CTokenEmit is called)
	 *		- use in case aim is minimal code size and there's more than one type of CStreamReader
	 *		  so GetToken specializes once only (but there's virtual function overhead)
	 */
	inline bool GetToken_V(CStreamReader &r_input)
	{
		return GetToken(r_input);
	}

	/*
	 *	template <class CReaderClass>
	 *	bool CScanner::GetToken(CReaderClass &r_input)
	 *		- template function for generating token from input stream
	 *		- r_input is function object, featuring bool ReadChar(int &r_n_char) function
	 *		  which returns true on success and false on EOF, character is put to r_n_char and
	 *		  bool UnreadChar(int n_char) function which is used for returning last read char
	 *		  back to be read next time. this function is called maximally once until calling
	 *		  ReadChar again. return values true success, false failure (reserved for possible
	 *		  i/o errors, etc)
	 *		- returns true on success, false on failure (lexical error, not enough memory)
	 *		- note generated token must be returned by CTokenEmit object called by this function
	 *		  (in case of empty file function returns true but no CTokenEmit is called)
	 *		- use in case aim is maximal speed (but in case you use more CReaderClass types,
	 *		  there will be more instances of this function meaning larger code)
	 */
	template <class CReaderClass>
	bool GetToken(CReaderClass &r_input)
	{
		int n_line = m_n_cur_line;
		int n_column = m_n_cur_column;
		// here lies new token

		m_p_buffer_ptr = m_p_buffer;
		// "clean" input buffer

		const TStateType *p_state = m_p_state;
		do {
			int n_char;
			if(!r_input.ReadChar(n_char)) {
				if(m_p_buffer_ptr == m_p_buffer)
					return true;
				// empty file

				break;
				// EOF in the middle of something
			}
			// get char

			const TStateType *p_next_state = 0;
			for(const TTransitionType *p_transition = p_state->p_transition,
			   *p_end = p_state->p_transition + p_state->n_transition_num;
			   p_transition != p_end; ++ p_transition) {
				if(n_char >= p_transition->n_char_min && n_char <= p_transition->n_char_max) {
					p_next_state = m_p_state + p_transition->n_state;
					break;
				}
			}
			// find next state

			if(!p_next_state) {
				if(!r_input.UnreadChar(n_char))
					return false;
				// this character is not to be processed

				break;
			}
			p_state = p_next_state;

			if(m_p_buffer_ptr == m_p_buffer_end && !GrowBuffer())
				return false;
			*m_p_buffer_ptr ++ = (TCharType)n_char;
			// accumulate character

			if(n_char == '\n') {
				m_n_cur_column = 0;
				++ m_n_cur_line;
			} else {
				if(n_char == '\t')
					m_n_cur_column = (m_n_cur_column + 4) & ~3;
				else
					++ m_n_cur_column;
			}
			// count lines / columns
		} while(p_state->n_transition_num);
		// no point in reading more characters when there are no transitions from current state

		if(p_state->n_regexp_id < 0 || p_state->n_regexp_id >= m_n_token_emit_num)
			return false;
		// not accepting state - error (unexpected character / unexpected end of file)

		if(m_p_buffer_ptr == m_p_buffer_end && !GrowBuffer())
			return false;
		*m_p_buffer_ptr ++ = 0;
		// add terminating zero

		return (*m_p_token_emit[p_state->n_regexp_id])(m_p_buffer,
			p_state->n_regexp_id, n_line, n_column);
		// emit token
	}

	/*
	 *	inline int CScanner::n_Cur_Line() const
	 *		- returns current line (zero-based index)
	 */
	inline int n_Cur_Line() const
	{
		return m_n_cur_line;
	}

	/*
	 *	inline int CScanner::n_Cur_Column() const
	 *		- returns current column (zero-based index)
	 */
	inline int n_Cur_Column() const
	{
		return m_n_cur_column;
	}

protected:
	bool GrowBuffer()
	{
		unsigned long n_size = m_p_buffer_end - m_p_buffer;
		if(!n_size)
			n_size = 64;
		else {
			if(n_size >= (unsigned long)-1 / 2)
				return false;
			// n_size *= 2 would cause owerflow

			n_size *= 2;
		}
		// determine new size

		TCharType *p_new_buffer;
		if(!(p_new_buffer = new TCharType[n_size]))
			return false;
		// alloc new buffer

		if(m_p_buffer) {
			memcpy(p_new_buffer, m_p_buffer, (m_p_buffer_ptr - m_p_buffer) * sizeof(TCharType));
			delete[] m_p_buffer;
		}
		// copy old data

		m_p_buffer_ptr = p_new_buffer + (m_p_buffer_ptr - m_p_buffer);
		m_p_buffer = p_new_buffer;
		m_p_buffer_end = p_new_buffer + n_size;
		// update pointers

		return true;
	}
};

#endif //__SCANNER_INCLUDED
