/*
								+----------------------------------+
								|                                  |
								|    ***  Task description  ***    |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|             Task.cpp             |
								|                                  |
								+----------------------------------+
*/

#if defined(_MSC_VER) && !defined(__MWERKS__)
#pragma warning(disable:4786)
#endif
// disable warning C4786 "identifier was truncated to '255' characters in the debug information"

#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#include <crtdbg.h>
#else
#include <assert.h>
#define _ASSERTE(x) assert(x)
#endif
//#include <GL/gl.h>
#include <stdio.h>
#include <math.h>
#include <vector>
#include <algorithm>
#include "PNoise.h"
#include "MiniVector.h"
#include "Scanner.h"
#include "Material.h"
#include "Terrain.h"
#include "Color.h"
#include "Task.h"

/*
 *								=== TGaussDistribution ===
 */

/*
 *	TGaussDistribution::TGaussDistribution()
 *		- default constructor; has no effect
 */
TGaussDistribution::TGaussDistribution()
{}

/*
 *	TGaussDistribution::TGaussDistribution(float _f_bias, float _f_multiplier,
 *		float _f_center, float _f_sigma2)
 *		- constructor
 *		- initializes gauss distribution structure
 *		- f_bias is bias (constant value, making the distribution in fact non-gauss)
 *		- f_multiplier, f_center and f_sigma2 are parameteres in gauss distribution equation:
 *		  g(x) = f_multiplier * exp(-(x - f_center)^2 / f_sigma2)
 */
TGaussDistribution::TGaussDistribution(float _f_bias, float _f_multiplier,
	float _f_center, float _f_sigma2)
	:f_bias(_f_bias), f_multiplier(_f_multiplier), f_center(_f_center), f_sigma2(_f_sigma2)
{}

/*
 *	bool TGaussDistribution::Parse(CMaterialLexer &r_lexer,
 *		const char *&r_p_s_error_description)
 *		- expects current token to be token right after colon,
 *		  upon successful return leaves last token before semicolon
 *		- returns true on success, false on failure
 */
bool TGaussDistribution::Parse(CMaterialLexer &r_lexer, const char *&r_p_s_error_description)
{
	if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Float && 
	   r_lexer.n_Cur_Token() != CMaterialLexer::token_Int) {
		r_p_s_error_description = "bias must be float value";
		return false;
	}
	f_bias = r_lexer.t_Cur_Token().f_Value_AllowCast();
	// read bias

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
		r_p_s_error_description = "multiplier must be float value";
		return false;
	}
	f_multiplier = r_lexer.t_Cur_Token().f_Value_AllowCast();
	// read multiplier

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
		r_p_s_error_description = "center must be float value";
		return false;
	}
	f_center = r_lexer.t_Cur_Token().f_Value_AllowCast();
	// read center

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
		r_p_s_error_description = "sigma^2 must be float value";
		return false;
	}
	f_sigma2 = r_lexer.t_Cur_Token().f_Value_AllowCast();
	// read sigma^2

	return true;
}

/*
 *								=== ~TGaussDistribution ===
 */

/*
 *								=== TNoiseParams ===
 */

/*
 *	TNoiseParams::TNoiseParams()
 *		- default constructor; has no effect
 */
TNoiseParams::TNoiseParams()
{}

/*
 *	TNoiseParams::TNoiseParams(float _f_bias, int _n_octave_num, float _f_scale,
 *		float _f_amplitude, float _f_octave_exponent)
 *		- constructor
 *		- initializes noise parameters structure
 *		- f_bias is noise DC value
 *		- n_octave_num is number of noise octaves
 *		- f_scale is noise input coordinate scaling
 *		- f_amplitude is noise multiplier
 *		- individual octaves are attenuatet by f_octave_exponent,
 *		  raised to power of index of octave
 */
TNoiseParams::TNoiseParams(float _f_bias, int _n_octave_num, float _f_scale,
	float _f_amplitude, float _f_octave_exponent)
	:f_bias(_f_bias), n_octave_num(_n_octave_num), f_scale(_f_scale),
	f_amplitude(_f_amplitude), f_octave_exponent(_f_octave_exponent)
{}

/*
 *	bool TNoiseParams::Parse(CMaterialLexer &r_lexer, const char *&r_p_s_error_description)
 *		- expects current token to be token right after colon,
 *		  upon successful return leaves last token before semicolon
 *		- returns true on success, false on failure
 */
bool TNoiseParams::Parse(CMaterialLexer &r_lexer, const char *&r_p_s_error_description)
{
	if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Float && 
	   r_lexer.n_Cur_Token() != CMaterialLexer::token_Int) {
		r_p_s_error_description = "bias must be float value";
		return false;
	}
	f_bias = r_lexer.t_Cur_Token().f_Value_AllowCast();
	// read bias

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Int)) {
		r_p_s_error_description = "octave count must be int value";
		return false;
	}
	n_octave_num = r_lexer.t_Cur_Token().n_Value();
	// read number of octaves

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
		r_p_s_error_description = "scale must be float value";
		return false;
	}
	f_scale = r_lexer.t_Cur_Token().f_Value_AllowCast();
	// read scale

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
		r_p_s_error_description = "amplitude must be float value";
		return false;
	}
	f_amplitude = r_lexer.t_Cur_Token().f_Value_AllowCast();
	// read amplitude

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
		r_p_s_error_description = "octave exponent must be float value";
		return false;
	}
	f_octave_exponent = r_lexer.t_Cur_Token().f_Value_AllowCast();
	// read amplitude

	return true;
}

/*
 *								=== ~TNoiseParams ===
 */

/*
 *	static bool b_PoT(unsigned int n_value)
 *		- returns true in case n_value is power of two
 */
static bool b_PoT(unsigned int n_value)
{
	for(; n_value; n_value >>= 1) {
		if(n_value & 1) {
			n_value ^= 1; // flip first bit
			return !n_value; // must be zero (only 1 bit is allowed to be one)
		}
	}
	return true;
}

/*
 *								=== CTaskDescription::TOutputDescription ===
 */

/*
 *	void CTaskDescription::TOutputDescription::Defaults()
 *		- fills-in default values
 */
void CTaskDescription::TOutputDescription::Defaults()
{
	n_texture_size = 512;
	n_sample_num = 32;
	n_ao_sample_num = 32;
	n_shadow_sample_num = 10;
	f_shadow_sample_offset = 1;
	b_normalize = true;
	f_gamma = 2.2f;
	f_bias = 0;
	f_scale = 1;
	n_color_system_name = CColorSystem::system_Rec709;
	/*s_sky_bitmap = "";
	s_horizon_bitmap = "";
	s_output_bitmap = "skybox_%s.tga";*/ // deprecated
}

/*
 *	bool CTaskDescription::TOutputDescription::Parse(CMaterialLexer &r_lexer,
 *		const char *&r_p_s_error_description)
 *		- expects current token to be the node name token,
 *		  upon successful return leaves the end brace token
 *		- returns true on success, false on failure
 */
bool CTaskDescription::TOutputDescription::Parse(CMaterialLexer &r_lexer,
	const char *&r_p_s_error_description)
{
	_ASSERTE(r_lexer.n_Cur_Token() == CMaterialLexer::token_Node &&
		!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "output"));
	// make sure what is being parsed really is output node

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Begin)) {
		r_p_s_error_description = "unexpected token before \'{\'";
		return false;
	}
	// get begin token

	for(;;) {
		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Node, CMaterialLexer::token_End)) {
			r_p_s_error_description = "unexpected token before node-name";
			return false;
		}
		if(r_lexer.n_Cur_Token() == CMaterialLexer::token_End)
			return true;
		// make sure we've got node token

		CMaterialLexer::TToken t_node(r_lexer.t_Cur_Token());
		if(!t_node.p_s_Value()) {
			r_p_s_error_description = "not enough memory";
			return false;
		}
		// remember node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Colon)) {
			r_p_s_error_description = "unexpected token before \':\'";
			return false;
		}
		// must be followed by colon

		if(!r_lexer.Get_NextToken())
			return false;
		// and another token

		if(!strcmp(t_node.p_s_Value(), "texture-size")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int) {
				r_p_s_error_description = "texture-size must be integer value";
				return false;
			}
			n_texture_size = r_lexer.t_Cur_Token().n_Value();
			if(n_texture_size <= 0 || !b_PoT(n_texture_size)) {
				r_p_s_error_description = "texture-size must be power-of-two";
				return false;
			}
		} else if(!strcmp(t_node.p_s_Value(), "sample-num")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int) {
				r_p_s_error_description = "sample-num must be integer value";
				return false;
			}
			n_sample_num = r_lexer.t_Cur_Token().n_Value();
		} else if(!strcmp(t_node.p_s_Value(), "ambient-occlusion-sample-num")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int) {
				r_p_s_error_description = "ambient-occlusion-sample-num must be integer value";
				return false;
			}
			n_ao_sample_num = r_lexer.t_Cur_Token().n_Value();
		} else if(!strcmp(t_node.p_s_Value(), "shadow-sample-num")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int) {
				r_p_s_error_description = "shadow-sample-num must be integer value";
				return false;
			}
			n_shadow_sample_num = r_lexer.t_Cur_Token().n_Value();
		} else if(!strcmp(t_node.p_s_Value(), "shadow-sample-offset")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "shadow-sample-offset must be float value";
				return false;
			}
			f_shadow_sample_offset = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "normalize")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Bool) {
				r_p_s_error_description = "normalize must be boolean value";
				return false;
			}
			b_normalize = r_lexer.t_Cur_Token().n_Value() != 0;
		} else if(!strcmp(t_node.p_s_Value(), "gamma")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "gamma must be float value";
				return false;
			}
			f_gamma = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "bias")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "bias must be float value";
				return false;
			}
			f_bias = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "scale")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "scale must be float value";
				return false;
			}
			f_scale = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "color-system")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_String) {
				r_p_s_error_description = "color-system must be string";
				return false;
			}
			const char *p_s_color_sys_name = r_lexer.t_Cur_Token().p_s_Value();
			const struct {
				const char *p_s_name;
				int n_name;
			} p_color_system_list[] = {
				{"NTSC", CColorSystem::system_NTSC},
				{"EBU", CColorSystem::system_EBU},
				{"SMPTE", CColorSystem::system_SMPTE},
				{"HDTV", CColorSystem::system_HDTV},
				{"CIE", CColorSystem::system_CIE},
				{"Rec709", CColorSystem::system_Rec709}
			};
			for(unsigned int i = 0; i < sizeof(p_color_system_list) / sizeof(p_color_system_list[0]); ++ i) {
				if(!strcmp(p_s_color_sys_name, p_color_system_list[i].p_s_name)) {
					n_color_system_name = p_color_system_list[i].n_name;
					p_s_color_sys_name = 0; // gotcha
					break;
				}
			}
			if(p_s_color_sys_name) {
				r_p_s_error_description = "unknown color system";
				return false;
			}
		} else if(!strcmp(t_node.p_s_Value(), "sky-bitmap")) {
			r_p_s_error_description = "sky-bitmap was declared deprecated";
			return false;
		} else if(!strcmp(t_node.p_s_Value(), "horizon-bitmap")) {
			r_p_s_error_description = "horizon-bitmap was declared deprecated";
			return false;
		} else if(!strcmp(t_node.p_s_Value(), "output-bitmap")) {
			r_p_s_error_description = "output-bitmap was declared deprecated";
			return false;
		} else {
			r_p_s_error_description = "unknown node";
			return false;
		}
		// copy data from value token to appropriate
		// member variable, according to node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Semicolon)) {
			r_p_s_error_description = "unexpected token before \';\'";
			return false;
		}
		// must be followed by semicolon
	}
}

/*
 *								=== ~CTaskDescription::TOutputDescription ===
 */

/*
 *								=== CTaskDescription::TSkyDescription ===
 */

/*
 *	void CTaskDescription::TSkyDescription::Defaults()
 *		- fills-in default values
 */
void CTaskDescription::TSkyDescription::Defaults()
{
	f_planet_radius = 6378e3;
	f_atmosphere_radius = 6478e3;
	v_sun_direction = Vector3f(0, 1, 0);
	f_u = .7f;
	f_k_mie = .25e-9f;
	f_k_raileigh = .15e-31f;
	f_observer_altitude = 500;
}

/*
 *	bool CTaskDescription::TSkyDescription::Parse(CMaterialLexer &r_lexer,
 *		const char *&r_p_s_error_description)
 *		- expects current token to be the node name token,
 *		  upon successful return leaves the end brace token
 *		- returns true on success, false on failure
 */
bool CTaskDescription::TSkyDescription::Parse(CMaterialLexer &r_lexer,
	const char *&r_p_s_error_description)
{
	_ASSERTE(r_lexer.n_Cur_Token() == CMaterialLexer::token_Node &&
		!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "atmosphere"));
	// make sure what is being parsed really is sky description node

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Begin)) {
		r_p_s_error_description = "unexpected token before \'{\'";
		return false;
	}
	// get begin token

	for(;;) {
		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Node, CMaterialLexer::token_End)) {
			r_p_s_error_description = "unexpected token before node-name";
			return false;
		}
		if(r_lexer.n_Cur_Token() == CMaterialLexer::token_End)
			return true;
		// make sure we've got node token

		CMaterialLexer::TToken t_node(r_lexer.t_Cur_Token());
		if(!t_node.p_s_Value()) {
			r_p_s_error_description = "not enough memory";
			return false;
		}
		// remember node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Colon)) {
			r_p_s_error_description = "unexpected token before \':\'";
			return false;
		}
		// must be followed by colon

		if(!r_lexer.Get_NextToken())
			return false;
		// and another token

		if(!strcmp(t_node.p_s_Value(), "planet-r")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "planet-r must be float value";
				return false;
			}
			f_planet_radius = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "atmosphere-r")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "atmosphere-r must be float value";
				return false;
			}
			f_atmosphere_radius = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "sun-direction")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "sun-direction must be trinity of float values";
				return false;
			}
			v_sun_direction.f_X() = r_lexer.t_Cur_Token().f_Value_AllowCast();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
				r_p_s_error_description = "sun-direction must be trinity of float values";
				return false;
			}
			v_sun_direction.f_Y() = r_lexer.t_Cur_Token().f_Value_AllowCast();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
				r_p_s_error_description = "sun-direction must be trinity of float values";
				return false;
			}
			v_sun_direction.f_Z() = r_lexer.t_Cur_Token().f_Value_AllowCast();
			v_sun_direction.Normalize(); // normalize!
		} else if(!strcmp(t_node.p_s_Value(), "u")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "u must be float value";
				return false;
			}
			f_u = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "k-mie")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "k-mie must be float value";
				return false;
			}
			f_k_mie = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "k-raileigh")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "k-raileigh must be float value";
				return false;
			}
			f_k_raileigh = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "observer-altitude")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "observer-altitude must be float value";
				return false;
			}
			f_observer_altitude = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else {
			r_p_s_error_description = "unknown node";
			return false;
		}
		// copy data from value token to appropriate
		// member variable, according to node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Semicolon)) {
			r_p_s_error_description = "unexpected token before \';\'";
			return false;
		}
		// must be followed by semicolon
	}
}

/*
 *								=== ~CTaskDescription::TSkyDescription ===
 */

/*
 *								=== CTaskDescription::TSkyColor ===
 */

/*
 *	void CTaskDescription::TSkyColor::Defaults()
 *		- fills-in default values
 */
void CTaskDescription::TSkyColor::Defaults()
{
	n_texture_size = 256;
	f_y_cutoff = -1.5f; // nothing gets cut-off
	//f_overshoot = 1;
}

/*
 *	bool CTaskDescription::TSkyColor::Parse(CMaterialLexer &r_lexer,
 *		const char *&r_p_s_error_description)
 *		- expects current token to be the node name token,
 *		  upon successful return leaves the end brace token
 *		- returns true on success, false on failure
 */
bool CTaskDescription::TSkyColor::Parse(CMaterialLexer &r_lexer,
	const char *&r_p_s_error_description)
{
	_ASSERTE(r_lexer.n_Cur_Token() == CMaterialLexer::token_Node &&
		!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "sky-color"));
	// make sure what is being parsed really is sky-color node

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Begin)) {
		r_p_s_error_description = "unexpected token before \'{\'";
		return false;
	}
	// get begin token

	for(;;) {
		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Node, CMaterialLexer::token_End)) {
			r_p_s_error_description = "unexpected token before node-name";
			return false;
		}
		if(r_lexer.n_Cur_Token() == CMaterialLexer::token_End)
			return true;
		// make sure we've got node token

		CMaterialLexer::TToken t_node(r_lexer.t_Cur_Token());
		if(!t_node.p_s_Value()) {
			r_p_s_error_description = "not enough memory";
			return false;
		}
		// remember node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Colon)) {
			r_p_s_error_description = "unexpected token before \':\'";
			return false;
		}
		// must be followed by colon

		if(!r_lexer.Get_NextToken())
			return false;
		// and another token

		if(!strcmp(t_node.p_s_Value(), "texture-size")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int) {
				r_p_s_error_description = "texture-size must be int value";
				return false;
			}
			n_texture_size = r_lexer.t_Cur_Token().n_Value();
			if(n_texture_size <= 0 || !b_PoT((int)n_texture_size)) {
				r_p_s_error_description = "texture-size must be power-of-two";
				return false;
			}
		} else if(!strcmp(t_node.p_s_Value(), "y-cutoff")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "y-cutoff must be float value";
				return false;
			}
			f_y_cutoff = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "overshoot")) {
			/*if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "overshoot must be float value";
				return false;
			}
			f_overshoot = r_lexer.t_Cur_Token().f_Value_AllowCast();*/
			r_p_s_error_description = "the \'overshoot\' property was deprecated";
			return false;
		} else {
			r_p_s_error_description = "unknown node";
			return false;
		}
		// copy data from value token to appropriate
		// member variable, according to node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Semicolon)) {
			r_p_s_error_description = "unexpected token before \';\'";
			return false;
		}
		// must be followed by semicolon
	}
}

/*
 *								=== ~CTaskDescription::TSkyColor ===
 */

/*
 *								=== CTaskDescription::THorizonSampling ===
 */

/*
 *	void CTaskDescription::THorizonSampling::Defaults()
 *		- fills-in default values
 */
void CTaskDescription::THorizonSampling::Defaults()
{
	t_horizon_filter = TGaussDistribution(0, 1, 0, .5f);
	f_horizon_filter_cutoff = .1f;
	f_min_y = -.1f;
	f_max_y = .1f;
}

/*
 *	bool CTaskDescription::THorizonSampling::Parse(CMaterialLexer &r_lexer,
 *		const char *&r_p_s_error_description)
 *		- expects current token to be the node name token,
 *		  upon successful return leaves the end brace token
 *		- returns true on success, false on failure
 */
bool CTaskDescription::THorizonSampling::Parse(CMaterialLexer &r_lexer,
	const char *&r_p_s_error_description)
{
	_ASSERTE(r_lexer.n_Cur_Token() == CMaterialLexer::token_Node &&
		!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "horizon-color"));
	// make sure what is being parsed really is horizon sampling node

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Begin)) {
		r_p_s_error_description = "unexpected token before \'{\'";
		return false;
	}
	// get begin token

	for(;;) {
		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Node, CMaterialLexer::token_End)) {
			r_p_s_error_description = "unexpected token before node-name";
			return false;
		}
		if(r_lexer.n_Cur_Token() == CMaterialLexer::token_End)
			return true;
		// make sure we've got node token

		CMaterialLexer::TToken t_node(r_lexer.t_Cur_Token());
		if(!t_node.p_s_Value()) {
			r_p_s_error_description = "not enough memory";
			return false;
		}
		// remember node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Colon)) {
			r_p_s_error_description = "unexpected token before \':\'";
			return false;
		}
		// must be followed by colon

		if(!r_lexer.Get_NextToken())
			return false;
		// and another token

		if(!strcmp(t_node.p_s_Value(), "cutoff")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "cutoff must be float value";
				return false;
			}
			f_horizon_filter_cutoff = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "y-range")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "min-y must be float value";
				return false;
			}
			f_min_y = r_lexer.t_Cur_Token().f_Value_AllowCast();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
				r_p_s_error_description = "max-y must be float value";
				return false;
			}
			f_max_y = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "gauss")) {
			if(!t_horizon_filter.Parse(r_lexer, r_p_s_error_description))
				return false;
		} else {
			r_p_s_error_description = "unknown node";
			return false;
		}
		// copy data from value token to appropriate
		// member variable, according to node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Semicolon)) {
			r_p_s_error_description = "unexpected token before \';\'";
			return false;
		}
		// must be followed by semicolon
	}
}

/*
 *								=== ~CTaskDescription::THorizonSampling ===
 */

/*
 *								=== CTaskDescription::TCloudLayerDescription ===
 */

/*
 *	void CTaskDescription::TCloudLayerDescription::Defaults()
 *		- fills-in default values
 */
void CTaskDescription::TCloudLayerDescription::Defaults()
{
	f_altitude = 10e3; // 10km
	f_thickness = 100; // 100m

	t_perturb = TNoiseParams(0, 0, 1, 1, .5f);
	t_noise = TNoiseParams(0, 1, 1, 1, .5f);
	n_function = func_Plain;
	// default perturbation and noise function

}

/*
 *	bool CTaskDescription::TCloudLayerDescription::Parse(CMaterialLexer &r_lexer,
 *		const char *&r_p_s_error_description)
 *		- expects current token to be the node name token,
 *		  upon successful return leaves the end brace token
 *		- returns true on success, false on failure
 */
bool CTaskDescription::TCloudLayerDescription::Parse(CMaterialLexer &r_lexer,
	const char *&r_p_s_error_description)
{
	_ASSERTE(r_lexer.n_Cur_Token() == CMaterialLexer::token_Node &&
		!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "layer"));
	// make sure what is being parsed really is clouds layer node

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Begin)) {
		r_p_s_error_description = "unexpected token before \'{\'";
		return false;
	}
	// get begin token

	for(;;) {
		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Node, CMaterialLexer::token_End)) {
			r_p_s_error_description = "unexpected token before node-name";
			return false;
		}
		if(r_lexer.n_Cur_Token() == CMaterialLexer::token_End)
			return true;
		// make sure we've got node token

		CMaterialLexer::TToken t_node(r_lexer.t_Cur_Token());
		if(!t_node.p_s_Value()) {
			r_p_s_error_description = "not enough memory";
			return false;
		}
		// remember node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Colon)) {
			r_p_s_error_description = "unexpected token before \':\'";
			return false;
		}
		// must be followed by colon

		if(!r_lexer.Get_NextToken())
			return false;
		// and another token

		if(!strcmp(t_node.p_s_Value(), "perturb")) {
			if(!t_perturb.Parse(r_lexer, r_p_s_error_description))
				return false;
		} else if(!strcmp(t_node.p_s_Value(), "noise")) {
			if(!t_noise.Parse(r_lexer, r_p_s_error_description))
				return false;
		} else if(!strcmp(t_node.p_s_Value(), "function")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Node) {
				r_p_s_error_description = "function must be one of "
					"plain, abs, exp, exp-abs, force-image";
				return false;
			}
			if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "plain"))
				n_function = func_Plain;
			else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "abs"))
				n_function = func_Abs;
			else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "exp"))
				n_function = func_Exp;
			else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "exp-abs"))
				n_function = func_ExpAbs;
			else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "force-image"))
				n_function = func_ForceImage;
			else {
				r_p_s_error_description = "function must be one of "
					"plain, abs, exp, exp-abs, force-image";
				return false;
			}
		} else if(!strcmp(t_node.p_s_Value(), "altitude")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "altitude must be float value";
				return false;
			}
			f_altitude = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "thickness")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "thickness must be float value";
				return false;
			}
			f_thickness = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else {
			r_p_s_error_description = "unknown node";
			return false;
		}
		// copy data from value token to appropriate
		// member variable, according to node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Semicolon)) {
			r_p_s_error_description = "unexpected token before \';\'";
			return false;
		}
		// must be followed by semicolon
	}
}

/*
 *								=== ~CTaskDescription::TCloudLayerDescription ===
 */

/*
 *								=== CTaskDescription::TCloudsDescription ===
 */

/*
 *	void CTaskDescription::TCloudsDescription::Defaults()
 *		- fills-in default values
 */
void CTaskDescription::TCloudsDescription::Defaults()
{
	cloud_layers.clear();
}

/*
 *	bool CTaskDescription::TCloudsDescription::Parse(CMaterialLexer &r_lexer,
 *		const char *&r_p_s_error_description)
 *		- expects current token to be the node name token,
 *		  upon successful return leaves the end brace token
 *		- returns true on success, false on failure
 */
bool CTaskDescription::TCloudsDescription::Parse(CMaterialLexer &r_lexer,
	const char *&r_p_s_error_description)
{
	_ASSERTE(r_lexer.n_Cur_Token() == CMaterialLexer::token_Node &&
		!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "clouds"));
	// make sure what is being parsed really is clouds description node

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Begin)) {
		r_p_s_error_description = "unexpected token before \'{\'";
		return false;
	}
	// get begin token

	for(;;) {
		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Node, CMaterialLexer::token_End)) {
			r_p_s_error_description = "unexpected token before node-name";
			return false;
		}
		if(r_lexer.n_Cur_Token() == CMaterialLexer::token_End)
			return true;
		// make sure we've got node token

		CMaterialLexer::TToken t_node(r_lexer.t_Cur_Token());
		if(!t_node.p_s_Value()) {
			r_p_s_error_description = "not enough memory";
			return false;
		}
		// remember node name

		if(strcmp(t_node.p_s_Value(), "layer")) { // layer node starts with '{', not ':'
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Colon)) {
				r_p_s_error_description = "unexpected token before \':\'";
				return false;
			}
			// must be followed by colon

			if(!r_lexer.Get_NextToken())
				return false;
			// and another token
		}

		if(!strcmp(t_node.p_s_Value(), "layer")) {
			TCloudLayerDescription t_layer;
			t_layer.Defaults(); // this allows fields to be filled voluntarily
			if(!t_layer.Parse(r_lexer, r_p_s_error_description))
				return false;
			cloud_layers.reserve(cloud_layers.size() + 1);
			if(cloud_layers.capacity() < cloud_layers.size() + 1) {
				r_p_s_error_description = "not enough memory";
				return false;
			}
			cloud_layers.push_back(t_layer);
			continue;
		} else {
			r_p_s_error_description = "unknown node";
			return false;
		}
		// copy data from value token to appropriate
		// member variable, according to node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Semicolon)) {
			r_p_s_error_description = "unexpected token before \';\'";
			return false;
		}
		// must be followed by semicolon
	}
}

/*
 *								=== ~CTaskDescription::TCloudsDescription ===
 */

/*
 *								=== CTaskDescription::TTerrainDescription::TTerrainOp ===
 */

/*
 *	CTaskDescription::TTerrainDescription::TTerrainOp::TTerrainOp(int _n_type, float f_arg0 = 0,
 *		float f_arg1 = 0, float f_arg2 = 0, float f_arg3 = 0, float f_arg4 = 0, int _n_seed = 0)
 *		- terrain operator default constructor
 *		- n_type is one of op_Fill, op_Normalize, ... op_Perlin
 *		- f_arg0 .. f_arg4 are operator arguments
 *		- n_seed is random number generator seed
 */
CTaskDescription::TTerrainDescription::TTerrainOp::TTerrainOp(int _n_type, float f_arg0,
	float f_arg1, float f_arg2, float f_arg3, float f_arg4, int _n_seed)
	:n_type(_n_type), n_seed(_n_seed)
{
	p_args[0] = f_arg0;
	p_args[1] = f_arg1;
	p_args[2] = f_arg2;
	p_args[3] = f_arg3;
	p_args[4] = f_arg4;
}

/*
 *	bool CTaskDescription::TTerrainDescription::TTerrainOp::Apply(CHeightField &r_target) const
 *		- apply operator on terrain
 */
bool CTaskDescription::TTerrainDescription::TTerrainOp::Apply(CHeightField &r_target) const
{
	switch(n_type) {
	case op_Fill:
		r_target.Fill(p_args[0]);
		// fill (1 param value)
		break;
	case op_Normalize:
		r_target.Normalize(p_args[0], p_args[1]);
		// normalize (2 params: min max)
		break;
	case op_Mean:
		r_target.Mean((int)p_args[0]);
		// mean filter (1 param kernel size)
		break;
	case op_Exp:
		{
			for(int y = 0, h = r_target.n_Height(); y < h; ++ y) {
				for(int x = 0, w = r_target.n_Width(); x < w; ++ x)
					r_target[y][x] = exp(-r_target[y][x]);
			}
		}
		// exp(-height) (0 params)
		break;
	case op_FForm:
		srand(n_seed); // seed
		r_target.FaultFormation((int)p_args[0], p_args[1]);
		// fault formation (2 params: no of iters and displacement)
		break;
	case op_PDepo:
		srand(n_seed); // seed
		r_target.ParticleDeposition((int)(r_target.n_Width() * p_args[0]),
			(int)(r_target.n_Height() * p_args[1]), p_args[2],
			(int)p_args[3], (int)p_args[4]);
		// particle deposition (5 params: xy of seed: shift proba: max walk: iters)
		break;
	case op_Perlin:
		r_target.PerlinNoise((int)p_args[0], p_args[1],
			p_args[2], p_args[3], p_args[4]);
		// perlin noise (5 params: octave num: size: amplitude: z: durability)
		break;
	default:
		_ASSERTE(0); // shouldn't happen, it had to pass trough parser
		return false;
	}
	return true;
}

/*
 *								=== ~CTaskDescription::TTerrainDescription::TTerrainOp ===
 */

/*
 *								=== CTaskDescription::TTerrainDescription::TTextureOp ===
 */

/*
 *	void CTaskDescription::TTerrainDescription::TTextureOp::Defaults()
 *		- fills-in default values
 */
void CTaskDescription::TTerrainDescription::TTextureOp::Defaults()
{
	n_texcoord_style = map_Volume;
	t_perturb = TNoiseParams(0, 0, 1, 1, .5f);
	t_noise = TNoiseParams(0, 1, 1, 1, .5f);
	n_function = func_Plain;
	palette.clear();
	t_height_matching = TGaussDistribution(1, 0, 0, 1);
	v_ref_normal = Vector3f(0, 1, 0);
	t_slope_matching = TGaussDistribution(0, 0, 0, 1);;
	t_matching_distortion = TNoiseParams(0, 0, 1, 1, .5f);
}

/*
 *	bool CTaskDescription::TTerrainDescription::TTextureOp::Parse(
 *		CMaterialLexer &r_lexer, const char *&r_p_s_error_description)
 *		- expects current token to be the node name token,
 *		  upon successful return leaves the end brace token
 *		- returns true on success, false on failure
 */
bool CTaskDescription::TTerrainDescription::TTextureOp::Parse(CMaterialLexer &r_lexer,
	const char *&r_p_s_error_description)
{
	_ASSERTE(r_lexer.n_Cur_Token() == CMaterialLexer::token_Node &&
		!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "texture"));
	// make sure what is being parsed really is texture op node

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Begin)) {
		r_p_s_error_description = "unexpected token before \'{\'";
		return false;
	}
	// get begin token

	for(;;) {
		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Node, CMaterialLexer::token_End)) {
			r_p_s_error_description = "unexpected token before node-name";
			return false;
		}
		if(r_lexer.n_Cur_Token() == CMaterialLexer::token_End) {
			std::sort(palette.begin(), palette.end());
			// sort palette

			return true;
		}
		// make sure we've got node token

		CMaterialLexer::TToken t_node(r_lexer.t_Cur_Token());
		if(!t_node.p_s_Value()) {
			r_p_s_error_description = "not enough memory";
			return false;
		}
		// remember node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Colon)) {
			r_p_s_error_description = "unexpected token before \':\'";
			return false;
		}
		// must be followed by colon

		if(!r_lexer.Get_NextToken())
			return false;
		// and another token

		if(!strcmp(t_node.p_s_Value(), "texgen")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Node) {
				r_p_s_error_description = "texgen must be one of volume, xz-plane or y-axis";
				return false;
			}
			if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "xz-plane"))
				n_texcoord_style = map_XZPlane;
			else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "y-axis"))
				n_texcoord_style = map_YAxis;
			else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "volume"))
				n_texcoord_style = map_Volume;
			else {
				r_p_s_error_description = "texgen must be one of volume, xz-plane or y-axis";
				return false;
			}
		} else if(!strcmp(t_node.p_s_Value(), "perturb")) {
			if(!t_perturb.Parse(r_lexer, r_p_s_error_description))
				return false;
		} else if(!strcmp(t_node.p_s_Value(), "noise")) {
			if(!t_noise.Parse(r_lexer, r_p_s_error_description))
				return false;
		} else if(!strcmp(t_node.p_s_Value(), "function")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Node) {
				r_p_s_error_description = "function must be one of "
					"plain, abs, exp, exp-abs, force-image";
				return false;
			}
			if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "plain"))
				n_function = func_Plain;
			else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "abs"))
				n_function = func_Abs;
			else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "exp"))
				n_function = func_Exp;
			else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "exp-abs"))
				n_function = func_ExpAbs;
			else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "force-image"))
				n_function = func_ForceImage;
			else {
				r_p_s_error_description = "function must be one of "
					"plain, abs, exp, exp-abs, force-image";
				return false;
			}
		} else if(!strcmp(t_node.p_s_Value(), "height-gauss")) {
			if(!t_height_matching.Parse(r_lexer, r_p_s_error_description))
				return false;
		} else if(!strcmp(t_node.p_s_Value(), "slope-gauss")) {
			if(!t_slope_matching.Parse(r_lexer, r_p_s_error_description))
				return false;
		} else if(!strcmp(t_node.p_s_Value(), "match-distort")) {
			if(!t_matching_distortion.Parse(r_lexer, r_p_s_error_description))
				return false;
		} else if(!strcmp(t_node.p_s_Value(), "ref-normal")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "normal must be trinity of float values";
				return false;
			}
			v_ref_normal.f_X() = r_lexer.t_Cur_Token().f_Value_AllowCast();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
				r_p_s_error_description = "normal must be trinity of float values";
				return false;
			}
			v_ref_normal.f_Y() = r_lexer.t_Cur_Token().f_Value_AllowCast();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
				r_p_s_error_description = "normal must be trinity of float values";
				return false;
			}
			v_ref_normal.f_Z() = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "palette")) {
			TPaletteEntry t_pal_entry;
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "intensity must be float value";
				return false;
			}
			t_pal_entry.f_pos = r_lexer.t_Cur_Token().f_Value_AllowCast();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
				r_p_s_error_description = "color must be trinity of float values";
				return false;
			}
			t_pal_entry.v_color.f_X() = r_lexer.t_Cur_Token().f_Value_AllowCast();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
				r_p_s_error_description = "color must be trinity of float values";
				return false;
			}
			t_pal_entry.v_color.f_Y() = r_lexer.t_Cur_Token().f_Value_AllowCast();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
				r_p_s_error_description = "color must be trinity of float values";
				return false;
			}
			t_pal_entry.v_color.f_Z() = r_lexer.t_Cur_Token().f_Value_AllowCast();
			palette.reserve(palette.size() + 1);
			if(palette.capacity() < palette.size() + 1) {
				r_p_s_error_description = "not enough memory";
				return false;
			}
			palette.push_back(t_pal_entry);
		} else {
			r_p_s_error_description = "unknown node";
			return false;
		}
		// copy data from value token to appropriate
		// member variable, according to node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Semicolon)) {
			r_p_s_error_description = "unexpected token before \';\'";
			return false;
		}
		// must be followed by semicolon
	}
}

/*
 *								=== ~CTaskDescription::TTerrainDescription::TTextureOp ===
 */

/*
 *								=== CTaskDescription::TTerrainDescription ===
 */

/*
 *	void CTaskDescription::TTerrainDescription::Defaults()
 *		- fills-in default values
 */
void CTaskDescription::TTerrainDescription::Defaults()
{
	n_width = 256;
	n_height = 256;
	f_tile_size = 100;
	op_list.clear();
	texture_list.clear();
	f_fog_base = 0;
	f_fog_height = 1000;
	f_fog_thickness = 0;
	f_ambient = 1;
	f_diffuse = 1;
}

/*
 *	bool CTaskDescription::TTerrainDescription::Parse(CMaterialLexer &r_lexer,
 *		const char *&r_p_s_error_description)
 *		- expects current token to be the node name token,
 *		  upon successful return leaves the end brace token
 *		- returns true on success, false on failure
 */
bool CTaskDescription::TTerrainDescription::Parse(CMaterialLexer &r_lexer,
	const char *&r_p_s_error_description)
{
	_ASSERTE(r_lexer.n_Cur_Token() == CMaterialLexer::token_Node &&
		!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "terrain"));
	// make sure what is being parsed really is terrain description node

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Begin)) {
		r_p_s_error_description = "unexpected token before \'{\'";
		return false;
	}
	// get begin token

	for(;;) {
		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Node, CMaterialLexer::token_End)) {
			r_p_s_error_description = "unexpected token before node-name";
			return false;
		}
		if(r_lexer.n_Cur_Token() == CMaterialLexer::token_End)
			return true;
		// make sure we've got node token

		CMaterialLexer::TToken t_node(r_lexer.t_Cur_Token());
		if(!t_node.p_s_Value()) {
			r_p_s_error_description = "not enough memory";
			return false;
		}
		// remember node name

		if(strcmp(t_node.p_s_Value(), "generator") && // generator node starts with '{', not ':'
		   strcmp(t_node.p_s_Value(), "texture")) { // texture node starts with '{', not ':'
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Colon)) {
				r_p_s_error_description = "unexpected token before \':\'";
				return false;
			}
			// must be followed by colon

			if(!r_lexer.Get_NextToken())
				return false;
			// and another token
		}

		if(!strcmp(t_node.p_s_Value(), "width")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int) {
				r_p_s_error_description = "width must be int value";
				return false;
			}
			n_width = r_lexer.t_Cur_Token().n_Value();
		} else if(!strcmp(t_node.p_s_Value(), "height")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int) {
				r_p_s_error_description = "height must be int value";
				return false;
			}
			n_height = r_lexer.t_Cur_Token().n_Value();
		} else if(!strcmp(t_node.p_s_Value(), "tile-size")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "tile-size must be float value";
				return false;
			}
			f_tile_size = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "ambient")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "ambient must be float value";
				return false;
			}
			f_ambient = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "diffuse")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "diffuse must be float value";
				return false;
			}
			f_diffuse = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "fog")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "fog-base must be float value";
				return false;
			}
			f_fog_base = r_lexer.t_Cur_Token().f_Value_AllowCast();

			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
				r_p_s_error_description = "fog-height must be float value";
				return false;
			}
			f_fog_height = r_lexer.t_Cur_Token().f_Value_AllowCast();

			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
				r_p_s_error_description = "fog-thickness must be float value";
				return false;
			}
			f_fog_thickness = r_lexer.t_Cur_Token().f_Value_AllowCast();
		} else if(!strcmp(t_node.p_s_Value(), "generator")) {
			if(!ParseGenerator(r_lexer, r_p_s_error_description))
				return false;
			continue;
		} else if(!strcmp(t_node.p_s_Value(), "texture")) {
			TTextureOp t_op;
			t_op.Defaults(); // this allows fields to be filled voluntarily
			if(!t_op.Parse(r_lexer, r_p_s_error_description))
				return false;
			texture_list.reserve(texture_list.size() + 1);
			if(texture_list.capacity() < texture_list.size() + 1) {
				r_p_s_error_description = "not enough memory";
				return false;
			}
			texture_list.push_back(t_op);
			continue;
		} else {
			r_p_s_error_description = "unknown node";
			return false;
		}
		// copy data from value token to appropriate
		// member variable, according to node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Semicolon)) {
			r_p_s_error_description = "unexpected token before \';\'";
			return false;
		}
		// must be followed by semicolon
	}
}

/*
 *	bool CTaskDescription::TTerrainDescription::ParseGenerator(CMaterialLexer &r_lexer,
 *		const char *&r_p_s_error_description)
 *		- expects current token to be the node name token,
 *		  upon successful return leaves the end brace token
 *		- returns true on success, false on failure
 */
bool CTaskDescription::TTerrainDescription::ParseGenerator(CMaterialLexer &r_lexer,
	const char *&r_p_s_error_description)
{
	_ASSERTE(r_lexer.n_Cur_Token() == CMaterialLexer::token_Node &&
		!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "generator"));
	// make sure what is being parsed really is terrain generator node

	if(!r_lexer.Get_NextToken(CMaterialLexer::token_Begin)) {
		r_p_s_error_description = "unexpected token before \'{\'";
		return false;
	}
	// get begin token

	for(;;) {
		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Node, CMaterialLexer::token_End)) {
			r_p_s_error_description = "unexpected token before node-name";
			return false;
		}
		if(r_lexer.n_Cur_Token() == CMaterialLexer::token_End)
			return true;
		// make sure we've got node token

		CMaterialLexer::TToken t_node(r_lexer.t_Cur_Token());
		if(!t_node.p_s_Value()) {
			r_p_s_error_description = "not enough memory";
			return false;
		}
		// remember node name

		if(strcmp(t_node.p_s_Value(), "exp")) { // exp node has no contents, semicolon follows
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Colon)) {
				r_p_s_error_description = "unexpected token before \':\'";
				return false;
			}
			// must be followed by colon

			if(!r_lexer.Get_NextToken())
				return false;
			// and another token
		}

		if(!strcmp(t_node.p_s_Value(), "fill")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "fill must be float value";
				return false;
			}
			op_list.reserve(op_list.size() + 1);
			if(op_list.capacity() < op_list.size() + 1) {
				r_p_s_error_description = "not enough memory";
				return false;
			}
			op_list.push_back(TTerrainOp(op_Fill, r_lexer.t_Cur_Token().f_Value_AllowCast()));
		} else if(!strcmp(t_node.p_s_Value(), "mean")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int) {
				r_p_s_error_description = "kernel size must be int value";
				return false;
			}
			op_list.reserve(op_list.size() + 1);
			if(op_list.capacity() < op_list.size() + 1) {
				r_p_s_error_description = "not enough memory";
				return false;
			}
			op_list.push_back(TTerrainOp(op_Mean, r_lexer.t_Cur_Token().n_Value()));
		} else if(!strcmp(t_node.p_s_Value(), "normalize")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "lower bound must be float value";
				return false;
			}
			float f_lower = r_lexer.t_Cur_Token().f_Value_AllowCast();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
				r_p_s_error_description = "upper bound must be float value";
				return false;
			}
			float f_upper = r_lexer.t_Cur_Token().f_Value_AllowCast();
			op_list.reserve(op_list.size() + 1);
			if(op_list.capacity() < op_list.size() + 1) {
				r_p_s_error_description = "not enough memory";
				return false;
			}
			op_list.push_back(TTerrainOp(op_Normalize, f_lower, f_upper));
		} else if(!strcmp(t_node.p_s_Value(), "exp")) {
			op_list.reserve(op_list.size() + 1);
			if(op_list.capacity() < op_list.size() + 1) {
				r_p_s_error_description = "not enough memory";
				return false;
			}
			op_list.push_back(TTerrainOp(op_Exp));
		} else if(!strcmp(t_node.p_s_Value(), "fault-formation")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int) {
				r_p_s_error_description = "iteration count must be int value";
				return false;
			}
			int n_iter_num = r_lexer.t_Cur_Token().n_Value();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
				r_p_s_error_description = "displacement must be float value";
				return false;
			}
			float f_displacement = r_lexer.t_Cur_Token().f_Value_AllowCast();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Int)) {
				r_p_s_error_description = "seed must be int value";
				return false;
			}
			int n_seed = r_lexer.t_Cur_Token().n_Value();
			op_list.reserve(op_list.size() + 1);
			if(op_list.capacity() < op_list.size() + 1) {
				r_p_s_error_description = "not enough memory";
				return false;
			}
			op_list.push_back(TTerrainOp(op_FForm, n_iter_num, f_displacement, 0, 0, 0, n_seed));
		} else if(!strcmp(t_node.p_s_Value(), "particle-depo")) {
			if(r_lexer.n_Cur_Token() != CMaterialLexer::token_Int &&
			   r_lexer.n_Cur_Token() != CMaterialLexer::token_Float) {
				r_p_s_error_description = "x-pos must be float value";
				return false;
			}
			float f_x = r_lexer.t_Cur_Token().f_Value_AllowCast();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
				r_p_s_error_description = "y-pos must be float value";
				return false;
			}
			float f_y = r_lexer.t_Cur_Token().f_Value_AllowCast();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Float, CMaterialLexer::token_Int)) {
				r_p_s_error_description = "shift probability must be float value";
				return false;
			}
			float f_shift_proba = r_lexer.t_Cur_Token().f_Value_AllowCast();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Int)) {
				r_p_s_error_description = "max walk must be int value";
				return false;
			}
			int n_max_walk = r_lexer.t_Cur_Token().n_Value();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Int)) {
				r_p_s_error_description = "iteration count must be int value";
				return false;
			}
			int n_iter_num = r_lexer.t_Cur_Token().n_Value();
			if(!r_lexer.Get_NextToken(CMaterialLexer::token_Int)) {
				r_p_s_error_description = "seed must be int value";
				return false;
			}
			int n_seed = r_lexer.t_Cur_Token().n_Value();
			op_list.reserve(op_list.size() + 1);
			if(op_list.capacity() < op_list.size() + 1) {
				r_p_s_error_description = "not enough memory";
				return false;
			}
			op_list.push_back(TTerrainOp(op_PDepo, f_x, f_y,
				f_shift_proba, n_max_walk, n_iter_num, n_seed));
		} else if(!strcmp(t_node.p_s_Value(), "perlin")) {
			TNoiseParams t_params;
			if(!t_params.Parse(r_lexer, r_p_s_error_description))
				return false;
			op_list.reserve(op_list.size() + 1);
			if(op_list.capacity() < op_list.size() + 1) {
				r_p_s_error_description = "not enough memory";
				return false;
			}
			op_list.push_back(TTerrainOp(op_Perlin, t_params.n_octave_num,
				t_params.f_scale, t_params.f_amplitude, .5f, t_params.f_octave_exponent));
		} else {
			r_p_s_error_description = "unknown node";
			return false;
		}
		// copy data from value token to appropriate
		// member variable, according to node name

		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Semicolon)) {
			r_p_s_error_description = "unexpected token before \';\'";
			return false;
		}
		// must be followed by semicolon
	}
}

/*
 *								=== ~CTaskDescription::TTerrainDescription ===
 */

/*
 *								=== CTaskDescription ===
 */

/*
 *	CTaskDescription::CTaskDescription()
 *		- default constructor
 *		- fills-in default values
 */
CTaskDescription::CTaskDescription()
{
	Defaults();
}

/*
 *	void CTaskDescription::Defaults()
 *		- fills-in default values
 */
void CTaskDescription::Defaults()
{
	m_output.Defaults();
	m_sky.Defaults();
	m_sky_color.Defaults();
	m_horizon.Defaults();
	m_terrain.Defaults();
}

/*
 *	bool CTaskDescription::Parse(const char *p_s_file, const char *&r_p_s_error_description,
 *		int &r_n_error_line, int &r_n_error_column)
 *		- parses task description from file p_s_file
 *		- r_p_s_error_description is filled with error description,
 *		  r_n_error_line and r_n_error_column are filled with error
 *		  line and columnt zero-based index (if error occured, otherwise
 *		  r_p_s_error_description is 0 and values of r_n_error_line and
 *		  r_n_error_column are undefined)
 *		- note this automatically resets all data to defaults before actual parsing
 *		- returns true on success, false on failure
 */
bool CTaskDescription::Parse(const char *p_s_file, const char *&r_p_s_error_description,
	int &r_n_error_line, int &r_n_error_column)
{
	CMaterialLexer lexer(p_s_file);
	if(!lexer.b_Status())
		return false;

	Defaults();
	// defaults

	if(!Parse_Loop(lexer, r_p_s_error_description)) {
		if(!r_p_s_error_description)
			r_p_s_error_description = "lexical error";
		r_n_error_line = lexer.t_Cur_Token().n_Line();
		r_n_error_column = lexer.t_Cur_Token().n_Column();
		return false;
	}

	r_p_s_error_description = 0;

	return true;
}

/*
 *	bool CTaskDescription::Parse_Loop(CMaterialLexer &r_lexer,
 *		const char *&r_p_s_error_description)
 *		- main parsing loop
 *		- returns true on success, false on failure
 */
bool CTaskDescription::Parse_Loop(CMaterialLexer &r_lexer, const char *&r_p_s_error_description)
{
	for(;;) {
		if(!r_lexer.Get_NextToken(CMaterialLexer::token_Node, CMaterialLexer::token_EOF)) {
			r_p_s_error_description = "unexpected token before node-name";
			return false;
		}
		if(r_lexer.n_Cur_Token() == CMaterialLexer::token_EOF)
			return true;
		// make sure we've got node token

		if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "output")) {
			if(!m_output.Parse(r_lexer, r_p_s_error_description))
				return false;
		} else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "atmosphere")) {
			if(!m_sky.Parse(r_lexer, r_p_s_error_description))
				return false;
		} else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "sky-color")) {
			if(!m_sky_color.Parse(r_lexer, r_p_s_error_description))
				return false;
		} else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "clouds")) {
			if(!m_clouds.Parse(r_lexer, r_p_s_error_description))
				return false;
		} else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "horizon-color")) {
			if(!m_horizon.Parse(r_lexer, r_p_s_error_description))
				return false;
		} else if(!strcmp(r_lexer.t_Cur_Token().p_s_Value(), "terrain")) {
			if(!m_terrain.Parse(r_lexer, r_p_s_error_description))
				return false;
		} else {
			r_p_s_error_description = "unknown node";
			return false;
		}
	}
	// parse individual nodes
}

/*
 *								=== ~CTaskDescription ===
 */
