/*
								+----------------------------------+
								|                                  |
								|    ***  Task description  ***    |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|              Task.h              |
								|                                  |
								+----------------------------------+
*/

#ifndef __TASK_INCLUDED
#define __TASK_INCLUDED

class CMaterialLexer;

/*
 *	struct TGaussDistribution
 *		- gaussian distribution
 */
struct TGaussDistribution {
	float f_bias;
	float f_multiplier;
	float f_center;
	float f_sigma2;

	/*
	 *	TGaussDistribution::TGaussDistribution()
	 *		- default constructor; has no effect
	 */
	TGaussDistribution();

	/*
	 *	TGaussDistribution::TGaussDistribution(float _f_bias, float _f_multiplier,
	 *		float _f_center, float _f_sigma2)
	 *		- constructor
	 *		- initializes gauss distribution structure
	 *		- f_bias is bias (constant value, making the distribution in fact non-gauss)
	 *		- f_multiplier, f_center and f_sigma2 are parameteres in gauss distribution equation:
	 *		  g(x) = f_multiplier * exp(-(x - f_center)^2 / f_sigma2)
	 */
	TGaussDistribution(float _f_bias, float _f_multiplier, float _f_center, float _f_sigma2);

	/*
	 *	bool TGaussDistribution::Parse(CMaterialLexer &r_lexer,
	 *		const char *&r_p_s_error_description)
	 *		- expects current token to be token right after colon,
	 *		  upon successful return leaves last token before semicolon
	 *		- returns true on success, false on failure
	 */
	bool Parse(CMaterialLexer &r_lexer, const char *&r_p_s_error_description);

	/*
	 *	inline float TGaussDistribution::f_Value(float f_x) const
	 *		- returns function value at f_x
	 */
	inline float f_Value(float f_x) const
	{
		f_x -= f_center;
		return f_bias + f_multiplier * exp(-f_x * f_x / f_sigma2);
	}
};

/*
 *	struct TNoiseParams
 *		- perlin noise parameters
 */
struct TNoiseParams {
	float f_bias;
	int n_octave_num;
	float f_scale;
	float f_amplitude;
	float f_octave_exponent;

	/*
	 *	TNoiseParams::TNoiseParams()
	 *		- default constructor; has no effect
	 */
	TNoiseParams();

	/*
	 *	TNoiseParams::TNoiseParams(float _f_bias, int _n_octave_num, float _f_scale,
	 *		float _f_amplitude, float _f_octave_exponent)
	 *		- constructor
	 *		- initializes noise parameters structure
	 *		- f_bias is noise DC value
	 *		- n_octave_num is number of noise octaves
	 *		- f_scale is noise input coordinate scaling
	 *		- f_amplitude is noise multiplier
	 *		- individual octaves are attenuatet by f_octave_exponent,
	 *		  raised to power of index of octave
	 */
	TNoiseParams(float _f_bias, int _n_octave_num, float _f_scale,
		float _f_amplitude, float _f_octave_exponent);

	/*
	 *	bool TNoiseParams::Parse(CMaterialLexer &r_lexer, const char *&r_p_s_error_description)
	 *		- expects current token to be token right after colon,
	 *		  upon successful return leaves last token before semicolon
	 *		- returns true on success, false on failure
	 */
	bool Parse(CMaterialLexer &r_lexer, const char *&r_p_s_error_description);

	/*
	 *	float f_Value(Vector3f v_pos) const
	 *		- calculates noise value at position v_pos
	 */
	float f_Value(Vector3f v_pos) const
	{
		if(!n_octave_num)
			return f_bias;

		float f_rescale = 0;
		float f_result = 0, f_exponent = 1;
		v_pos *= f_scale;
		for(int i = 0; i < n_octave_num; ++ i, v_pos *= 2, f_exponent *= f_octave_exponent) {
			f_result += CPerlinNoise::f_Noise(v_pos.f_X(), v_pos.f_Y(), v_pos.f_Z()) * f_exponent;
			f_rescale += f_exponent;
		}
		return f_result * f_amplitude / f_rescale + f_bias;
	}

	/*
	 *	Vector3f v_Value(Vector3f v_pos) const
	 *		- calculates vector noise value at position v_pos
	 */
	Vector3f v_Value(Vector3f v_pos) const
	{
		if(!n_octave_num)
			return Vector3f(f_bias, f_bias, f_bias);

		Vector3f v_result(0, 0, 0);
		float f_rescale = 0;
		float f_exponent = 1;
		v_pos *= f_scale;
		for(int i = 0; i < n_octave_num; ++ i, v_pos *= 2, f_exponent *= f_octave_exponent) {
			v_result.f_X() += CPerlinNoise::f_Noise(v_pos.f_X(),
				v_pos.f_Y(), v_pos.f_Z(), .5f) * f_exponent;
			v_result.f_Y() += CPerlinNoise::f_Noise(v_pos.f_X(),
				v_pos.f_Y(), v_pos.f_Z(), 2.5f) * f_exponent;
			v_result.f_Z() += (CPerlinNoise::f_Noise(v_pos.f_X(),
				v_pos.f_Y(), v_pos.f_Z(), 4.5f) * f_exponent);
			f_rescale += f_exponent; // calculate rescale coeff
		}
		return v_result * (f_amplitude / f_rescale) + Vector3f(f_bias, f_bias, f_bias);
	}

	/*
	 *	template <class CFunction>
	 *	float f_Value(Vector3f v_pos, CFunction function) const
	 *		- calculates noise value at position v_pos
	 *		- perlin noise results are first passed to function and
	 *		  it's results are used in octave sum instead
	 */
	template <class CFunction>
	float f_Value(Vector3f v_pos, CFunction function) const
	{
		if(!n_octave_num)
			return f_bias;

		float f_rescale = 0;
		float f_result = 0, f_exponent = 1;
		v_pos *= f_scale;
		for(int i = 0; i < n_octave_num; ++ i, v_pos *= 2, f_exponent *= f_octave_exponent) {
			f_result += function(CPerlinNoise::f_Noise(v_pos.f_X(),
				v_pos.f_Y(), v_pos.f_Z())) * f_exponent;
			f_rescale += f_exponent;
		}
		return f_result * f_amplitude / f_rescale + f_bias;
	}
};

/*
 *	struct CTaskDescription
 *		- skybox generator task description; consists
 *		  of several tasks to create sky, terrain and clouds
 */
struct CTaskDescription {
public:
	/*
	 *	struct CTaskDescription::TOutputDescription
	 *		- description of output images
	 */
	struct TOutputDescription {
		int n_texture_size; // output resolution
		int n_sample_num; // number of samples per pixel (1 to inf, utilizing repeatable poisson disc patterns)
		int n_ao_sample_num; // number of ambient-occlusion samples
		int n_shadow_sample_num; // number of shadow samples
		float f_shadow_sample_offset; // offset between shadow samples
		bool b_normalize; // normalize colors before saving them?
		float f_gamma; // additional gamma correction (after normalizing)
		float f_bias; // additional biasing (after normalizing)
		float f_scale; // additional multiplication (after normalizing)
		int n_color_system_name; // one of CColorSystem::<unnamed> color system name enums

		/*std::string s_sky_bitmap; // sky color only
		std::string s_horizon_bitmap; // horizon color strip
		std::string s_output_bitmap;*/ // complete skybox
		// depreacated

		/*
		 *	void CTaskDescription::TOutputDescription::Defaults()
		 *		- fills-in default values
		 */
		void Defaults();

		/*
		 *	bool CTaskDescription::TOutputDescription::Parse(CMaterialLexer &r_lexer,
		 *		const char *&r_p_s_error_description)
		 *		- expects current token to be the node name token,
		 *		  upon successful return leaves the end brace token
		 *		- returns true on success, false on failure
		 */
		bool Parse(CMaterialLexer &r_lexer, const char *&r_p_s_error_description);
	};

	/*
	 *	struct CTaskDescription::TSkyDescription
	 *		- description of the sky / atmosphere
	 */
	struct TSkyDescription {
		float f_planet_radius;
		float f_atmosphere_radius;
		Vector3f v_sun_direction;
		float f_u;
		float f_k_mie;
		float f_k_raileigh;
		float f_observer_altitude;

		/*
		 *	void CTaskDescription::TSkyDescription::Defaults()
		 *		- fills-in default values
		 */
		void Defaults();

		/*
		 *	bool CTaskDescription::TSkyDescription::Parse(CMaterialLexer &r_lexer,
		 *		const char *&r_p_s_error_description)
		 *		- expects current token to be the node name token,
		 *		  upon successful return leaves the end brace token
		 *		- returns true on success, false on failure
		 */
		bool Parse(CMaterialLexer &r_lexer, const char *&r_p_s_error_description);
	};

	/*
	 *	struct CTaskDescription::TSkyColor
	 *		- description of some non-physical aspects of sky-color generation
	 */
	struct TSkyColor {
		float n_texture_size; // sky color is pre-calcuated to texture with lover size than result
		float f_y_cutoff; // anything below this won't be calculated and is set to black
		//float f_overshoot; // sky color is automatically equalized to [0, 1] range, this sets overshoot (multiplier)

		/*
		 *	void CTaskDescription::TSkyColor::Defaults()
		 *		- fills-in default values
		 */
		void Defaults();

		/*
		 *	bool CTaskDescription::TSkyColor::Parse(CMaterialLexer &r_lexer,
		 *		const char *&r_p_s_error_description)
		 *		- expects current token to be the node name token,
		 *		  upon successful return leaves the end brace token
		 *		- returns true on success, false on failure
		 */
		bool Parse(CMaterialLexer &r_lexer, const char *&r_p_s_error_description);
	};

	/*
	 *	struct CTaskDescription::THorizonSampling
	 *		- description of horizon color sampling (horizon color is used for terrain fog)
	 */
	struct THorizonSampling {
		TGaussDistribution t_horizon_filter; // horizon convolution kernel
		float f_horizon_filter_cutoff; // to determine kernel size
		float f_min_y, f_max_y; // texcoord range

		/*
		 *	void CTaskDescription::THorizonSampling::Defaults()
		 *		- fills-in default values
		 */
		void Defaults();

		/*
		 *	bool CTaskDescription::THorizonSampling::Parse(CMaterialLexer &r_lexer,
		 *		const char *&r_p_s_error_description)
		 *		- expects current token to be the node name token,
		 *		  upon successful return leaves the end brace token
		 *		- returns true on success, false on failure
		 */
		bool Parse(CMaterialLexer &r_lexer, const char *&r_p_s_error_description);
	};

	/*
	 *	CTaskDescription::<unnamed>
	 *		- noise function names
	 */
	enum {
		func_Plain,
		func_Abs,
		func_Exp,
		func_ExpAbs,
		func_ForceImage // load image from file
	};

	/*
	 *	struct TCloudLayerDescription
	 *		- description of single layer of clouds
	 */
	struct TCloudLayerDescription {
		float f_altitude;
		float f_thickness;
		// clouds position

		TNoiseParams t_perturb;
		TNoiseParams t_noise;
		int n_function;
		// noise function characteristics

		/*
		 *	void CTaskDescription::TCloudLayerDescription::Defaults()
		 *		- fills-in default values
		 */
		void Defaults();

		/*
		 *	bool CTaskDescription::TCloudLayerDescription::Parse(CMaterialLexer &r_lexer,
		 *		const char *&r_p_s_error_description)
		 *		- expects current token to be the node name token,
		 *		  upon successful return leaves the end brace token
		 *		- returns true on success, false on failure
		 */
		bool Parse(CMaterialLexer &r_lexer, const char *&r_p_s_error_description);
	};

	/*
	 *	struct TCloudsDescription
	 *		- layered clouds description
	 */
	struct TCloudsDescription {
		std::vector<TCloudLayerDescription> cloud_layers;

		/*
		 *	void CTaskDescription::TCloudsDescription::Defaults()
		 *		- fills-in default values
		 */
		void Defaults();

		/*
		 *	bool CTaskDescription::TCloudsDescription::Parse(CMaterialLexer &r_lexer,
		 *		const char *&r_p_s_error_description)
		 *		- expects current token to be the node name token,
		 *		  upon successful return leaves the end brace token
		 *		- returns true on success, false on failure
		 */
		bool Parse(CMaterialLexer &r_lexer, const char *&r_p_s_error_description);
	};

	/*
	 *	struct CTaskDescription::TTerrainDescription
	 *		- terrain description
	 */
	struct TTerrainDescription {
		int n_width, n_height;
		float f_tile_size;
		float f_fog_base, f_fog_height, f_fog_thickness;
		float f_ambient, f_diffuse;

		/*
		 *	struct CTaskDescription::TTerrainDescription::TTerrainOp
		 *		- terrain operator
		 */
		struct TTerrainOp {
			int n_type;
			float p_args[5];
			int n_seed;

			/*
			 *	CTaskDescription::TTerrainDescription::TTerrainOp::TTerrainOp(int _n_type,
			 *		float f_arg0 = 0, float f_arg1 = 0, float f_arg2 = 0,
			 *		float f_arg3 = 0, float f_arg4 = 0, int _n_seed = 0)
			 *		- terrain operator default constructor
			 *		- _n_type is one of op_Fill, op_Normalize, ... op_Perlin
			 *		- f_arg0 .. f_arg4 are operator arguments
			 *		- _n_seed is random number generator seed
			 */
			TTerrainOp(int _n_type, float f_arg0 = 0, float f_arg1 = 0,
				float f_arg2 = 0, float f_arg3 = 0, float f_arg4 = 0, int _n_seed = 0);

			/*
			 *	bool CTaskDescription::TTerrainDescription::TTerrainOp::Apply(
			 *		CHeightField &r_target) const
			 *		- apply operator on terrain
			 *		- returns true on success, false on failure (unknown op-type)
			 */
			bool Apply(CHeightField &r_target) const;
		};

		/*
		 *	CTaskDescription::TTerrainDescription::<unnamed>
		 *		- terrain operator names
		 */
		enum {
			op_Fill, // fill (1 param value)
			op_Normalize, // normalize (2 params, min max)
			op_Mean, // mean filter (1 param kernel size)
			op_Exp, // exp(-height) (0 params)
			op_FForm, // fault formation (2 params, no of iters and displacement)
			op_PDepo, // particle deposition (5 params, xy of seed, shift proba, max walk, iters)
			op_Perlin // perlin noise (5 params, octave num, size, amplitude, z, durability)
		};

		std::vector<TTerrainOp> op_list;

		/*
		 *	struct CTaskDescription::TTerrainDescription::TTextureOp
		 *		- texturing operator
		 */
		struct TTextureOp {
			int n_texcoord_style;
			// texture-coordinate mapping style

			/*
			 *	CTaskDescription::TTerrainDescription::TTextureOp::<unnamed>
			 *		- texture mapping style names
			 */
			enum {
				map_XZPlane,
				map_YAxis,
				map_Volume
			};

			TNoiseParams t_perturb;
			TNoiseParams t_noise;
			int n_function;
			// noise function characteristics

			/*
			 *	CTaskDescription::TTerrainDescription::TTextureOp::TPaletteEntry
			 *		- simple piecewise interpolated palette entry structure
			 */
			struct TPaletteEntry {
				float f_pos;
				Vector3f v_color;

				TPaletteEntry()
				{}

				TPaletteEntry(float f)
					:f_pos(f)
				{}

				inline bool operator <(const TPaletteEntry &r_t_entry) const
				{
					return f_pos < r_t_entry.f_pos;
				}
			};

			std::vector<TPaletteEntry> palette;
			// color palette

			TGaussDistribution t_height_matching;
			Vector3f v_ref_normal;
			TGaussDistribution t_slope_matching;
			TNoiseParams t_matching_distortion;
			// texture application

			/*
			 *	void CTaskDescription::TTerrainDescription::TTextureOp::Defaults()
			 *		- fills-in default values
			 */
			void Defaults();

			/*
			 *	bool CTaskDescription::TTerrainDescription::TTextureOp::Parse(
			 *		CMaterialLexer &r_lexer, const char *&r_p_s_error_description)
			 *		- expects current token to be the node name token,
			 *		  upon successful return leaves the end brace token
			 *		- returns true on success, false on failure
			 */
			bool Parse(CMaterialLexer &r_lexer, const char *&r_p_s_error_description);

			/*
			 *	inline float f_Weight(Vector3f v_pos, Vector3f v_normal) const
			 *		- returns weight of this texture layer for point at v_pos with normal v_normal
			 */
			inline float f_Weight(Vector3f v_pos, Vector3f v_normal) const
			{
				return t_height_matching.f_Value(v_pos.f_Y()) +
					t_slope_matching.f_Value(/*acos*/(v_normal.f_Dot(v_ref_normal))) +
					t_matching_distortion.f_Value(v_pos);
			}

			/*
			 *	Vector3f v_Color(Vector3f v_pos) const
			 *		- returns texture color at position v_pos
			 */
			Vector3f v_Color(Vector3f v_pos) const
			{
				v_pos += t_perturb.v_Value(v_pos);
				// offset by perturbation (before mapping!)

				switch(n_texcoord_style) {
				case map_XZPlane:
					v_pos.f_Y() = 0;
					break;
				case map_YAxis:
					v_pos.f_X() = 0;
					v_pos.f_Z() = 0;
					break;
				}
				// texture "mapping" styles

				// todo - fix palette, add multiplicative noise

				float f_intensity = 0;
				switch(n_function) {
				case func_ForceImage:
					return Vector3f(0, 0, 0); // todo
				case func_Plain:
					f_intensity = t_noise.f_Value(v_pos);
					break;
				case func_Abs:
					f_intensity = t_noise.f_Value(v_pos, f_Abs);
					break;
				case func_Exp:
					f_intensity = t_noise.f_Value(v_pos, f_Exp);
					break;
				case func_ExpAbs:
					f_intensity = t_noise.f_Value(v_pos, f_ExpAbs);
					break;
				}
				// calculate intensity

				if(palette.size() >= 2) {
					TPaletteEntry t_e0 = *(std::lower_bound(palette.begin() + 1,
						palette.end(), TPaletteEntry(f_intensity)) - 1);
					TPaletteEntry t_e1 = *std::upper_bound(palette.begin(),
						palette.end() - 1, TPaletteEntry(f_intensity));
					// find two palette components

					f_intensity -= t_e0.f_pos;
					if(t_e1.f_pos - t_e0.f_pos == 0)
						return t_e0.v_color;
					f_intensity /= t_e1.f_pos - t_e0.f_pos;
					// scale intensity so it becomes weight factor

					return t_e0.v_color + (t_e1.v_color - t_e0.v_color) * f_intensity;
					// lerp
				} else if(palette.size() == 1) {
					return palette[0].v_color;
					// single palette entry means constant color
				} else {
					return Vector3f(f_intensity, f_intensity, f_intensity);
					// no palette entries means greyscale
				}
				// final color processing
			}

			static inline float f_Abs(float f_x)
			{
				return fabs(f_x);
			}

			static inline float f_Exp(float f_x)
			{
				return exp(-f_x);
			}

			static inline float f_ExpAbs(float f_x)
			{
				return exp(-fabs(f_x));
			}
		};

		std::vector<TTextureOp> texture_list;

		/*
		 *	void CTaskDescription::TTerrainDescription::Defaults()
		 *		- fills-in default values
		 */
		void Defaults();

		/*
		 *	bool CTaskDescription::TTerrainDescription::Parse(CMaterialLexer &r_lexer,
		 *		const char *&r_p_s_error_description)
		 *		- expects current token to be the node name token,
		 *		  upon successful return leaves the end brace token
		 *		- returns true on success, false on failure
		 */
		bool Parse(CMaterialLexer &r_lexer, const char *&r_p_s_error_description);

		/*
		 *	bool CTaskDescription::TTerrainDescription::ParseGenerator(CMaterialLexer &r_lexer,
		 *		const char *&r_p_s_error_description)
		 *		- expects current token to be the node name token,
		 *		  upon successful return leaves the end brace token
		 *		- returns true on success, false on failure
		 */
		bool ParseGenerator(CMaterialLexer &r_lexer, const char *&r_p_s_error_description);

		/*
		 *	Vector3f CTaskDescription::TTerrainDescription::v_Color(Vector3f v_pos,
		 *		Vector3f v_normal) const
		 *		- returns (unshaded) texture color at position v_pos and normal v_normal
		 */
		Vector3f v_Color(Vector3f v_pos, Vector3f v_normal) const
		{
			float f_weight_sum = 0;
			Vector3f v_color(0, 0, 0);

			for(int i = 0, n = texture_list.size(); i < n; ++ i) {
				float f_weight = texture_list[i].f_Weight(v_pos, v_normal);
				f_weight_sum += f_weight;
				v_color += texture_list[i].v_Color(v_pos) * f_weight;
			}

			if(!f_weight_sum)
				return Vector3f(1, 1, 1);
			return v_color / f_weight_sum;
		}
	};

protected:
	TOutputDescription m_output;
	TSkyDescription m_sky;
	TSkyColor m_sky_color;
	TCloudsDescription m_clouds;
	THorizonSampling m_horizon;
	TTerrainDescription m_terrain;

public:
	/*
	 *	CTaskDescription::CTaskDescription()
	 *		- default constructor
	 *		- fills-in default values
	 */
	CTaskDescription();

	/*
	 *	void CTaskDescription::Defaults()
	 *		- fills-in default values
	 */
	void Defaults();

	/*
	 *	bool CTaskDescription::Parse(const char *p_s_file, const char *&r_p_s_error_description,
	 *		int &r_n_error_line, int &r_n_error_column)
	 *		- parses task description from file p_s_file
	 *		- r_p_s_error_description is filled with error description,
	 *		  r_n_error_line and r_n_error_column are filled with error
	 *		  line and columnt zero-based index (if error occured, otherwise
	 *		  r_p_s_error_description is 0 and values of r_n_error_line and
	 *		  r_n_error_column are undefined)
	 *		- note this automatically resets all data to defaults before actual parsing
	 *		- returns true on success, false on failure
	 */
	bool Parse(const char *p_s_file, const char *&r_p_s_error_description,
		int &r_n_error_line, int &r_n_error_column);

	/*
	 *	inline const TOutputDescription &CTaskDescription::r_Output() const
	 *		- returns output description data
	 */
	inline const TOutputDescription &r_Output() const
	{
		return m_output;
	}

	/*
	 *	inline const TSkyDescription &CTaskDescription::r_Atmosphere() const
	 *		- returns atmosphere / sky description data
	 */
	inline const TSkyDescription &r_Atmosphere() const
	{
		return m_sky;
	}

	/*
	 *	inline const TSkyColor &CTaskDescription::r_SkyColor() const
	 *		- returns non-physical sky color processing description data
	 */
	inline const TSkyColor &r_SkyColor() const
	{
		return m_sky_color;
	}

	/*
	 *	inline const THorizonSampling &CTaskDescription::r_HorizonSampling() const
	 *		- returns horizon sampling description data
	 */
	inline const THorizonSampling &r_HorizonSampling() const
	{
		return m_horizon;
	}

	/*
	 *	inline const TCloudsDescription &CTaskDescription::r_Clouds() const
	 *		- returns clouds description data
	 */
	inline const TCloudsDescription &r_Clouds() const
	{
		return m_clouds;
	}

	/*
	 *	inline const TTerrainDescription &CTaskDescription::r_Terrain() const
	 *		- returns terrain description data
	 */
	inline const TTerrainDescription &r_Terrain() const
	{
		return m_terrain;
	}

protected:
	bool Parse_Loop(CMaterialLexer &r_lexer, const char *&r_p_s_error_description);
};

#endif //__TASK_INCLUDED
