/*
								+-----------------------------------+
								|                                   |
								|    ***  Terrain generator  ***    |
								|                                   |
								|    Design by Lukas Polok 2007     |
								|    Implemented by Jiri Hanus      |
								|                                   |
								|            Terrain.cpp            |
								|                                   |
								+-----------------------------------+
*/

#ifdef WIN32
#include <crtdbg.h>
#else
#include <assert.h>
#define _ASSERTE(x) assert(x)
#endif
#include <stdlib.h>
#include <math.h>
#include "MiniVector.h"
#include "Terrain.h"
#include "PNoise.h"

#if defined(_MSC_VER) && !defined(_MWERKS)
#define for if(0) {} else for
#endif
// msvc 'for' scoping hack

/*
 *								=== CHeightField ===
 */

/*
 *	CHeightField::CHeightField(int n_width, int n_height)
 *		- default constructor, prepares heightfield with
 *		  size n_width per n_height pixels and unspecified contents
 *		- call b_Status() to determine wheter constructor succeeded allocating memory
 *
 *		- zadn
 *			- vytvo vkovou mapu o rozmrech n_width x n_height s pedem neurenm obsahem
 *			- dnou chybovou hlku pi nedostatku pamti nevypisuje, ale mus bt schopn
 *			  njak oznait chybov stav, aby funkce b_Status() mohla vrtit false
 *			- ostatn operace (krom b_Status a destuktoru) nemus uvaovat nezdaenou alokaci
 *
 */
CHeightField::CHeightField(int n_width, int n_height)
	:m_n_width(n_width), m_n_height(n_height), m_p_data(new float[n_width * n_height])
{}

/*
 *	CHeightField::~CHeightField()
 *		- default destructor
 *
 *		- zadn
 *			- uvoln naalokovanou pam (pokud pi alokaci nedolo k chyb)
 *
 */
CHeightField::~CHeightField()
{
	if(m_p_data)
		delete[] m_p_data;
}

/*
 *	bool CHeightField::b_Status() const
 *		- if constructor succeeded returns true, otherwise returns false
 *
 *		- zadn
 *			- pedv vsledek kontruktoru. true je spch, false je nespch (nedostatek ram)
 *
 */
bool CHeightField::b_Status() const
{
	return m_p_data != 0;
}

/*
 *	void CHeightField::Fill(float f_value = 0)
 *		- fills terrain with constant value f_value
 *
 *		- zadn
 *			- vypln celou heightmapu zadanou konstantn hodnotou
 *
 */
void CHeightField::Fill(float f_value)
{
	for(float *p_data = m_p_data, *p_end = m_p_data +
	   (m_n_width * m_n_height); p_data != p_end; ++ p_data)
		*p_data = f_value;
}

/*
 *	void CHeightField::Normalize(float f_min_height = 0, float f_max_height = 1)
 *		- scales heightfield so minimum is f_min_height and maximum is f_max_height
 *		- note in case heightfield is flat, it will be positioned to f_min_height
 *
 *		- zadn
 *			- uprav hodnoty heightmapy tak, aby minimum bylo f_min_height a maximum bylo f_max_height
 *			- co kdy heightmapa obsahuje konstantn hodnotu (potom min = max)?
 *
 */
void CHeightField::Normalize(float f_min_height, float f_max_height)
{
	float f_min = *m_p_data, f_max = *m_p_data;
	for(const float *p_data = m_p_data + 1, *p_end = m_p_data +
	   (m_n_width * m_n_height); p_data != p_end; ++ p_data) {
		float f_data = *p_data;
		if(f_min > f_data)
			f_min = f_data;
		if(f_max < f_data)
			f_max = f_data;
	}
	// determine min / max

	f_max -= f_min;
	if(f_max < 1e-3f)
		f_max_height = 0; // position flat heightfield at f_min_height
	else {
		f_max_height -= f_min_height;
		f_max_height /= f_max;
		f_min_height -= f_min * f_max_height;
	}
	// calculate bias and scale

	for(float *p_data = m_p_data, *p_end = m_p_data +
	   (m_n_width * m_n_height); p_data != p_end; ++ p_data)
		*p_data = *p_data * f_max_height + f_min_height;
	// normalize
}

/*
 *	bool CHeightField::Mean(int n_kernel_size)
 *		- applies mean filter on heightfield, n_kernel_size is size
 *		  of mean window; window is centered above output pixel
 *		- needs auxiliary buffer, returns false in case there
 *		  is not enough memory for it, otherwise returns true
 *
 *		- zadn
 *			- aplikovat na vzorky heightmapy filtr typu prmr pod oknem (tedy spot prmr pixel
 *			  pod tvercem o hran n_kernel_size a vsledek dosad do pixelu pod sted tverce)
 *			- pro efektivnost operaci provede naped vodorovn a pot svisle (tedy na dva prchody)
 *			- je poteba pomocn buffer na mezivsledky, pokud jeho alokace sele, vrac false,
 *			  jinak true (nevypisuje dn chybov hlen)
 *
 */
bool CHeightField::Mean(int n_kernel_size)
{
	CHeightField temp(m_n_width, m_n_height);
	if(!temp.b_Status())
		return false;
	// need temporary field to do this

	Mean_Separable(n_kernel_size, 1, m_n_width,
		m_n_width, m_n_height, m_p_data, temp.m_p_data);
	Mean_Separable(n_kernel_size, m_n_width, 1,
		m_n_height, m_n_width, temp.m_p_data, m_p_data);
	// apply separate mean filters first for scanlines, then for columns

	return true;
}

/*
 *	void CHeightField::FaultFormation(int n_iteration_num, float f_step_size)
 *		- draws n_iteration_num iterations of fault formation
 *		- f_step_size is size of fault formation step, in each step it is applied
 *		  both positively and negatively so the result has mean arround 0
 *
 *		- zadn
 *			- simulace pohybu zemskch desek;  n_iteration_num je poet iterac a
 *			  f_step_size je zven / snen ternu. jeden prchod fault formation
 *			  nhodn ur ru, prochzejc pes heightmapu a vechny vzorky nalevo sn,
 *			  kdeto vechny vzorky napravo zv (-> urduje se prmrn hodnota)
 *			- vce o fault formation teba na:
 *			  http://collective.valve-erc.com/index.php?doc=1040644009-23694100
 *
 */
void CHeightField::FaultFormation(int n_iteration_num, float f_step_size)
{
	for(; n_iteration_num; -- n_iteration_num) {
		int n_x = rand() % m_n_width;
		int n_y = rand() % m_n_height;
		float f_normal_a = 1 - (rand() / (float)(RAND_MAX / 2));
		float f_normal_b = 1 - (rand() / (float)(RAND_MAX / 2));
		float f_normal_c = -f_normal_a * n_x - f_normal_b * n_y;
		// determines division normal

		float f_step = (f_normal_a < 0)? -f_step_size : f_step_size;
		// turn step in case normal's x is negative

		float *p_dest = m_p_data;
		for(int y = 0, m = m_n_height, n = m_n_width; y < m; ++ y) {
			float f_division_x = (f_normal_b * y + f_normal_c) / -f_normal_a;
			int n_division_x = (f_division_x >= 0)?
				((f_division_x < n)? (int)f_division_x : n - 1) : 0;
			// determine x where plane intersects the scanline (have to recalc
			// equation every time because of precission / large number issues
			// when plane is too steep)

			if(n_division_x) {
				for(int x = 0; x <= n_division_x; ++ x, ++ p_dest)
					*p_dest -= f_step;
				for(int x = n_division_x + 1; x < n; ++ x, ++ p_dest)
					*p_dest += f_step;
			} else {
				for(int x = 0; x < n_division_x; ++ x, ++ p_dest)
					*p_dest -= f_step;
				for(int x = n_division_x; x < n; ++ x, ++ p_dest)
					*p_dest += f_step;
			}
			// offset pixels
		}
	}
}

/*
 *	void CHeightField::ParticleDeposition(int n_drop_x, int n_drop_y, float f_move_proba,
 *		int n_max_walk, int n_drop_num)
 *		- generates terrain using particle deposition
 *		- n_drop_x and n_drop_y are initial positions of particle drop-point
 *		- f_move_proba is drop-point movement probability (relative, per iteration)
 *		- n_max_walk is maximal number of particle stabilization walk steps
 *		- n_drop_num is number of iterations
 *
 *		- zadn
 *			- vytven nhodnho ternu pomoc sticovho systmu
 *			- stice se sypou z njak pozice na tern a zvyuj jej tak. pokud je tern v mst
 *			  dopadu pli strm, stice se 'kutl' dokud nedoshne stabiln polohu.
 *			- pozice sypn je n_drop_x, n_drop_y. f_move_proba je pravdpodobnost (0 - 1) e
 *			  po vysypn stice se pozice sypn zmn (zmna je potom ist nhodn, pozor
 *			  na pekroen hranic ternu!)
 *			- n_max_walk omezuje poet krok kutlen stice (kvli rychlosti)
 *			- n_drop_num je celkov poet vysypanch stic
 *			- nco o particle deposition je na:
 *			  http://www.lighthouse3d.com/opengl/terrain/index.php3?particle
 */
void CHeightField::ParticleDeposition(int n_drop_x, int n_drop_y, float f_move_proba,
	int n_max_walk, int n_drop_num)
{
	if(n_drop_x < 0)
		n_drop_x = 0;
	else if(n_drop_x >= m_n_width)
		n_drop_x = m_n_width - 1;
	if(n_drop_y < 0)
		n_drop_y = 0;
	else if(n_drop_y >= m_n_height)
		n_drop_y = m_n_height - 1;
	// check droppoint corectness

	for(; n_drop_num; -- n_drop_num) {
		if(rand() <= RAND_MAX * f_move_proba) {
			if(rand() & 1) {
				n_drop_x += 1 - 2 * (rand() & 1);
				if(n_drop_x < 0)
					n_drop_x = 0;
				else if(n_drop_x >= m_n_width)
					n_drop_x = m_n_width - 1;
			} else {
				n_drop_y += 1 - 2 * (rand() & 1);
				if(n_drop_y < 0)
					n_drop_y = 0;
				else if(n_drop_y >= m_n_height)
					n_drop_y = m_n_height - 1;
			}
		}
		// move the emitter

		int x = n_drop_x;
		int y = n_drop_y;
		for(int n_walk = 0; n_walk < n_max_walk; ++ n_walk) {
			int x1 = (x < m_n_width - 1)? x + 1 : m_n_width - 1;
			int y1 = (y < m_n_height - 1)? y + 1 : m_n_height - 1;
			int x0 = (x)? x - 1 : 0;
			int y0 = (y)? y - 1 : 0;

			float f_laplacian = m_p_data[x0 + y0 * m_n_width] +
								m_p_data[x  + y0 * m_n_width] +
								m_p_data[x1 + y0 * m_n_width] +
								m_p_data[x0 + y  * m_n_width] +
								m_p_data[x1 + y  * m_n_width] +
								m_p_data[x0 + y1 * m_n_width] +
								m_p_data[x  + y1 * m_n_width] +
								m_p_data[x1 + y1 * m_n_width] +
								m_p_data[x  + y  * m_n_width] * -8;
			// calculate laplacian

			if(f_laplacian < 0) {
				// have to move

				float f_sobel_x = (m_p_data[x0 + y0 * m_n_width] +
								   m_p_data[x0 + y  * m_n_width] * 2 +
								   m_p_data[x0 + y1 * m_n_width]) -
								  (m_p_data[x1 + y0 * m_n_width] +
								   m_p_data[x1 + y  * m_n_width] * 2 +
								   m_p_data[x1 + y1 * m_n_width]);
				float f_sobel_y = (m_p_data[x0 + y0 * m_n_width] +
								   m_p_data[x  + y0 * m_n_width] * 2 +
								   m_p_data[x1 + y0 * m_n_width]) -
								  (m_p_data[x0 + y1 * m_n_width] +
								   m_p_data[x  + y1 * m_n_width] * 2 +
								   m_p_data[x1 + y1 * m_n_width]);
				// calculate x and y sobel to decide where to move
				// subtracting 'left' - 'right' -> positive means right is lower -> roll right

				if(fabs(f_sobel_x) > fabs(f_sobel_y)) {
					if(f_sobel_x > 0) {
						if(x < m_n_width - 1)
							++ x;
					} else if(x > 0)
						-- x;
				} else {
					if(f_sobel_y > 0) {
						if(y < m_n_height - 1)
							++ y;
					} else if(y > 0)
						-- y;
				}
				// roll the particle 'downhill'
			} else {
				++ m_p_data[x + y * m_n_width];
				// particle had settled

				break;
			}
		}
	}
}

/*
 *	void CHeightField::PerlinNoise(int n_octave_num, float f_size, float f_amplitude,
 *		float f_z, float f_durability)
 *		- adds perlin noise with n_octave_num octaves, f_size is spatial acaling
 *		  ratio, f_amplitude is noise amplitude multiplier, f_z is third noise function
 *		  coordinate (like random seed) and f_durability controls power of each octave
 *		  using equation: octave(i) = noise(x, y, z) * pow(f_durability, i)
 *
 *		- zadn
 *			- pite k hodnotm heightmapy fraktln perlinv um. n_octave_num je poet
 *			  oktv, f_size je velikost umu (klovn pozice), f_amplitude je amplituda
 *			  umu, f_z je tet souadnice pro umovou funkci (heightmapa je dvojrozmrn),
 *			  f_durability uruje jak siln jsou jednotliv oktvy
 *			- vce o perlinov umu na:
 *			  http://www.noisemachine.com/talk1/
 *			- implementace je dostupn v PNoise.h
 *
 */
void CHeightField::PerlinNoise(int n_octave_num, float f_size, float f_amplitude,
	float f_z, float f_durability)
{
	float *p_data = m_p_data;
	for(int y = 0, m = m_n_height; y < m; ++ y) {
		for(int x = 0, n = m_n_width; x < n; ++ x, ++ p_data) {
			float f_noise = 0;
			for(int i = 0; i < n_octave_num; ++ i) {
				f_noise += (float)pow(f_durability, i) * CPerlinNoise::f_Noise(
					x * (1 << i) * f_size, y * (1 << i) * f_size, f_z);
			}
			*p_data += f_noise * f_amplitude;
		}
	}
}

/*
 *	inline void CHeightField::PassNormal(int x, int y, float f_scale) const
 *		- calculates normal over point x, y in heightmap
 *		- f_scale is size of side of one pixel quad in worldspace
 *		- passes result to OpenGL using glNormal3f
 */
/*inline void CHeightField::PassNormal(int x, int y, float f_scale) const
{
	int x1 = (x < m_n_width - 1)? x + 1 : m_n_width - 1;
	int y1 = (y < m_n_height - 1)? y + 1 : m_n_height - 1;
	int x0 = (x)? x - 1 : 0;
	int y0 = (y)? y - 1 : 0;
	float f_dx = m_p_data[x0 + y * m_n_width] - m_p_data[x1 + y * m_n_width];
	float f_dy = m_p_data[x + y0 * m_n_width] - m_p_data[x + y1 * m_n_width];
	float f_dz = f_scale;
	float f_size = 1.0f / (float)sqrt(f_dx * f_dx + f_dy * f_dy + f_dz * f_dz);
	//glNormal3f(f_dx * f_size, f_dz * f_size, f_dy * f_size);
}*/

/*
 *	static void CHeightField::Mean_Separable(int n_kernel_size, int n_pixel_shift,
 *		int n_scan_shift, int n_width, int n_height, const float *p_src, float *p_dest)
 *		- calculates one pass of separable mean filter (ie. filters in one dimension)
 *		- n_kernel_size is size of filtering kernel
 *		- n_pixel_shift is number of items in array to skip to get to next pixel
 *		- n_scan_shift is number of items in array to skip to get to next scanline
 *		- n_width and n_height are heightmap dimensions
 *		- p_src is source heightmap, p_dest is destination heightmap
 */
void CHeightField::Mean_Separable(int n_kernel_size, int n_pixel_shift,
	int n_scan_shift, int n_width, int n_height, const float *p_src, float *p_dest)
{
	int n_kernel_half = n_kernel_size / 2;
	int n_kernel_half2 = n_kernel_size - n_kernel_half;
	float f_scale = 1.0f / n_kernel_size;
	// it's propably cheaper to calculate these than put them on the stack

	for(int y = 0; y < n_height; ++ y, p_src += n_scan_shift, p_dest += n_scan_shift) {
		float f_left_border = *p_src;
		float f_right_border = p_src[((n_width - 1) * n_pixel_shift)];
		// left and right border heights

		float f_accumulator = n_kernel_half * f_left_border;
		for(int x = 0; x < n_kernel_half2; ++ x)
			f_accumulator += (x < n_width)? p_src[x * n_pixel_shift] : f_right_border;
		// eat-in the accumulator for the first pixel

		for(int x = 0; x < n_width; ++ x) {
			p_dest[x * n_pixel_shift] = f_accumulator * f_scale;
			f_accumulator -= (x - n_kernel_half >= 0)?
				p_src[(x - n_kernel_half) * n_pixel_shift] : f_left_border;
			f_accumulator += (x + n_kernel_half2 < n_width)?
				p_src[(x + n_kernel_half2) * n_pixel_shift] : f_right_border;
		}
		// somewhat unoptimized

		/*int n_headpsace = max(0, n_width - n_kernel_half2);
		for(int x = 0; x < min(n_kernel_half, n_headpsace); ++ x) {
			p_dest[x * n_pixel_shift] = f_accumulator * f_scale;
			f_accumulator -= f_left_border;
			f_accumulator += p_src[(x + n_kernel_half2) * n_pixel_shift];
		}
		// move kernel until left border is shifted out

		if(n_width < n_kernel_half) {
			for(int x = n_headpsace; x < min(n_width, n_kernel_half); ++ x) {
				p_dest[x * n_pixel_shift] = f_accumulator * f_scale;
				f_accumulator -= f_left_border;
				f_accumulator += f_right_border;
			}
			// move kernel until left border is shifted out, scanline is shorter than kernel
		} else {
			for(int so = 0, x = n_kernel_half, si = n_kernel_size * n_pixel_shift,
			   n = n_width - n_kernel_half2; x < n; so += n_pixel_shift, ++ x, si += n_pixel_shift) {
				p_dest[x * n_pixel_shift] = f_accumulator * f_scale;
				f_accumulator -= p_src[so];
				f_accumulator += p_src[si];
			}
			// move kernel until reaching right border
			// (the 'middle' of scanline where no kernel pixel is clipped)

			for(int so = (n_width > n_kernel_size)? (n_width - n_kernel_size) * n_pixel_shift : 0,
			   x = (n_width > n_kernel_size)? n_width - n_kernel_half2 : n_kernel_half;
			   x < n_width; so += n_pixel_shift, ++ x) {
				p_dest[x * n_pixel_shift] = f_accumulator * f_scale;
				f_accumulator -= p_src[so];
				f_accumulator += f_right_border;
			}
			// move kernel past right border until the oputput is filled
		}*/
		// buggy :(
	}
}

/*
 *								=== ~CHeightField ===
 */
