/*
								+-----------------------------------+
								|                                   |
								|    ***  Terrain generator  ***    |
								|                                   |
								|    Design by Lukas Polok 2007     |
								|    Implemented by Jiri Hanus      |
								|                                   |
								|             Terrain.h             |
								|                                   |
								+-----------------------------------+
*/

#ifndef __TERRAIN_GENERATOR_INCLUDED
#define __TERRAIN_GENERATOR_INCLUDED

/*
 *	class CHeightField
 *		- heightfield class with some simple operators
 */
class CHeightField {
protected:
	int m_n_width, m_n_height;
	float *m_p_data;

public:
	/*
	 *	CHeightField::CHeightField(int n_width, int n_height)
	 *		- default constructor, prepares heightfield with
	 *		  size n_width per n_height pixels and unspecified contents
	 *		- call b_Status() to determine wheter constructor succeeded allocating memory
	 *
	 *		- zadn
	 *			- vytvo vkovou mapu o rozmrech n_width x n_height s pedem neurenm obsahem
	 *			- dnou chybovou hlku pi nedostatku pamti nevypisuje, ale mus bt schopn
	 *			  njak oznait chybov stav, aby funkce b_Status() mohla vrtit false
	 *			- ostatn operace (krom b_Status a destuktoru) nemus uvaovat nezdaenou alokaci
	 *
	 */
	CHeightField(int n_width, int n_height);

	/*
	 *	CHeightField::~CHeightField()
	 *		- default destructor
	 *
	 *		- zadn
	 *			- uvoln naalokovanou pam (pokud pi alokaci nedolo k chyb)
	 *
	 */
	~CHeightField();

	/*
	 *	bool CHeightField::b_Status() const
	 *		- if constructor succeeded returns true, otherwise returns false
	 *
	 *		- zadn
	 *			- pedv vsledek kontruktoru. true je spch, false je nespch (nedostatek ram)
	 *
	 */
	bool b_Status() const;

	/*
	 *	inline int CHeightField::n_Width() const
	 *		- returns heightfield width
	 */
	inline int n_Width() const
	{
		return m_n_width;
	}

	/*
	 *	inline int CHeightField::n_Height() const
	 *		- returns heightfield width
	 */
	inline int n_Height() const
	{
		return m_n_height;
	}

	/*
	 *	inline const float *CHeightField::operator [](int n_y) const
	 *		- returns const pointer to scanline with zero based index n_y
	 */
	inline const float *operator [](int n_y) const
	{
		return m_p_data + (n_y * m_n_width);
	}

	/*
	 *	inline float *CHeightField::operator [](int n_y)
	 *		- returns pointer to scanline with zero based index n_y
	 */
	inline float *operator [](int n_y)
	{
		return m_p_data + (n_y * m_n_width);
	}

	/*
	 *	void CHeightField::Fill(float f_value = 0)
	 *		- fills terrain with constant value f_value
	 *
	 *		- zadn
	 *			- vypln celou heightmapu zadanou konstantn hodnotou
	 *
	 */
	void Fill(float f_value = 0);

	/*
	 *	void CHeightField::Normalize(float f_min_height = 0, float f_max_height = 1)
	 *		- scales heightfield so minimum is f_min_height and maximum is f_max_height
	 *		- note in case heightfield is flat, it will be positioned to f_min_height
	 *
	 *		- zadn
	 *			- uprav hodnoty heightmapy tak, aby minimum bylo f_min_height a maximum bylo f_max_height
	 *			- co kdy heightmapa obsahuje konstantn hodnotu (potom min = max)?
	 *
	 */
	void Normalize(float f_min_height = 0, float f_max_height = 1);

	/*
	 *	bool CHeightField::Mean(int n_kernel_size)
	 *		- applies mean filter on heightfield, n_kernel_size is size
	 *		  of mean window; window is centered above output pixel
	 *		- needs auxiliary buffer, returns false in case there
	 *		  is not enough memory for it, otherwise returns true
	 *
	 *		- zadn
	 *			- aplikovat na vzorky heightmapy filtr typu prmr pod oknem (tedy spot prmr pixel
	 *			  pod tvercem o hran n_kernel_size a vsledek dosad do pixelu pod sted tverce)
	 *			- pro efektivnost operaci provede naped vodorovn a pot svisle (tedy na dva prchody)
	 *			- je poteba pomocn buffer na mezivsledky, pokud jeho alokace sele, vrac false,
	 *			  jinak true (nevypisuje dn chybov hlen)
	 *
	 */
	bool Mean(int n_kernel_size);

	/*
	 *	void CHeightField::FaultFormation(int n_iteration_num, float f_step_size)
	 *		- draws n_iteration_num iterations of fault formation
	 *		- f_step_size is size of fault formation step, in each step it is applied
	 *		  both positively and negatively so the result has mean arround 0
	 *
	 *		- zadn
	 *			- simulace pohybu zemskch desek;  n_iteration_num je poet iterac a
	 *			  f_step_size je zven / snen ternu. jeden prchod fault formation
	 *			  nhodn ur ru, prochzejc pes heightmapu a vechny vzorky nalevo sn,
	 *			  kdeto vechny vzorky napravo zv (-> urduje se prmrn hodnota)
	 *			- vce o fault formation teba na:
	 *			  http://collective.valve-erc.com/index.php?doc=1040644009-23694100
	 *
	 */
	void FaultFormation(int n_iteration_num, float f_step_size);

	/*
	 *	void CHeightField::ParticleDeposition(int n_drop_x, int n_drop_y, float f_move_proba,
	 *		int n_max_walk, int n_drop_num)
	 *		- generates terrain using particle deposition
	 *		- n_drop_x and n_drop_y are initial positions of particle drop-point
	 *		- f_move_proba is drop-point movement probability (relative, per iteration)
	 *		- n_max_walk is maximal number of particle stabilization walk steps
	 *		- n_drop_num is number of iterations
	 *
	 *		- zadn
	 *			- vytven nhodnho ternu pomoc sticovho systmu
	 *			- stice se sypou z njak pozice na tern a zvyuj jej tak. pokud je tern v mst
	 *			  dopadu pli strm, stice se 'kutl' dokud nedoshne stabiln polohu.
	 *			- pozice sypn je n_drop_x, n_drop_y. f_move_proba je pravdpodobnost (0 - 1) e
	 *			  po vysypn stice se pozice sypn zmn (zmna je potom ist nhodn, pozor
	 *			  na pekroen hranic ternu!)
	 *			- n_max_walk omezuje poet krok kutlen stice (kvli rychlosti)
	 *			- n_drop_num je celkov poet vysypanch stic
	 *			- nco o particle deposition je na:
	 *			  http://www.lighthouse3d.com/opengl/terrain/index.php3?particle
	 *
	 */
	void ParticleDeposition(int n_drop_x, int n_drop_y, float f_move_proba,
		int n_max_walk, int n_drop_num);

	/*
	 *	void CHeightField::PerlinNoise(int n_octave_num, float f_size, float f_amplitude,
	 *		float f_z, float f_durability)
	 *		- adds perlin noise with n_octave_num octaves, f_size is spatial acaling
	 *		  ratio, f_amplitude is noise amplitude multiplier, f_z is third noise function
	 *		  coordinate (like random seed) and f_durability controls power of each octave
	 *		  using equation: octave(i) = noise(x, y, z) * pow(f_durability, i)
	 *
	 *		- zadn
	 *			- pite k hodnotm heightmapy fraktln perlinv um. n_octave_num je poet
	 *			  oktv, f_size je velikost umu (klovn pozice), f_amplitude je amplituda
	 *			  umu, f_z je tet souadnice pro umovou funkci (heightmapa je dvojrozmrn),
	 *			  f_durability uruje jak siln jsou jednotliv oktvy
	 *			- vce o perlinov umu na:
	 *			  http://www.noisemachine.com/talk1/
	 *			- implementace je dostupn v PNoise.h
	 *
	 */
	void PerlinNoise(int n_octave_num, float f_size, float f_amplitude,
		float f_z, float f_durability);

protected:
	static void Mean_Separable(int n_kernel_size, int n_pixel_shift,
		int n_scan_shift, int n_width, int n_height, const float *p_src, float *p_dest);
};

#endif //__TERRAIN_GENERATOR_INCLUDED
