/*
								+----------------------------------+
								|                                  |
								|      ***  Terrain mesh  ***      |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|          TerrainMesh.cpp         |
								|                                  |
								+----------------------------------+
*/

#ifdef WIN32
#include <crtdbg.h>
#else
#include <assert.h>
#define _ASSERTE(x) assert(x)
#endif
#include <math.h>
#include <vector>
#include "MiniVector.h"
#include "Terrain.h"
#include "TerrainMesh.h"

#if defined(_MSC_VER) && !defined(_MWERKS)
#define for if(0) {} else for
#endif
// msvc 'for' scoping hack

/*
 *								=== CUniformGrid ===
 */

/*
 *	CUniformGrid::CUniformGrid(int n_target_cell_num,
 *		int n_triangle_num, const TTriangle *p_triangle)
 *		- default constructor
 *		- creates a new uniform grid with n_target_cell_num (or propably more) cells
 *		  from mesh with n_triangle_num triangles in p_triangle array
 *		- call b_Status() to determine wheter constructor succeeded
 */
CUniformGrid::CUniformGrid(int n_target_cell_num, int n_triangle_num, const TTriangle *p_triangle)
	:m_p_grid_data(0), m_p_cell_pointer_list(0)
{
	BoundingBox(p_triangle, p_triangle + n_triangle_num, m_v_min, m_v_max);
	Vector3f v_tmp = m_v_max - m_v_min;
	float f_volume = v_tmp.f_X() * v_tmp.f_Y() * v_tmp.f_Z();
	float f_cell_side = (float)pow(f_volume / n_target_cell_num, 1.0f / 3);
	if(f_cell_side < 1e-2f)
		f_cell_side = 1e-2f; // not so small
	for(int n = 0; n < 3; ++ n) {
		if(!(m_p_grid_size[n] = (int)ceil(v_tmp[n] / f_cell_side)))
			m_p_grid_size[n] = 1;
		m_v_cell_size[n] = v_tmp[n] / m_p_grid_size[n];
	}
	// determine dimensions of the grid

	int n_cell_num = m_p_grid_size[0] * m_p_grid_size[1] * m_p_grid_size[2];
	if(!(m_p_cell_pointer_list = new int*[n_cell_num]))
		return;
	// alloc pointers for individual cells

	std::vector<int> cell_size_list;
	cell_size_list.reserve(n_cell_num);
	if(cell_size_list.capacity() < (unsigned)n_cell_num)
		return;
	cell_size_list.insert(cell_size_list.end(), n_cell_num, 0);
	// cell sizes

	int n_grid_data_size = n_cell_num;
	// start with number of cells (there are triangle counts in the array)

	for(int n_pass = 0; n_pass < 2; ++ n_pass) {
		for(int i = 0; i < n_triangle_num; ++ i) {
			Vector3f v_min, v_max;
			BoundingBox(p_triangle + i, p_triangle + i + 1, v_min, v_max);
			// calculate bounding box of i-th triangle

			v_min -= m_v_min;
			v_max -= m_v_min;
			// shift to grid origin

			v_min /= m_v_cell_size;
			v_max /= m_v_cell_size;
			// scale so coordinates are in cells

			int p_min_cell[3];
			int p_max_cell[3];
			for(int n = 0; n < 3; ++ n) {
				p_min_cell[n] = (int)(v_min[n] - .01f);
				if(p_min_cell[n] < 0)
					p_min_cell[n] = 0;
				else if(p_min_cell[n] >= m_p_grid_size[n])
					p_min_cell[n] = m_p_grid_size[n] - 1;

				p_max_cell[n] = (int)(v_max[n] + 1.01f);
				if(p_max_cell[n] < 0)
					p_max_cell[n] = 0;
				else if(p_max_cell[n] >= m_p_grid_size[n])
					p_max_cell[n] = m_p_grid_size[n] - 1;
				++ p_max_cell[n]; // max is supposed be one beyond
			}
			// calculate indices of occupied cells

			if(!n_pass) {
				n_grid_data_size += (p_max_cell[0] - p_min_cell[0]) *
					(p_max_cell[1] - p_min_cell[1]) * (p_max_cell[2] - p_min_cell[2]);
			}
			// sum up number of cells

			for(int z = p_min_cell[2]; z < p_max_cell[2]; ++ z) {
				for(int y = p_min_cell[1]; y < p_max_cell[1]; ++ y) {
					for(int x = p_min_cell[0]; x < p_max_cell[0]; ++ x) {
						int n_cell_index = x + (y + z * m_p_grid_size[1]) * m_p_grid_size[0];
						if(!n_pass) {
							++ cell_size_list[n_cell_index];
							// just sum up triangles in this cell
						} else {
							(m_p_cell_pointer_list[n_cell_index])
								[++ cell_size_list[n_cell_index]] = i;
							// add triangle to the proper cell
						}
					}
				}
			}
			// add triangle index to all cells
		}

		if(!n_pass) {
			if(!(m_p_grid_data = new int[n_grid_data_size]))
				return;
			// alloc grid data

			int *p_pointer = m_p_grid_data;
			for(int i = 0; i < n_cell_num; ++ i) {
				m_p_cell_pointer_list[i] = p_pointer;
				int n = cell_size_list[i];
				cell_size_list[i] = 0;
				*p_pointer = n;
				p_pointer += n + 1;
			}
			_ASSERTE(p_pointer == m_p_grid_data + n_grid_data_size);
			// set pointer to grid cells, write triangle counts
			// and reset cell sizes, they're going to be used
			// as offsets to cell index arrays in the next pass
		}
	}
	// in two passes firstly count neccessary cell array lengths,
	// secondly alloc them and fill them
}

/*
 *	CUniformGrid::~CUniformGrid()
 *		- default destructor
 */
CUniformGrid::~CUniformGrid()
{
	if(m_p_grid_data)
		delete[] m_p_grid_data;
	if(m_p_cell_pointer_list)
		delete[] m_p_cell_pointer_list;
}

/*
 *	bool CUniformGrid::b_Status() const
 *		- returns true in case constructor succeeded, otherwise returns false
 */
bool CUniformGrid::b_Status() const
{
	return m_p_grid_data && m_p_cell_pointer_list;
}

/*
 *	static Vector3<int> CUniformGrid::v_WorldToVoxel(Vector3f v_world,
 *		Vector3f v_min, Vector3f v_cell_size)
 *		- converts from world coordinates v_world to voxel coordinates
 */
Vector3<int> CUniformGrid::v_WorldToVoxel(Vector3f v_world, Vector3f v_min, Vector3f v_cell_size)
{
	Vector3f v_tmp((v_world - v_min) / v_cell_size);
	return Vector3<int>((int)v_tmp.f_X(), (int)v_tmp.f_Y(), (int)v_tmp.f_Z());
}

/*
 *	static bool CUniformGrid::b_Intersect(Vector3f v_org, Vector3f v_dir,
 *		Vector3f v_min, Vector3f v_max, float &r_f_t_min, float &r_f_t_max)
 *		- tests for ray-box intersection
 *		- ray is defined by it's origin v_org and direction v_dir
 *		- bounding box is axis-aligned and given by v_min and v_max corner coordinates
 *		- returns true on intersection, false on no intersection
 *		- if intersection occured, r_f_t_min and r_f_t_max are filled with interval
 *		  of distance along the ray that lies inside box
 */
bool CUniformGrid::b_Intersect(Vector3f v_org, Vector3f v_dir,
	Vector3f v_min, Vector3f v_max, float &r_f_t_min, float &r_f_t_max)
{
	v_min -= v_org;
	v_max -= v_org;
	// shift box so origin of ray is now O

	float f_t_min, f_t_max;
	float f_inv_x = 1.0f / v_dir.f_X();
	if(f_inv_x >= 0) {
		f_t_min = v_min.f_X() * f_inv_x;
		f_t_max = v_max.f_X() * f_inv_x;
	} else {
		f_t_max = v_min.f_X() * f_inv_x;
		f_t_min = v_max.f_X() * f_inv_x;
	}
	// calculate interval where ray goes trough y-z planes of the box

	float f_t_min_y, f_t_max_y;
	float f_inv_y = 1.0f / v_dir.f_Y();
	if(f_inv_y >= 0) {
		f_t_min_y = v_min.f_Y() * f_inv_y;
		f_t_max_y = v_max.f_Y() * f_inv_y;
	} else {
		f_t_max_y = v_min.f_Y() * f_inv_y;
		f_t_min_y = v_max.f_Y() * f_inv_y;
	}
	// calculate interval where ray goes trough x-z planes of the box

	if(f_t_min > f_t_max_y || f_t_min_y > f_t_max)
		return false;
	// in case intervals are disjunct, no intersection occurs

	if(f_t_min_y > f_t_min)
		f_t_min = f_t_min_y;
	if(f_t_max_y < f_t_max)
		f_t_max = f_t_max_y;
	// intersect overlapping intervals

	float f_t_min_z, f_t_max_z;
	float f_inv_z = 1.0f / v_dir.f_Z();
	if(f_inv_z >= 0) {
		f_t_min_z = v_min.f_Z() * f_inv_z;
		f_t_max_z = v_max.f_Z() * f_inv_z;
	} else {
		f_t_max_z = v_min.f_Z() * f_inv_z;
		f_t_min_z = v_max.f_Z() * f_inv_z;
	}
	// calculate interval where ray goes trough x-y planes of the box

	if((f_t_min > f_t_max_z) || (f_t_min_z > f_t_max))
		return false;
	// in case intervals are disjunct, no intersection occurs

	if(f_t_min_z > f_t_min)
		f_t_min = f_t_min_z;
	if(f_t_max_z < f_t_max)
		f_t_max = f_t_max_z;
	// intersect overlapping intervals to get interval where ray intersects the box

	r_f_t_min = f_t_min;
	r_f_t_max = f_t_max;
	// output the interval

	return true;
}

/*
 *	static void CUniformGrid::BoundingBox(const TTriangle *p_triangle, const TTriangle *p_end,
 *		Vector3f &r_v_min, Vector3f &r_v_max)
 *		- calculates bounding box of triangles p_triangle to p_end (not including p_end)
 *		- results are set to r_v_min and r_v_max
 */
void CUniformGrid::BoundingBox(const TTriangle *p_triangle, const TTriangle *p_end,
	Vector3f &r_v_min, Vector3f &r_v_max)
{
	if(p_triangle == p_end)
		return;
	// no triangles?

	Vector3f v_min(p_triangle->p_vertex[0]->v_position),
		v_max(p_triangle->p_vertex[0]->v_position);
	for(;p_triangle != p_end; ++ p_triangle) {
		for(int n = 0; n < 3; ++ n) {
			v_min = v_min.v_Min(p_triangle->p_vertex[n]->v_position);
			v_max = v_max.v_Max(p_triangle->p_vertex[n]->v_position);
		}
	}
	r_v_min = v_min;
	r_v_max = v_max;
	// find the bounding box
}

/*
 *								=== ~CUniformGrid ===
 */

/*
 *								=== CTerrainMesh ===
 */

/*
 *	CTerrainMesh::CTerrainMesh(const CHeightField &r_heightfield, float f_quad_scale)
 *		- creates a new terrain mesh from heightfield r_heightfield
 *		  while side of one mesh quad is f_quad_scale units long
 *		- call b_Status() to see wheter constructor succeeded
 */
CTerrainMesh::CTerrainMesh(const CHeightField &r_heightfield, float f_quad_scale)
	:m_n_vertex_num(r_heightfield.n_Width() * r_heightfield.n_Height()),
	m_p_vertex(0), m_n_triangle_num((r_heightfield.n_Width() - 1) * (r_heightfield.n_Height() - 1) * 2),
	m_p_triangle(0)
{
	if(!(m_p_vertex = new TVertex[m_n_vertex_num]))
		return;

	for(int y = 0, m = r_heightfield.n_Height(); y < m; ++ y) {
		for(int x = 0, n = r_heightfield.n_Width(); x < n; ++ x) {
			m_p_vertex[x + n * y].v_position = Vector3f((x - n / 2) * f_quad_scale,
				r_heightfield[y][x], (y - m / 2) * f_quad_scale);
			m_p_vertex[x + n * y].v_normal = Vector3f(0, 0, 0);
		}
	}
	// fill vertex array, for now keep normals zero

	if(!(m_p_triangle = new TTriangle[m_n_triangle_num]))
		return;

	TTriangle *p_triangle = m_p_triangle;
	for(int y = 0, m = r_heightfield.n_Height() - 1; y < m; ++ y) {
		for(int x = 0, n = r_heightfield.n_Width(); x + 1 < n; ++ x) {
			TVertex *p_quad[4] = {
				m_p_vertex + (x + n * y),
				m_p_vertex + (x + n * (y + 1)),
				m_p_vertex + ((x + 1) + n * (y + 1)),
				m_p_vertex + ((x + 1) + n * y)
			};
			// get indices of vertices forming quadriliteral 

			p_triangle->p_vertex[0] = p_quad[0];
			p_triangle->p_vertex[1] = p_quad[1];
			p_triangle->p_vertex[2] = p_quad[2];
			++ p_triangle;
			// emit first triangle

			p_triangle->p_vertex[0] = p_quad[2];
			p_triangle->p_vertex[1] = p_quad[3];
			p_triangle->p_vertex[2] = p_quad[0];
			++ p_triangle;
			// emit second triangle
		}
	}
	// fill triangle array

	for(TTriangle *p_triangle = m_p_triangle, *p_end = m_p_triangle +
	   m_n_triangle_num; p_triangle != p_end; ++ p_triangle) {
		Vector3f v_vert0 = p_triangle->p_vertex[0]->v_position;
		Vector3f v_edge1 = p_triangle->p_vertex[1]->v_position - v_vert0;
		Vector3f v_edge2 = p_triangle->p_vertex[2]->v_position - v_vert0;
		Vector3f v_normal = v_edge1.v_Cross(v_edge2);
		v_normal.Normalize();
		// calculate normalized face normal

		for(int n = 0; n < 3; ++ n)
			p_triangle->p_vertex[n]->v_normal += v_normal;
		// add to vertex normals
	}
	// add face normals to vertex normals

	for(TVertex *p_vertex = m_p_vertex, *p_end = m_p_vertex +
	   m_n_vertex_num; p_vertex != p_end; ++ p_vertex)
		p_vertex->v_normal.Normalize();
	// normalize vertex normals, thus averaging face
	// normals referencing each particular vertex
}

/*
 *	CTerrainMesh::~CTerrainMesh()
 *		- destructor
 */
CTerrainMesh::~CTerrainMesh()
{
	if(m_p_vertex)
		delete[] m_p_vertex;
	if(m_p_triangle)
		delete[] m_p_triangle;
}

/*
 *	bool CTerrainMesh::b_Status() const
 *		- if constructor succeeded and the mesh is ready
 *		  returs true, otherwise returns false
 */
bool CTerrainMesh::b_Status() const
{
	return m_p_vertex && m_p_triangle;
}

/*
 *	void CTerrainMesh::ResetEnvelopes(unsigned int n_value)
 *		- sets envelope values of all triangles to n_value
 */
void CTerrainMesh::ResetEnvelopes(unsigned int n_value)
{
	for(TTriangle *p_tri = m_p_triangle,
	   *p_end = m_p_triangle + m_n_triangle_num; p_tri != p_end; ++ p_tri)
		p_tri->n_envelope = n_value;
}

/*
 *	void CTerrainMesh::Draw() const
 *		- draws mesh using OpenGL, passes normals and positions only
 */
/*void CTerrainMesh::Draw() const
{
	glBegin(GL_TRIANGLES);
	for(const TTriangle *p_triangle = m_p_triangle, *p_end = m_p_triangle +
	   m_n_triangle_num; p_triangle != p_end; ++ p_triangle) {
		for(int n = 0; n < 3; ++ n) {
			Vector3f v_normal = p_triangle->p_vertex[n]->v_normal;
			Vector3f v_position = p_triangle->p_vertex[n]->v_position;
			glNormal3f(v_normal.f_X(), v_normal.f_Y(), v_normal.f_Z());
			glVertex3f(v_position.f_X(), v_position.f_Y(), v_position.f_Z());
		}
	}
	glEnd();
}*/

/*
 *								=== ~CTerrainMesh ===
 */
