/*
								+----------------------------------+
								|                                  |
								|      ***  Terrain mesh  ***      |
								|                                  |
								|    Written by Lukas Polok 2007   |
								|                                  |
								|           TerrainMesh.h          |
								|                                  |
								+----------------------------------+
*/

#ifndef __TERRAIN_MESH_INCLUDED
#define __TERRAIN_MESH_INCLUDED

/*
 *	struct TVertex
 *		- simple vertex structure, bearing position and normal vector
 */
struct TVertex {
	Vector3f v_position, v_normal;
};

/*
 *	struct TTriangle
 *		- simple face structure, bearing it's three vertices and envelope id for raytracing
 *		- capable of calculating intersection with ray (Moller-Trumbore)
 */
struct TTriangle {
	TVertex *p_vertex[3];
	unsigned int n_envelope;

	/*
	 *	bool Cast_Ray(Vector3f v_org, Vector3f v_dir,
	 *		float &r_f_t, float &r_f_u, float &r_f_v) const
	 *		- test for intersection with ray v_org, v_dir
	 *		- returns true in case intersection occured, r_f_t is filled with intersection
	 *		  distance along the ray, r_f_u and r_f_v are filled with barycentric coordinates
	 *		  along triangle's edges (to interpolate normal / texcoords / etc)
	 *		- if no intersection occurs returns false
	 *		- note this version throws away backfacing triangles, even if they'd intersect
	 */
	bool Cast_Ray(Vector3f v_org, Vector3f v_dir, float &r_f_t, float &r_f_u, float &r_f_v) const
	{
		Vector3f v_edge1 = p_vertex[1]->v_position - p_vertex[0]->v_position,
                 v_edge2 = p_vertex[2]->v_position - p_vertex[0]->v_position;
		Vector3f v_p = v_dir.v_Cross(v_edge2);
		float f_det = v_edge1.f_Dot(v_p);
		if(f_det < 1e-3f)
			return false;
		// cull backfaces

		Vector3f v_t = v_org - p_vertex[0]->v_position;
		r_f_u = v_t.f_Dot(v_p);
		if(r_f_u < -1e-2f || r_f_u > f_det + 1e-2f)
			return false;
		// first edge failed

		Vector3f v_q = v_t.v_Cross(v_edge1);
		r_f_v = v_dir.f_Dot(v_q);
		if(r_f_v < -1e-2f || r_f_u + r_f_v > f_det + 1e-2f)
			return false;
		// second edge failed

		f_det = 1.0f / f_det;
		r_f_t = v_edge2.f_Dot(v_q) * f_det;
		// calculate t

		r_f_u *= f_det;
		r_f_v *= f_det;
		// lazy rescale u, v

		return true;
	}
};

#if defined(_MSC_VER) && !defined(_MWERKS)
#define for if(0) {} else for
#endif
// msvc 'for' scoping hack

/*
 *	class CUniformGrid
 *		- uniform grid raytracing acceleration structure
 */
class CUniformGrid {
protected:
	Vector3f m_v_min, m_v_max;
	Vector3f m_v_cell_size;
	int m_p_grid_size[3];
	int *m_p_grid_data;
	int **m_p_cell_pointer_list;

public:
	/*
	 *	CUniformGrid::CUniformGrid(int n_target_cell_num,
	 *		int n_triangle_num, const TTriangle *p_triangle)
	 *		- default constructor
	 *		- creates a new uniform grid with n_target_cell_num (or propably more) cells
	 *		  from mesh with n_triangle_num triangles in p_triangle array
	 *		- call b_Status() to determine wheter constructor succeeded
	 */
	CUniformGrid(int n_target_cell_num, int n_trianlge_num, const TTriangle *p_triangle);

	/*
	 *	CUniformGrid::~CUniformGrid()
	 *		- default destructor
	 */
	~CUniformGrid();

	/*
	 *	bool CUniformGrid::b_Status() const
	 *		- returns true in case constructor succeeded, otherwise returns false
	 */
	bool b_Status() const;

	/*
	 *	template <class CCellListener>
	 *	CCellListener CUniformGrid::Traverse(Vector3f v_org, Vector3f v_dir,
	 *		CCellListener listener) const
	 *		- traverses grid, following ray v_org, v_dir
	 *		- listener is cell listener, it's passed an int array, containing array length
	 *		  as first element and then indices of faces contained in the cell
	 *		- in case listener returns true, traversal ends, otherwise it continues until
	 *		  reaching far end of the grid
	 */
	template <class CCellListener>
	CCellListener Traverse(Vector3f v_org, Vector3f v_dir, CCellListener listener) const
	{
		float f_t_min, f_t_max;
		if(!b_Intersect(v_org, v_dir, m_v_min, m_v_max, f_t_min, f_t_max))
			return listener;
		// does ray intersect the box at all?

		if(f_t_max < 0)
			return listener;
		// maybe it does, but the intersection is not in ray's direction

		if(f_t_min > 0) {
			v_org += v_dir * f_t_min;
			f_t_max -= f_t_min;
			f_t_min = 0;
		} else
			f_t_min = 0;
		// in case the ray doesn't start inside the box, shift it inside

		Vector3<int> v_voxel = v_WorldToVoxel(v_org, m_v_min, m_v_cell_size);
		// determine voxel position (conert to int to get indices)

		for(int i = 0; i < 3; ++ i) {
			if(v_voxel[i] >= m_p_grid_size[i])
				v_voxel[i] = m_p_grid_size[i] - 1;
			else 
			if(v_voxel[i] < 0)
				v_voxel[i] = 0;
		}
		// sometimes it slips off when rounding

		Vector3f v_t_max;
		for(int n = 0; n < 3; ++ n) {
			float f_raydir = v_dir[n];
			v_t_max[n] = ((v_voxel[n] + (f_raydir > 0)) *
				m_v_cell_size[n] + m_v_min[n] - v_org[n]) / f_raydir;
		}
		// determine values until voxel boundary is crossed

		Vector3f v_delta(m_v_cell_size / v_dir);
		for(int n = 0; n < 3; ++ n)
			v_delta[n] = (float)fabs(v_delta[n]);
		// determine t step to cross single cell

		Vector3<int> v_step((v_dir.f_X() > 0)? 1 : -1,
			(v_dir.f_Y() > 0)? 1 : -1, (v_dir.f_Z() > 0)? 1 : -1);
		// determine direction in which the ray is shifted

		Vector3<int> v_boundary((v_dir.f_X() > 0)? m_p_grid_size[0] : -1,
			(v_dir.f_Y() > 0)? m_p_grid_size[1] : -1, (v_dir.f_Z() > 0)? m_p_grid_size[2] : -1);
		// determine first invalid voxel indices (to break the loop)

		for(int i = 0; i < 3; ++ i) {
			if(v_voxel[i] == v_boundary[i])
				return listener;
		}
		// beginning on the boundary already?

		for(int n_width = m_p_grid_size[0], n_height = m_p_grid_size[1];;) {
			int n_cell = v_voxel.f_X() + (v_voxel.f_Y() + v_voxel.f_Z() * n_height) * n_width;
			if(listener(v_org, v_dir, m_p_cell_pointer_list[n_cell]))
				return listener;
			// pass the cell to the listener, along with ray

			bool b_x_below_y = v_t_max.f_X() < v_t_max.f_Y();
			if(b_x_below_y && v_t_max.f_X() < v_t_max.f_Z()) {
				if((v_voxel.f_X() += v_step.f_X()) == v_boundary.f_X())
					return listener;
				v_t_max.f_X() += v_delta.f_X();
			} else if(!b_x_below_y && v_t_max.f_Y() < v_t_max.f_Z()) {
				if((v_voxel.f_Y() += v_step.f_Y()) == v_boundary.f_Y())
					return listener;
				v_t_max.f_Y() += v_delta.f_Y();
			} else {
				if((v_voxel.f_Z() += v_step.f_Z()) == v_boundary.f_Z())
					return listener;
				v_t_max.f_Z() += v_delta.f_Z();
			}
			// shift along the axis where is least space
		}
	}

protected:
	static Vector3<int> v_WorldToVoxel(Vector3f v_world, Vector3f v_min, Vector3f v_cell_size);
	static bool b_Intersect(Vector3f v_org, Vector3f v_dir,
		Vector3f v_min, Vector3f v_max, float &r_f_t_min, float &r_f_t_max);
	static void BoundingBox(const TTriangle *p_triangle, const TTriangle *p_end,
		Vector3f &r_v_min, Vector3f &r_v_max);
};

/*
 *	class CTerrainMesh
 *		- simple triangular mesh, created from heightmap
 *		- offers vertex normals
 */
class CTerrainMesh {
protected:
	int m_n_vertex_num;
	TVertex *m_p_vertex;
	int m_n_triangle_num;
	TTriangle *m_p_triangle;

public:
	/*
	 *	CTerrainMesh::CTerrainMesh(const CHeightField &r_heightfield, float f_quad_scale)
	 *		- creates a new terrain mesh from heightfield r_heightfield
	 *		  while side of one mesh quad is f_quad_scale units long
	 *		- call b_Status() to see wheter constructor succeeded
	 */
	CTerrainMesh(const CHeightField &r_heightfield, float f_quad_scale);

	/*
	 *	CTerrainMesh::~CTerrainMesh()
	 *		- destructor
	 */
	~CTerrainMesh();

	/*
	 *	bool CTerrainMesh::b_Status() const
	 *		- if constructor succeeded and the mesh is ready
	 *		  returs true, otherwise returns false
	 */
	bool b_Status() const;

	/*
	 *	inline int CTerrainMesh::n_Vertex_Num() const
	 *		- returns number of mesh vertices
	 */
	inline int n_Vertex_Num() const
	{
		return m_n_vertex_num;
	}

	/*
	 *	inline int CTerrainMesh::n_Triangle_Num() const
	 *		- returns number of mesh triangles
	 */
	inline int n_Triangle_Num() const
	{
		return m_n_triangle_num;
	}

	/*
	 *	inline TVertex *CTerrainMesh::p_Vertex()
	 *		- returns pointer to mesh vertices (don't delete it)
	 */
	inline TVertex *p_Vertex()
	{
		return m_p_vertex;
	}

	/*
	 *	inline const TVertex *CTerrainMesh::p_Vertex() const
	 *		- returns const pointer to mesh vertices
	 */
	inline const TVertex *p_Vertex() const
	{
		return m_p_vertex;
	}

	/*
	 *	inline TTriangle *CTerrainMesh::p_Triangle()
	 *		- returns pointer to mesh triangles (don't delete it)
	 */
	inline TTriangle *p_Triangle()
	{
		return m_p_triangle;
	}

	/*
	 *	inline const TTriangle *CTerrainMesh::p_Triangle() const
	 *		- returns const pointer to mesh triangles
	 */
	inline const TTriangle *p_Triangle() const
	{
		return m_p_triangle;
	}

	/*
	 *	void CTerrainMesh::ResetEnvelopes(unsigned int n_value)
	 *		- sets envelope values of all triangles to n_value
	 */
	void ResetEnvelopes(unsigned int n_value);
};

#endif //__TERRAIN_MESH_INCLUDED
